#include <math.h>

#include "api3d_bl.h"

#ifdef FILTER_MODE_TENT
#define FILTER_MODE_GAUSS
#endif

float
abs_w( float v, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return fabs( v );
#ifdef FILTER_MODE_RECT
    float a = v + w * 0.5f;
    float b = v - w * 0.5f;
    return ( a * fabs( a ) - b * fabs( b ) ) / ( 2.0f * w );
#endif
#ifdef FILTER_MODE_GAUSS
    return v * erf( v / ( w * sqrt( 2.0f ) ) ) + w * sqrt( 2.0f / M_PI ) * exp( - v*v / (2.0f * w*w) );
#endif
}

float
ceil_w( float a, float w )
{
#ifdef FILTER_MODE_NONE
    return ceil_f( a );
#else
    return floor_w( a, w ) + 1.0f;
#endif
}

float
cos_w( float x, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return cos( x );

#ifdef FILTER_MODE_RECT
    return cos( x ) * sin( w * 0.5f ) / ( w * 0.5f );
#endif
#ifdef FILTER_MODE_GAUSS
    return cos( x ) * exp( - w*w * 0.5f );
#endif
#ifdef DISABLE_MODE_SINC
    return cos( x ) * step_f( 0, M_PI - w );
#endif
}

float
floor_w( float a, float w )
{
#ifdef FILTER_MODE_NONE
    return floor_f( a );
#else
    return a - fract_w( a, w );
#endif
}

#ifdef FILTER_MODE_TENT
#undef FILTER_MODE_GAUSS
#endif

float
fract_w( float x, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return x - floor_f( x );
#ifdef FILTER_MODE_RECT
    float u = floor_f( x - w * 0.5f );
    float v = floor_f( x + w * 0.5f );
    return ( ( w - v + u ) * ( x - ( v + u ) * 0.5f ) + ( v - u ) * 0.5f ) / w;
#endif
#ifdef DISABLE_MODE_TENT
    float u = x - w;
    float v = x + w;
    float uf = u - floor( u );
    float vf = v - floor( v );
    float xf = x - floor( x );
    return (
        (3*u*u + 2*uf*uf*uf - 3*uf*uf + uf - u) +
        (3*v*v + 2*vf*vf*vf - 3*vf*vf + vf - v) -
        (3*x*x + 2*xf*xf*xf - 3*xf*xf + xf - x) * 2.0f
    ) / ( 12.0f * w * w );
#endif
#ifdef FILTER_MODE_TENT
    __attribute__((opencl(force_double))) double xd = x;
    __attribute__((opencl(force_double))) double wd = w;
    __attribute__((opencl(force_double))) double uf = floor_f( x - w );
    __attribute__((opencl(force_double))) double vf = floor_f( x + w );
    __attribute__((opencl(force_double))) double xf = floor_f( x );
    __attribute__((opencl(force_double))) double result =
        xd - 0.5 * (uf + vf) + (uf - vf) * (xd - 0.5) / wd + (
            (uf + vf - 2.0*xf) * (- 6.0*xd*xd + 6.0*xd - 1.0) +
            (uf*uf + vf*vf - 2.0*xf*xf) * (6.0*xd - 3.0) +
            (uf*uf) * (- 6.0*wd - 2.0*uf) +
            (vf*vf) * (+ 6.0*wd - 2.0*vf) +
            (4.0*xf*xf*xf)
        ) / ( 12.0 * wd * wd );
    return (float) result;
#endif
#ifdef FILTER_MODE_GAUSS
    unsigned int k;
    float y = 0.5;
    float dy = 1.0f;
    for ( k = 1; k < 12; ++k ) {
        dy = exp( -2 * w*w * M_PI*M_PI * k*k ) / ( k * M_PI );
        y -= dy * sin( 2 * M_PI * k * x );
    }
    return y;
#endif
#ifdef FILTER_MODE_SINC
    float y = 0.5f;
    unsigned int k = 1;
    while ( k < 0.5 / w && k < 40 ) {
        y -= sin( 2 * M_PI * k * x ) / ( k * M_PI );
        k += 1;
    }
    return y;
#endif
}

#ifdef FILTER_MODE_TENT
#define FILTER_MODE_GAUSS
#endif

float
saturate_w( float x, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return saturate_f( x );
#ifdef FILTER_MODE_RECT
    w = w * 0.5f;
    float a = fmax( 0.0f, x + w );
    float b = fmax( 0.0f, x - w );
    float c = fmax( 0.0f, x + w - 1.0f );
    float d = fmax( 0.0f, x - w - 1.0f );
    return ( a*a - b*b + d*d - c*c ) / ( 4.0f * w ); 
#endif
#ifdef FILTER_MODE_GAUSS
    float wsqrt2 = w * sqrt( 2.0f );
    float wsq2   = w * w * 2.0f;
    float xm1    = x - 1.0f;
    float a = x * erf( x / wsqrt2 );
    float b = xm1 * erf( xm1 / wsqrt2 );
    float c = w * sqrt( 2.0f / M_PI ) * (
            exp( - x*x / wsq2 ) - exp( - xm1*xm1 / wsq2 )
        );
    return ( a - b + c + 1.0f ) * 0.5f;
#endif
}

float
sin_w( float x, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return sin( x );

#ifdef FILTER_MODE_RECT
    return sin( x ) * sin( w * 0.5f ) / ( w * 0.5f );
#endif
#ifdef FILTER_MODE_GAUSS
    return sin( x ) * exp( - w*w * 0.5f );
#endif
#ifdef DISABLE_MODE_SINC
    return sin( x ) * step_f( 0, M_PI - w );
#endif
}

float
sq_w( float x, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return x*x;
#ifdef FILTER_MODE_RECT
    return x*x + w*w/12.0f;
#endif
#ifdef FILTER_MODE_GAUSS
    return x*x + w*w;
#endif
}

float
step_w( float a, float x, float w )
{
#ifndef FILTER_MODE_NONE
    if ( w <= 0.0f )
#endif
        return step_f( a, x );
#ifdef FILTER_MODE_RECT
    return ( fmax( a, x + w * 0.5f ) - fmax( a, x - w * 0.5f ) ) / w;
#endif
#ifdef FILTER_MODE_GAUSS
    return 0.5f + 0.5f * erf( ( x - a ) / ( w * sqrt( 2.0f ) ) );
#endif
}

#ifdef FILTER_MODE_TENT
#undef FILTER_MODE_GAUSS
#endif
