#ifndef GABOR_H
#define GABOR_H

#include <math.h>
#include <stdlib.h>

#include "api3d.h"
#include "factors.h"
#include "globals.h"
#include "prng.h"

float
gabor( float K, float a, float F_0, float omega_0, float x, float y )
{
    float gaussian = K * exp( -M_PI * a*a * (x*x + y*y) );
    float sinusoid =
        cos( 2.0f * M_PI * F_0 * ( x * cos(omega_0) + y * sin(omega_0) ) );
    return gaussian * sinusoid;
}

float
gabor_noise_impl(
    float K, float a,
    float F_0, float omega_0,
    float impulses, unsigned period,
    float x, float y
) {
    float radius = sqrt( -log(0.05f) / M_PI ) / a;
    x /= radius;
    y /= radius;
    float int_x = floor( x );
    float int_y = floor( y );
    float frac_x = x - int_x;
    float frac_y = y - int_y;
    int i = (int) int_x;
    int j = (int) int_y;
    float noise = 0.0f;
    for ( int di = -1; di <= 1; ++di ) {
        for ( int dj = -1; dj <= 1; ++dj ) {
            unsigned s = ( ( (unsigned)(j + dj) % period ) * period ) + ( (unsigned)(i + di) % period );
            float number_of_impulses_per_cell = impulses / M_PI;
            unsigned number_of_impulses = prng_poisson( &s, number_of_impulses_per_cell );
            for ( unsigned k = 0; k < number_of_impulses; ++k ) {
                float x_k = prng_uniform_0_1( &s );
                float y_k = prng_uniform_0_1( &s );
                float w_k = prng_uniform( &s, -1.0f, 1.0f );
                float omega_0_k = prng_uniform( &s, 0.0f, 2.0f * M_PI );
                float x_k_x = (frac_x - di) - x_k;
                float y_k_y = (frac_y - dj) - y_k;
                if ( (x_k_x * x_k_x) + (y_k_y * y_k_y) < 1.0f ) {
                    //noise += w_k * gabor( K, a, F_0, omega_0, x_k_x * radius, y_k_y * radius ); // anisotropic
                    noise += w_k * gabor( K, a, F_0, omega_0_k, x_k_x * radius, y_k_y * radius ); // isotropic
                }
            }
        }
    }
    return noise;
}

float
gabor_noise(
    float K, float a, float F_0, float omega_0,
    float impulses, unsigned period,
    float x, float y
)
__attribute__((noinline))
{
    float var = (K*K) / (4.0f*a*a) * (1.0f + exp(-2.0f*M_PI*F_0*F_0 / (a*a)));
    var *= 1.0f / 3.0f * impulses / (-log(0.05)/(a*a));
    float scale = 3.0f * sqrt(var);

    return gabor_noise_impl(
        K, a, F_0, omega_0,
        impulses, period,
        x, y
    ) / scale;
}

float
filtered_gabor_noise(
    float K, float a, float F_0, float omega_0,
    float impulses, unsigned period,
    float x, float y,
    float dx, float dy
)
__attribute__((noinline))
{
    float var = (K*K) / (4.0f*a*a) * (1.0f + exp(-2.0f*M_PI*F_0*F_0 / (a*a)));
    var *= 1.0f / 3.0f * impulses / (-log(0.05)/(a*a));
    float scale = 3.0f * sqrt(var);

#ifndef FILTER_MODE_NONE
    float sigmasq = GAUSS_SIGMA_FACTOR * GAUSS_SIGMA_FACTOR;
    float Sigma_G = a*a / (2.0f*M_PI);
    float Sigma_FG_x = a*a / (4.0f*M_PI*M_PI*sigmasq*dx*dx*a*a + 2.0f*M_PI);
    float Sigma_FG_y = a*a / (4.0f*M_PI*M_PI*sigmasq*dy*dy*a*a + 2.0f*M_PI);
    float Sigma_x = Sigma_G + 1.0f/(4*M_PI*M_PI*sigmasq*dx*dx);
    float Sigma_y = Sigma_G + 1.0f/(4*M_PI*M_PI*sigmasq*dy*dy);
    float mu_G_x = F_0 * cos( omega_0 );
    float mu_G_y = F_0 * sin( omega_0 );
    float mu_FG_x = Sigma_FG_x / Sigma_G * mu_G_x;
    float mu_FG_y = Sigma_FG_y / Sigma_G * mu_G_y;

    float apsq = 2 * M_PI * sqrt( Sigma_FG_x * Sigma_FG_y );

    float ap = sqrt( apsq );
    float Kp = K * apsq / (a*a) * exp( -0.5f*(mu_G_x*mu_G_x/Sigma_x + mu_G_y*mu_G_y/Sigma_y) );
    float F_0p = sqrt( mu_FG_x*mu_FG_x + mu_FG_y*mu_FG_y );
    float omega_0p = atan2( mu_FG_y, mu_FG_x );
#else
    float Kp = K;
    float ap = a;
    float F_0p = F_0;
    float omega_0p = omega_0;
#endif

    return gabor_noise_impl(
        Kp, ap, F_0p, omega_0p,
        impulses, period,
        x, y
    ) / scale;
}

#endif
