#include <math.h>

#include "api3d.h"
#include "factors.h"
#include "procedural_shaders.h"
#include "procedural_helpers.h"
#include "reprojection.h"

/* Frame order:
 * FILTER_MODE_GABOR
 * FILTER_MODE_NONE
 */

#include "gabor.h"

#include "api3d_bl.c"

/*
 * Renders a brick wall.
 */
float3
calcBlockPos(
    __global global_buffers * g,
    VS_OUTPUT * in,
    float3 * gridIn,
    float * sqGridOut,
    int staggered
) {
    float3 mortarwidth;
    float mortarLines, gridHard;
    float height, bumpAmount;

    *gridIn = mul_f3f( in->worldPos, 3.85f );
    gridIn->y *= 2.0f;

    gridHard = 500.0f;
    mortarwidth = makef3_fff( 0.07f, 0.07f, 0.07f );
    /* Choose grid pattern based upon whether we want staggered or regular */
    if ( staggered )
        height = staggeredGrid( *gridIn, mortarwidth, gridHard, makef3_fff( 0.0f, 1.0f, 0.0f ), makef3_fff( 1.0f, 0.0f, 1.0f ) );
    else
        height = grid( *gridIn, mortarwidth, gridHard );

    /* Make our lines sharper than they would normally be */
    *sqGridOut = 1.0f - pow( height, 16.0f );
    height = 1.0f - height;

    bumpAmount = 4.0f;
    return add_f3f3( in->worldPos, mul_f3f( in->worldNorm, height * bumpAmount ) );
}

float
gaborGridColor( float3 p )
{
    p = div_f3f( floor_f3( p ), 255.0f );
    p = add_f3f( p, 10.0f * p.z );

    float K = 1.0f;
    float a = 30.0f;
    float F_0 = 200.0f;
    float omega_0 = 0.0f;
    float number_of_impulses_per_kernel = 64.0f;
    unsigned period = 256;

    return 0.5f + 0.5f * gabor_noise(
        K, a, F_0, omega_0,
        number_of_impulses_per_kernel, period,
        p.x, p.y
    );
}

float
gaborStaggeredGridColor( float3 p, float3 staggerOnMask, float3 staggerByMask )
{
    float3 test;
    int3 intTest, mod2s;
    float maxMod;

    test = abs_f3( floor_f3( mul_f3f3( p, staggerOnMask ) ) );
    intTest = int3_f3( test );
    mod2s = and_i3i( intTest, 1 );
    maxMod = mod2s.x + mod2s.y + mod2s.z;
    p = add_f3f3( p, mul_f3f( staggerByMask, 0.5f * maxMod ) );

    return gaborGridColor( p );
}

/*
 * Renders a brick wall.
 */

float4
gabor_brick(
    __global global_buffers * g,
    VS_OUTPUT * in,
    float2 offset,
    __global float * c
) {
    axis_align_widths( &( in->dwdx ), &( in->dwdy ) );
    float x = in->objPos.x * 3.85f;
    float y = in->objPos.y * 3.85f;

    float rx, ry, noise;

    float4 color = makef4_f( 0.0f );
    float4 brick = makef4_ffff( 0.25f, 0.03f, 0.01f, 1.0f );
    float4 brick2 = makef4_ffff( 0.7f, 0.02f, 0.05f, 1.0f );
    float4 mortarcolor = makef4_ffff( 0.7f, 0.7f, 0.7f, 1.0f );

    float K = 1.0f;
#ifdef ROW2
    float a = 32.0f;
    float F_0 = 2.0f;
#else
    float a = 0.05f;
    float F_0 = 1.0f;
#endif
    float omega_0 = 0.0f;
    float number_of_impulses_per_kernel = 64.0f;
    unsigned period = 256;

#ifdef FILTER_MODE_GABOR
    noise = 0.5f + 0.5f * gabor_noise(
        K, a, F_0, omega_0,
        number_of_impulses_per_kernel, period,
        x, y
    );
    return lerp_f4f( brick, brick2, noise );
#else
    rx = floor_f( x + 0.93f ) - floor_f( x );
    ry = floor_f( y + 0.465f ) - floor_f( y );
    noise = 0.5f + 0.5f * gabor_noise(
        K, a, F_0, omega_0,
        number_of_impulses_per_kernel, period,
        floor_f( x ), floor_f( y )
    );

    color = add_f4f4( color, mul_f4f( lerp_f4f( brick, brick2, noise ), rx*ry ) );

    rx = floor_f( x + 1.43f ) - floor_f( x + 0.5f );
    ry = floor_f( y + 0.965f ) - floor_f( y + 0.5f );
    noise = 0.5f + 0.5f * gabor_noise(
        K, a, F_0, omega_0,
        number_of_impulses_per_kernel, period,
        floor_f( x + 0.5f), floor_f( y + 0.5f )
    );
    color = add_f4f4( color, mul_f4f( lerp_f4f( brick, brick2, noise ), rx*ry ) );

    ry = floor_f( y + 0.5f ) - floor_f( y + 0.465f );
    color = add_f4f4( color, mul_f4f( mortarcolor, ry ) );

    ry = floor_f( y + 1.0f ) - floor_f( y + 0.965f );
    color = add_f4f4( color, mul_f4f( mortarcolor, ry ) );

    rx = floor_f( x + 1.0f ) - floor_f( x + 0.93f );
    ry = floor_f( y + 0.465f ) - floor_f( y );
    color = add_f4f4( color, mul_f4f( mortarcolor, rx*ry ) );

    rx = floor_f( x + 1.5f ) - floor_f( x + 1.43f );
    ry = floor_f( y + 0.965f ) - floor_f( y + 0.5f );
    color = add_f4f4( color, mul_f4f( mortarcolor, rx*ry ) );

    return color;
#endif
}

