#include <math.h>

#include "api3d.h"
#include "factors.h"
#include "procedural_shaders.h"
#include "procedural_helpers.h"
#include "reprojection.h"

/* Frame order:
 *   FILTER_MODE_FBM
 *   FILTER_MODE_NONE
 */

#include "gabor.h"

#include "api3d_bl.c"

float
gaborGridColor( __global global_buffers * g, float3 p )
{
    p = div_f3f( floor_f3( p ), 255.0f );
    p = add_f3f( p, 10.0f * p.z );

#ifdef FILTER_MODE_FBM
    return sample_f_f2( g, swizzle_f3xy( p ) );
#else
    float K = 1.0f;
    float a = 30.0f;
    float F_0 = 200.0f;
    float omega_0 = 0.0f;
    float number_of_impulses_per_kernel = 64.0f;
    unsigned period = 256;

    return 0.5f + 0.5f * gabor_noise(
        K, a, F_0, omega_0,
        number_of_impulses_per_kernel, period,
        p.x, p.y
    );
#endif
}

float
gaborStaggeredGridColor(
    __global global_buffers * g, float3 p,
    float3 staggerOnMask, float3 staggerByMask
) {
    float3 test;
    int3 intTest, mod2s;
    float maxMod;

    test = abs_f3( floor_f3( mul_f3f3( p, staggerOnMask ) ) );
    intTest = int3_f3( test );
    mod2s = and_i3i( intTest, 1 );
    maxMod = mod2s.x + mod2s.y + mod2s.z;
    p = add_f3f3( p, mul_f3f( staggerByMask, 0.5f * maxMod ) );

    return gaborGridColor( g, p );
}

float
mylength( float2 pos )
{
    return abs_f( pos.x ) + abs_f( pos.y );
}

float
mymin( float a, float b )
{
    return ( a + b - abs_f( a - b ) ) * 0.5f;
}

float4
gabor_wood(
    __global global_buffers * g,
    VS_OUTPUT * in,
    float2 offset,
    __global float * c
) {
    float4 lightBrown, darkBrown;
    float4 wiggleNoiseIn, radialNoiseIn, colorOut;
    float3 worldNorm, centerPos, max, min;
    float outBlock, wiggleNoise, toLine, radialNoise;
    float dist0, dist1, dist2, dist3;
    const float centerScaler = 15.0f;

    axis_align_widths( &( in->dwdx ), &( in->dwdy ) );

    float K = 1.0f;
    float a = 1.4f;
    float F_0 = 0.7f;
    float omega_0 = M_PI / 4.0f;

    lightBrown = mul_f4f( makef4_ffff( 0.6f, 0.4f, 0.20f, 1.0f ), 0.8f );
    darkBrown  = mul_f4f( makef4_ffff( 0.2f, 0.1f, 0.05f, 1.0f ), 0.7f );

    in->objPos = shiftValue_f3( in, in->objPos, offset );

    /* Add some random variation on a per block (or plank) basis */
    /*
    outBlock = gaborStaggeredGridColor( g,
        mul_f3f3( in->objPos, makef3_fff( 3.8, 1.0, 1.0 ) ),
        makef3_fff( 1.0f, 0.0f, 1.0f ),
        makef3_fff( 0.0f, 1.0f, 0.0f )
    );
    */
    outBlock = gaborGridColor(
        g, mul_f3f3( in->objPos, makef3_fff( 3.8f, 1.0f, 1.0f ) )
    );

    wiggleNoiseIn = makef4_f3f( mul_f3f( in->objPos, 2.0f ), 0.0f );
#ifdef FILTER_MODE_FBM
    wiggleNoise = fBm( g, wiggleNoiseIn, 2.0f, 2.0f, 0.5f, NT_SIMPLEX3D );
#else
    wiggleNoise = gabor_noise(
        K, a, F_0, omega_0, 64.0f, 256,
        wiggleNoiseIn.x, wiggleNoiseIn.y
    );
#endif

    /* Then we'll shift the center of where the noise is for that block of wood */
    centerPos = mul_f3f( add_f3f( in->objPos, wiggleNoise * 0.06f ), 2.15f );

    /* We also want to shift the noise lookup on a per block basis */
    centerPos = add_f3f( centerPos, outBlock * 1.25f );
    max = ceil_f3( centerPos );
    min = floor_f3( centerPos );

    /* Find the distance to the center of the noise */
    dist0 = mylength( sub_f2f2( swizzle_f3xz( centerPos ), swizzle_f3xz( min ) ) );
    dist1 = mylength( sub_f2f2( swizzle_f3xz( centerPos ), makef2_ff( min.x, max.z ) ) );
    dist2 = mylength( sub_f2f2( swizzle_f3xz( centerPos ), makef2_ff( max.x, min.z ) ) );
    dist3 = mylength( sub_f2f2( swizzle_f3xz( centerPos ), swizzle_f3xz( max ) ) );

    toLine = mymin( mymin( dist0, dist1 ), mymin( dist2, dist3 ) );

    /* Look up some radial noise using the distance to this center */
    radialNoiseIn = makef4_f3f( makef3_f( toLine * 12.0f ), 0.0f );
#ifdef FILTER_MODE_FBM
    radialNoise = fBm( g, radialNoiseIn, 3, 2, 0.5f, NT_SIMPLEX3D );
#else
    radialNoise = gabor_noise(
        K, a, F_0, omega_0, 64.0f, 256,
        radialNoiseIn.x, radialNoiseIn.y
    );
#endif

    /* Alternate between some brownish wood colors */
    colorOut = lerp_f4f( lightBrown, darkBrown, saturate_f( radialNoise + 0.5f ) );
    colorOut.w = 1.0f;
    return colorOut;
}

