#ifndef NOISE_H
#define NOISE_H

#include <stdlib.h>

#include "api3d.h"
#include "globals.h"

#define TEXTURE_VEC_SIZE
#include "texture.h"
#undef TEXTURE_VEC_SIZE

#define TEXTURE_VEC_SIZE 4
#include "texture.h"
#undef TEXTURE_VEC_SIZE

/*
 * File: noise.h
 *
 * Port of Noise.fx to C. Noise.fx header follows:
 *
 * The noise effect file for the ProceduralMaterials sample.  This file
 * implements both Simplex noise in 3D and 4D and voronoi noise.
 *
 * For a good explanation of the simplex noise algorithm see
 * http://staffwww.itn.liu.se/~stegu/simplexnoise/simplexnoise.pdf
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

static const float g_TexDims = 256;
static const float g_TexRandomDims = 1024;

const float3 g_grad3[ 16 ] = {
    { 1, 1, 0 }, { -1, 1, 0 }, { 1, -1, 0 }, { -1, -1, 0 },
    { 1, 0, 1 }, { -1, 0, 1 }, { 1, 0, -1 }, { -1, 0, -1 },
    { 0, 1, 1 }, { 0, -1, 1 }, { 0, 1, -1 }, { 0, -1, -1 },

    /* Wrap around to let us use an AND instead of a MOD */
    { 1, 1, 0 }, { -1, 1, 0 }, { 1, -1, 0 }, { -1, -1, 0 }
};

const float4 g_grad4[ 32 ] = {
    { 0,  1, 1, 1 }, { 0,  1,  1, -1 }, { 0,  1, -1, 1 }, { 0,  1, -1, -1 },
    { 0, -1, 1, 1 }, { 0, -1,  1, -1 }, { 0, -1, -1, 1 }, { 0, -1, -1, -1 },
    { 1,  0, 1, 1 }, { 1,  0,  1, -1 }, { 1,  0, -1, 1 }, { 1,  0, -1, -1 },
    {-1,  0, 1, 1 }, {-1,  0,  1, -1 }, {-1,  0, -1, 1 }, {-1,  0, -1, -1 },
    { 1,  1, 0, 1 }, { 1,  1,  0, -1 }, { 1, -1,  0, 1 }, { 1, -1,  0, -1 },
    {-1,  1, 0, 1 }, {-1,  1,  0, -1 }, {-1, -1,  0, 1 }, {-1, -1,  0, -1 },
    { 1,  1, 1, 0 }, { 1,  1, -1,  0 }, { 1, -1,  1, 0 }, { 1, -1, -1,  0 },
    {-1,  1, 1, 0 }, {-1,  1, -1,  0 }, {-1, -1,  1, 0 }, {-1, -1, -1,  0 }
};

/*
 * Find which simplex we are in by magnitude sorting (3D case). This fast code
 * for finding the simplex comes from the Perlin Fire sample from NVidia
 * Corporation. For more information about it, visit NVidia's developer website.
 * http://developer.download.nvidia.com/whitepapers/2007/SDK10/PerlinFire.pdf
 */
void
simplex3d( float3 p, float3 * simplex_0, float3 * simplex_1, float3 * simplex_2, float3 * simplex_3 )
{
    float3 t;
    t = float3_i3( ge_f3f3( swizzle_f3xzy( p ), swizzle_f3yxz( p ) ) );
    *simplex_0 = makef3_f( 0 );
    *simplex_1 = float3_i3( sub_ii3( 0, gt_f3f3( swizzle_f3xzy( t ), swizzle_f3yxz( t ) ) ) );
    *simplex_2 = float3_i3( sub_ii3( 0, ge_f3f3( swizzle_f3xzy( t ), swizzle_f3yxz( t ) ) ) );
    *simplex_3 = makef3_f( 1 );
}

/*
 * Find which simplex we are in by magnitude sorting (4D case). This fast code
 * for finding the simplex comes from the Perlin Fire sample from NVidia
 * Corporation. For more information about it, visit NVidia's developer website.
 */
void
simplex4d( float4 p, float4 simplex[ 5 ] )
{
    float4 offset0;
    float3 isX;
    float2 isY;
    float  isZ;

    isX = step_f3( swizzle_f4yzw( p ), swizzle_f4xxx( p ) );
    offset0.x = dot_f3( isX, makef3_fff( 1, 1, 1 ) );
    swizzleset_f4yzw( &offset0, sub_ff3( 1, isX ) );

    isY = step_f2( swizzle_f4zw( p ), swizzle_f4yy( p ) );
    offset0.y += dot_f2( isY, makef2_ff( 1, 1 ) );
    swizzleset_f4zw( &offset0, add_f2f2( swizzle_f4zw( offset0 ), sub_ff2( 1, isY ) ) );

    isZ = step_f( p.w, p.z );
    offset0.z += isZ;
    offset0.w += 1 - isZ;

    //offset0 now contains the unique values 0,1,2,3 in each channel
    simplex[ 4 ] = makef4_f( 1 );
    simplex[ 3 ] = saturate_f4( offset0 );
    simplex[ 2 ] = saturate_f4( add_f4f( offset0, -1 ) );
    simplex[ 1 ] = saturate_f4( add_f4f( offset0, -2 ) );
    simplex[ 0 ] = makef4_f( 0 );
}

float
simplexNoise3d( __global global_buffers * g, float3 p0 )
{
    #define F3 0.333333333333f    /* 1 / 3: magic number for 3D space noise */
    #define G3 0.166666666667f    /* 1 / 6: magic number for 3D space noise */

    float u, v, totalNoise, t;
    float3 pi, po, tex;
    float3 p_0, p_1, p_2, p_3, corner_0, corner_1, corner_2, corner_3;
    float index;
    int i;

    /* Skew input space to determine which simplex we're in */
    u = ( p0.x + p0.y + p0.z ) * F3;
    pi = floor_f3( add_f3f( p0, u ) );
    v = ( pi.x + pi.y + pi.z ) * G3;

    /* Unskew the cell origin back to < x, y, z > space */
    po = add_f3f( pi, -v );

    p_0 = sub_f3f3( p0, po );  /* The x, y, z distances from the cell origin */

    /* For the 3d case, the simplex shape is a tetrahedron. Find the 4 corners
     * of the simplex */
    simplex3d( p_0, &corner_0, &corner_1, &corner_2, &corner_3 );

    p_1 = add_f3f( sub_f3f3( p_0, corner_1 ), 1 * G3 );
    p_2 = add_f3f( sub_f3f3( p_0, corner_2 ), 2 * G3 );
    p_3 = add_f3f( sub_f3f3( p_0, corner_3 ), 3 * G3 );

    totalNoise = 0;

    // i = 0
    tex = add_f3f3( pi, corner_0 );
    tex = div_f3f( tex, g_TexDims );

    /* We want an index, but the random values are normalized... so mul by
     * 255 */
    index = samplef_ff(
        g->g_txRand_f, g->sizex, g->sizey, tex.x + tex.z, tex.x + tex.z
    ) * 255.0f;
    index = 16.0f * fract_f( index / 16.0f );

    t = saturate_f( 0.6f - dot_f3( p_0, p_0 ) );
    t *= t;
    totalNoise += t * t * dot_f3( g_grad3[ (int)index ], p_0 );

    // i = 1
    tex = add_f3f3( pi, corner_1 );
    tex = div_f3f( tex, g_TexDims );

    /* We want an index, but the random values are normalized... so mul by
     * 255 */
    index = samplef_ff(
         g->g_txRand_f, g->sizex, g->sizey, tex.x + tex.z, tex.y + tex.z
    ) * 255.0f;
    index = 16.0f * fract_f( index / 16.0f );

    t = saturate_f( 0.6f - dot_f3( p_1, p_1 ) );
    t *= t;
    totalNoise += t * t * dot_f3( g_grad3[ (int)index ], p_1 );

    // i = 2
    tex = add_f3f3( pi, corner_2 );
    tex = div_f3f( tex, g_TexDims );

    /* We want an index, but the random values are normalized... so mul by
     * 255 */
    index = samplef_ff(
        g->g_txRand_f, g->sizex, g->sizey, tex.x + tex.z, tex.y + tex.z
    ) * 255.0f;
    index = 16.0f * fract_f( index / 16.0f );

    t = saturate_f( 0.6f - dot_f3( p_2, p_2 ) );
    t *= t;
    totalNoise += t * t * dot_f3( g_grad3[ (int)index ], p_2 );

    // i = 3
    tex = add_f3f3( pi, corner_3 );
    tex = div_f3f( tex, g_TexDims );

    /* We want an index, but the random values are normalized... so mul by
     * 255 */
    index = samplef_ff(
        g->g_txRand_f, g->sizex, g->sizey, tex.x + tex.z, tex.y + tex.z
    ) * 255.0f;
    index = 16.0f * fract_f( index / 16.0f );

    t = saturate_f( 0.6f - dot_f3( p_3, p_3 ) );
    t *= t;
    totalNoise += t * t * dot_f3( g_grad3[ (int)index ], p_3 );

    /* Scale by 32 to get us into unit range */
    return 32.0f * totalNoise;
}

float
simplexNoise4d( __global global_buffers * g, float4 p0 )
{
    #define F4 0.309016994375f /* ( sqrt( 5 ) - 1 ) / 4 */
    #define G4 0.138196601125f /* ( 5 - sqrt( 5 ) ) / 20 */

    float u, v, totalNoise, t;
    float4 pi, po, tex;
    float4 p[ 5 ], corner[ 5 ];
    int i;
    unsigned int index;

    /* Skew the (x,y,z) space to determine which cell of 24 simplices we're in */
    u = ( p0.x + p0.y + p0.z + p0.w ) * F4;
    pi = floor_f4( add_f4f( p0, u ) );
    v = ( pi.x + pi.y + pi.z + pi.w ) * G4;

    /* Unskew the cell origin back to < x, y, z, w > space */
    po = add_f4f( pi, -v );

    /* The x, y, z, w distances from the cell origin */
    p[ 0 ] = sub_f4f4( p0, po );

    /* To find out which of the 24 possible simplexes we're in */
    simplex4d( p[ 0 ], corner );

    p[ 1 ] = add_f4f( sub_f4f4( p[ 0 ], corner[ 1 ] ), 1 * G4 );
    p[ 2 ] = add_f4f( sub_f4f4( p[ 0 ], corner[ 2 ] ), 2 * G4 );
    p[ 3 ] = add_f4f( sub_f4f4( p[ 0 ], corner[ 3 ] ), 3 * G4 );
    p[ 4 ] = add_f4f( sub_f4f4( p[ 0 ], corner[ 4 ] ), 4 * G4 );

    totalNoise = 0;

    for ( i = 0; i < 5; ++i ) {
        tex = add_f4f4( pi, corner[ i ] );
        tex = div_f4f( tex, g_TexDims );

        /* We want an index, but the random values are normalized... so mul by
         * 255 */
        index = samplef_ff(
            g->g_txRand_f, g->sizex, g->sizey, tex.x + tex.z, tex.y + tex.w
        ) * 255.0f;
        index &= 31;

        t = saturate_f( 0.6f - dot_f4( p[ i ], p[ i ] ) );
        t *= t;
        totalNoise += t * t * dot_f4( g_grad4[ index ], p[ i ] );
    }

    /* Scale by 32 to get us into unit range */
    return 27.0f * totalNoise;
}

/* Noise types */
#define NT_SIMPLEX3D            0    /* Standard 3D simplex noise */
#define NT_SIMPLEX4D            1    /* 4D simplex noise; good for animating */
#define NT_SIMPLEX3D_TURB       2    /* 3D simplex noise with turbulence */
#define NT_SIMPLEX4D_TURB       3    /* 4D simplex noise with turulence */
#define NT_VORONOI3             4    /* Voronoi noise evaluated over a 3x3x3 grid */
#define NT_VORONOI3_2           5    /* Voronoi noise over a 3x3x3 grid, but using the second distance function */
#define NT_VORONOI2             6    /* Voronoi noise evaluated over a 2x2x2 grid (faster, but some artifacts) */
#define NT_VORONOI2_2           7    /* Voronoi noise over a 2x2x2 grid, but using the second distance function */


/*
 * Voronoi noise overview
 *
 * For each sample position, we are trying to determine the distance to the
 * closest point in a collection of randomly placed points in 3D space. Because
 * iterating over a series of points in space would take a very LONG time, we
 * divide space up into a regular grid. The coordinates of each cell of this
 * grid are used to create a hash that indexes into a set of random vectors
 * stored in g_txRandVector.  This random vector corresponds to the location of
 * a point inside that grid cell.  For each point in space, we examine all 9
 * surrounding grid cells and find the closest point for all 9 cells.  The
 * methods that implement this routine are the 3 methods (for 3x3x3 surrounding
 * cells).
 *
 * We also use a slightly faster method labelled the 2 method.  This method only
 * looks at 4 surrounding cells.  This means that we could miss the closest
 * point in some cases. To increase the odds that we'll find the closest point,
 * each cell in the 2X2 method contains 3 random points.  It's still not
 * foolproof, but it's faster than the 3X3 method.
 */


/*
 * W is a random variation to the lookup.  Changing W with time will animate the
 * Voronoi noise.
 */
float3
getCellPoint( __global global_buffers * g, float3 cell, float w )
{
    float3 newCell, randPt;
    float hashIndex;

    newCell = cell;
    /* This is the hash for which random index corresponds to which cell */
    hashIndex = ( ( newCell.z / newCell.y ) * newCell.x ) + ( newCell.x * newCell.y );
    hashIndex += w;
    hashIndex /= g_TexRandomDims;
    randPt = swizzle_f4xyz(
        samplef4_ff( g->g_txRand_f4, g->sizex, g->sizey, hashIndex, 0.0f )
    );

    /* Translate the random sampled point (randPt) to the center of the cell
     * and make sure it stays within the cell boundaries. */
    return add_f3f3( cell, add_f3f( mul_f3f( randPt, 0.5f ), 0.5f ) );
}

/*
 * Voronoi noise 3x3x3 block. W is a random variation to the lookup. Changing
 * w with time will animate the Voronoi noise.
 */
float
voronoi3( __global global_buffers * g, float3 p, float w )
{
    float3 cell, iter, delta;
    float mindistSq, testDist;

    cell = floor_f3( p );
    mindistSq = 100.0f; /* Set mindistSq to a high number */

    for ( iter.x = cell.x - 1; iter.x < cell.x + 2; iter.x += 1 ) {
        for ( iter.y = cell.y - 1; iter.y < cell.y + 2; iter.y += 1 ) {
            for ( iter.z = cell.z - 1; iter.z < cell.z + 2; iter.z += 1 ) {
                /* Get the squared distance to the random sample point. */
                delta = sub_f3f3( getCellPoint( g, iter, w ), p );
                testDist = dot_f3( delta, delta );
                mindistSq = fmin( mindistSq, testDist );
            }
        }
    }

    return mindistSq;
}

/*
 * Voronoi noise 3x3x3 block second distance. W is a random variation to the
 * lookup. Changing w with time will animate the voronoi noise.
 */
float
voronoi3_2( __global global_buffers * g, float3 p, float w )
{
    float3 cell, iter, delta;
    float mindistSq, mindistSq2, testDist;

    cell = floor_f3( p );
    mindistSq  = 100.0f; /* Set mindistSq to a high number */
    mindistSq2 = 100.0f; /* Set mindistSq to a high number */
    for ( iter.x = cell.x - 1; iter.x < cell.x + 2; iter.x += 1 ){
        for ( iter.y = cell.y - 1; iter.y < cell.y + 2; iter.y += 1 ) {
            for ( iter.z = cell.z - 1; iter.z < cell.z + 2; iter.z += 1 ) {
                /* Get the squared distance to the random sample point */
                delta = sub_f3f3( getCellPoint( g, iter, w ), p );
                testDist = dot_f3( delta, delta );
                if ( testDist < mindistSq ) {
                    mindistSq2 = mindistSq;
                    mindistSq = testDist;
                } else {
                    mindistSq2 = fmin( testDist, mindistSq2 );
                }
            }
        }
    }

    return mindistSq2;
}

/*
 * Voronoi noise 2x2x2 block approximation to the 3x3x3 version. W is a random
 * variation to the lookup. Changing w with time will animate the Voronoi noise
 */
float
voronoi2( __global global_buffers * g, float3 p, float w )
{
    float3 cell, iter, delta;
    float mindistSq, testDist;
    int wShift;

    /* Bias to the position between the 8 neighboring cells */
    cell = floor_f3( add_f3f( p, 0.5f ) );
    mindistSq = 100.0f;    /* Set mindistSq to a high number */

    for ( iter.x = cell.x - 1; iter.x < cell.x + 1; iter.x += 1 ) {
        for ( iter.y = cell.y - 1; iter.y < cell.y + 1; iter.y += 1 ) {
            for ( iter.z = cell.z - 1; iter.z < cell.z + 1; iter.z += 1 ) {
                /* Sample 3 random points by shifting w */
                for ( wShift = 0; wShift < 3; ++wShift ) {
                    /* Sample 3 points and shift w by wShift * 4. Why 4? Why
                     * not. */
                    delta = sub_f3f3( getCellPoint( g, iter, w + wShift * 4 ), p );
                    testDist = dot_f3( delta, delta );
                    mindistSq = fmin( mindistSq, testDist );
                }
            }
        }
    }

    return mindistSq;
}

/*
 * Voronoi noise 2x2x2 block second distance approximation to the 3x3x3 version.
 * W is a random variation to the lookup. Changing w with time will animate the
 * Voronoi noise.
 */
float
voronoi2_2( __global global_buffers * g, float3 p, float w )
{
    float3 cell, iter, delta;
    float mindistSq, mindistSq2, testDist;
    int wShift;

    /* Bias to the position between the 8 neighboring cells */
    cell = floor_f3( add_f3f( p, 0.5f ) );
    mindistSq  = 100.0f; /* Set mindistSq to a high number */
    mindistSq2 = 100.0f; /* Set mindistSq2 to a high number */

    for ( iter.x = cell.x - 1; iter.x < cell.x + 1; iter.x += 1 ) {
        for ( iter.y = cell.y - 1; iter.y < cell.y + 1; iter.y += 1 ) {
            for ( iter.z = cell.z - 1; iter.z < cell.z + 1; iter.z += 1 ) {
                /* Sample 3 random points by shifting w */
                for ( wShift = 0; wShift < 3; ++wShift ) {
                    /* Sample 3 points and shift w by wShift * 4. Why 4? Why
                     * not. */
                    delta = sub_f3f3( getCellPoint( g, iter, w + wShift * 4 ), p );
                    testDist = dot_f3( delta, delta );
                    if ( testDist < mindistSq ) {
                        mindistSq2 = mindistSq;
                        mindistSq  = testDist;
                    } else {
                        mindistSq2 = fmin( testDist, mindistSq2 );
                    }
                }
            }
        }
    }

    return mindistSq2;
}

/*
 * Select different Voronoi noises based upon the input
 */
float
voronoiNoise( __global global_buffers * g, float3 p, float w, unsigned int noiseType )
{
    float output;

    output = 0;
    if ( noiseType == NT_VORONOI3 )
        output = voronoi3( g, p, w );
    else if ( noiseType == NT_VORONOI3_2 )
        output = voronoi3_2( g, p, w );
    else if ( noiseType == NT_VORONOI2 )
        output = voronoi2( g, p, w );
    else if ( noiseType == NT_VORONOI2_2 )
        output = voronoi2_2( g, p, w );

    return output;
}

/*
 * Sum octaves of noise at different frequencies. For simplex noise, this
 * produces the traditional "cloud" look.
 */
float
fBm(
    __global global_buffers * g,
    float4 vInputCoords,
    unsigned int nNumOctaves,
    float fLacunarity,
    float fGain,
    unsigned int noiseType
) {
    unsigned int i;

    float fNoiseSum = 0;
    float fAmplitude = 1;
    float fAmplitudeSum = 0;
    float4 vSampleCoords = vInputCoords;

    for ( i = 0; i < nNumOctaves; ++i ) {
        if ( noiseType == NT_SIMPLEX3D )
            fNoiseSum += fAmplitude * simplexNoise3d( g, swizzle_f4xyz( vSampleCoords ) );
        else if ( noiseType == NT_SIMPLEX4D )
            fNoiseSum += fAmplitude * simplexNoise4d( g, vSampleCoords );
        else if ( noiseType == NT_SIMPLEX3D_TURB )
            fNoiseSum += fAmplitude * fabs( simplexNoise3d( g, swizzle_f4xyz( vSampleCoords ) ) );
        else if ( noiseType == NT_SIMPLEX4D_TURB )
            fNoiseSum += fAmplitude * fabs( simplexNoise4d( g, vSampleCoords ) );
        else if ( noiseType >= NT_VORONOI3 )
            fNoiseSum += fAmplitude * voronoiNoise(
                g, swizzle_f4xyz( vSampleCoords ), vSampleCoords.w, noiseType
            );

        fAmplitudeSum += fAmplitude;
        fAmplitude *= fGain;
        vSampleCoords = mul_f4f( vSampleCoords, fLacunarity );
    }

    fNoiseSum /= fAmplitudeSum;
    return fNoiseSum;
}

#endif /* NOISE_H */

