#ifndef PROCEDURAL_HELPERS_H
#define PROCEDURAL_HELPERS_H

#include <stdio.h>
#include "noise.h"

const float g_fSliderVal[] = { 0.5f, 2.0f, 1.0f, 1.0f };
const float g_fDerivDiscreteAccu = 0.5;

const int transformed = 0;

void
axis_align_widths( float3 * dx, float3 * dy )
__attribute__((noinline))
{
    float3 sum, diff;
    float maxx, maxy;

    if ( transformed ) {
        sum  = add_f3f3( *dx, *dy );
        diff = sub_f3f3( *dx, *dy );

        /* box around parallelogram */
        maxx = fmax( fabs( sum.x ), fabs( diff.x ) );
        maxy = fmax( fabs( sum.y ), fabs( diff.y ) );
        *dx = makef3_fff( maxx, 0.0f, 0.0f );
        *dy = makef3_fff( 0.0f, maxy, 0.0f );

        /* smallest axis-aligned ellipse containing joined vectors */
        /*
        if ( length_f3( sum ) > length_f3( diff ) ) {
            *dx = makef3_fff( fabs( sum.x ) * sqrt( 2.0f ), 0.0f, 0.0f );
            *dy = makef3_fff( 0.0f, fabs( sum.y ) * sqrt( 2.0f ), 0.0f );
        } else {
            *dx = makef3_fff( fabs( diff.x ) * sqrt( 2.0f ), 0.0f, 0.0f );
            *dy = makef3_fff( 0.0f, fabs( diff.y ) * sqrt( 2.0f ), 0.0f );
        }
        */

        /* circle around larger of joined vectors */
        /*
        maxx = fmax( length_f3( sum ), length_f3( diff ) );
        *dx = makef3_fff( maxx, 0.0f, 0.0f );
        *dy = makef3_fff( 0.0f, maxx, 0.0f );
        */
#ifdef FILTER_MODE_GAUSS
        *dx = mul_f3f( *dx, GAUSS_SIGMA_FACTOR );
        *dy = mul_f3f( *dy, GAUSS_SIGMA_FACTOR );
#endif
    }
}

/*
 * Creates a grid pattern in 3d. This is a relatively cheap grid technique.
 * LineWidth determines the width of the grid lines. Hardness controls the hard
 * look or "fuzziness" of the lines.
 */
float
grid( float3 worldPos, float3 lineWidth, float hardness )
{
    float3 nearPlanes1, nearPlanes2;
    float3 delta1, delta2;
    float3 intermediate1, intermediate2;
    float retval;

    /* Assume the world is full of axis aligned planes at integer values. Find
     * the 6 planes nearest the world point. */
    nearPlanes1 = ceil_f3( worldPos );
    nearPlanes2 = floor_f3( worldPos );

    /* Find the distance to each of these planes */
    delta1 = abs_f3( sub_f3f3( worldPos, nearPlanes1 ) );
    delta2 = abs_f3( sub_f3f3( worldPos, nearPlanes2 ) );

    /* Clamp to within the line width of the planes */
    intermediate1 = saturate_f3( sub_f3f3( lineWidth, delta1 ) );
    intermediate2 = saturate_f3( sub_f3f3( lineWidth, delta2 ) );

    /* This is basically squaring the distances to the planes toget and
     * summing them all up. */
    retval = 0.0f;
    retval += dot_f3( intermediate1, intermediate1 );
    retval += dot_f3( intermediate2, intermediate2 );

    /* Multiplying by harness will change how much of the value actually goes
     * above 1. This effectively changes the hardness look of the lines. */
    return saturate_f( retval * hardness );
}

/*
 * Creates a grid pattern in 3D where some of the blocks are staggered based
 * on the masks. LineWidth determines the width of the grid lines. Harness
 * controls the hard look or "fuzziness" of the lines.
 */
float
staggeredGrid(
    float3 worldPos,
    float3 lineWidth,
    float hardness,
    float3 staggerOnMask,
    float3 staggerByMask
) {
    float3 test;
    int3 intTest, mod2s;
    float maxMod;

    /* Shift some of the coordinates by applying the stagger masks.
     * StaggerOnMask squishes the world position based upon the mask.
     * StaggerByMask moves the position in increments of whole units. */
    test = abs_f3( floor_f3( mul_f3f3( worldPos, staggerOnMask ) ) );
    intTest = int3_f3( test );
    mod2s = and_i3i( intTest, 1 );
    maxMod = mod2s.x + mod2s.y + mod2s.z;
    worldPos = add_f3f3( worldPos, mul_f3f( staggerByMask, 0.5f * maxMod ) );

    /* Just call grid on the shifted coordinates */
    return grid( worldPos, lineWidth, hardness );
}

/*
 * Colored grid. This is similar to grid, except we look up our color in the
 * random texture.
 */
float
gridColor( __global global_buffers * g, float3 worldPos )
{
    float3 nearLine;

    /* This assumes that we change colors every 1 world unit. The world variable
     * should be pre-scaled before calling this function. */
    nearLine = div_f3f( floor_f3( worldPos ), 255.0f );

    /* We scale nearLine.z by 10 just to give some variation in the color. */
    return samplef_ff(
        g->g_txRand_f, g->sizex, g->sizey,
        nearLine.x + 10.0f * nearLine.z, nearLine.y + 10.0f * nearLine.z
    );
}

/*
 * Staggerd grid color
 */
float
staggeredGridColor( __global global_buffers * g, float3 worldPos, float3 staggerOnMask, float3 staggerByMask )
{
    float3 test;
    int3 intTest, mod2s;
    float maxMod;

    /* Shift some of the coordinates by applying some fancy masks */
    test = abs_f3( floor_f3( mul_f3f3( worldPos, staggerOnMask ) ) );
    intTest = int3_f3( test );
    mod2s = and_i3i( intTest, 1 );
    maxMod = mod2s.x + mod2s.y + mod2s.z;
    worldPos = add_f3f3( worldPos, mul_f3f( staggerByMask, 0.5f * maxMod ) );

    /* Then just call gridColor on the staggered coordinates */
    return gridColor( g, worldPos );
}

#endif /* PROCEDURAL_HELPERS_H */

