#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fftw3.h>

#include "ppm_util.h"

void
print_usage( char * prog )
{
    printf( "Usage: %s [options] filename rate\n", prog );
    printf( "   --box       apply a box filter instead of accumulating\n" );
    printf( "   --dft file  write the DFT image to the named file\n" );
    printf( "   --gray      transform the image to grayscale first\n" );
    printf( "   -o file     write the aliased image to the named file\n" );
    printf( "   --p+s N     apply p+s decomposition N times\n" );
    printf( "   -h,--help   show this help and exit\n" );
}

void
check_argc( int argc, char * option )
{
    if ( argc > 0 )
        return;

    fprintf( stderr, "ERROR: missing %s argument\n", option );
    exit( 1 );
}

int
parse_args(
    int argc,
    char ** argv,
    unsigned int * rate,
    unsigned int * box,
    unsigned int * gray,
    unsigned int * p_s,
    const char ** dftname,
    const char ** outfile,
    const char ** ppmname
) {
    char *endptr, *prog;

    prog = argv[ 0 ];
    for ( ++argv, --argc; argc > 0; ++argv, --argc ) {
        if ( strcmp( "--help", *argv ) == 0 || strcmp( "-h", *argv ) == 0 ) {
            print_usage( prog );
            exit( 0 );
        } else if ( strcmp( "--box", *argv ) == 0 ) {
            *box = 1;
        } else if ( strcmp( "--dft", *argv ) == 0 ) {
            check_argc( --argc, "--dft" );
            ++argv;
            *dftname = *argv;
        } else if ( strcmp( "--gray", *argv ) == 0 ) {
            *gray = 1;
        } else if ( strcmp( "-o", *argv ) == 0 ) {
            check_argc( --argc, "-o" );
            ++argv;
            *outfile = *argv;
        } else if ( strcmp( "--p+s", *argv ) == 0 ) {
            check_argc( --argc, "--p+s" );
            ++argv;
            *p_s = atoi( *argv );
        } else if ( ppmname != NULL ) {
            *ppmname = *argv;
            ppmname = NULL;
        } else if ( rate != NULL ) {
            *rate = strtod( *argv, &endptr );
            if ( endptr == *argv ) {
                fprintf( stderr, "ERROR: could not parse '%s'\n", *argv );
                return 1;
            }
            rate = NULL;
        } else {
            fprintf( stderr, "ERROR: unrecognized argument '%s'\n", *argv );
        }
    }
    if ( rate != NULL ) {
        print_usage( prog );
        return 1;
    }

    return 0;
}

float
lum( float * c, unsigned int depth )
{
    if ( depth == 3 )
        return 0.2126f * c[0] + 0.7152f * c[1] + 0.0722f * c[2];
    else
        return c[0];
}

float
modulus( fftw_complex c )
{
    return sqrt( c[ 0 ] * c[ 0 ] + c[ 1 ] * c[ 1 ] );
}

float
re( fftw_complex c )
{
    return c[ 0 ];
}

float
im( fftw_complex c )
{
    return c[ 1 ];
}

fftw_complex *
get_dft( float * raster, unsigned int w, unsigned int h, unsigned int d )
{
    fftw_complex * buffer;
    fftw_plan p;
    unsigned int i;

    int n[] = { h, w };

    buffer = (fftw_complex*) fftw_malloc( sizeof( fftw_complex ) * w * h * d );
    p = fftw_plan_many_dft(
        2,      /* dimensionality of data to transform */
        n,      /* size in each dimension */
        d,      /* number of transforms to compute */
        buffer, /* input buffer */
        n,      /*     embedded in buffer of same size */
        d,      /*     stride between consecutive elements of each series */
        1,      /*     distance between first elements of consecutive series */
        buffer, /* output buffer */
        n,      /*     embedded in buffer of same size */
        d,      /*     stride between consecutive elements of each series */
        1,      /*     distance between first elements of consecutive series */
        FFTW_FORWARD,
        FFTW_ESTIMATE
    );

    memset( buffer, 0, sizeof( fftw_complex ) * w * h * d );
    for ( i = 0; i < (w * h * d); ++i )
        buffer[ i ][ 0 ] = raster[ i ];
    fftw_execute( p );

    fftw_destroy_plan( p );

    return buffer;
}

float *
invert_dft(
    fftw_complex * dft,
    unsigned int w,
    unsigned int h,
    unsigned int d,
    float (*f)(fftw_complex)
) {
    fftw_complex * buffer;
    fftw_plan p;
    float * raster;
    float scale;
    int i;
    unsigned int n[] = { h, w };

    buffer = (fftw_complex*) fftw_malloc( sizeof( fftw_complex ) * w * h * d );
    p = fftw_plan_many_dft(
        2,      /* dimensionality of data to transform */
        n,      /* size in each dimension */
        d,      /* number of transforms to compute */
        buffer, /* input buffer */
        n,      /*     embedded in buffer of same size */
        d,      /*     stride between consecutive elements of each series */
        1,      /*     distance between first elements of consecutive series */
        buffer, /* output buffer */
        n,      /*     embedded in buffer of same size */
        d,      /*     stride between consecutive elements of each series */
        1,      /*     distance between first elements of consecutive series */
        FFTW_BACKWARD,
        FFTW_ESTIMATE
    );

    memcpy( buffer, dft, sizeof( fftw_complex ) * w * h * d );
    fftw_execute( p );

    scale = 1.0f / w / h;
    raster = (float*) malloc( sizeof( float ) * w * h * d );
    for ( i = 0; i < (w * h * d); ++i )
        raster[ i ] = f( buffer[ i ] ) * scale;

    fftw_destroy_plan( p );
    fftw_free( buffer );

    return raster;
}

float
energy_above_rate(
    fftw_complex * dft,
    unsigned int w,
    unsigned int h,
    unsigned int d,
    unsigned int rate,
    unsigned int box
) {
    fftw_complex * tmp;
    unsigned int nyq_x, nyq_y, nyq_x2, nyq_y2, count;
    unsigned int i, j, x, y, c;
    float accum, total, val, scale;

    nyq_x = ( w / rate );
    nyq_y = ( h / rate );
    nyq_x2 = nyq_x / 2;
    nyq_y2 = nyq_y / 2;

    tmp = (fftw_complex*) fftw_malloc( sizeof( fftw_complex ) * nyq_x * nyq_y * d );
    memset( tmp, 0, sizeof( fftw_complex ) * nyq_x * nyq_y * d );

    total = 0.0f;
    accum = 0.0f;
    count = 0;
    scale = 1.0f / ( rate * rate );

    for ( y = 0, i = 0; y < h; ++y ) {
        for ( x = 0; x < w; ++x ) {
            for ( c = 0; c < d; ++c, ++i ) {
                val = modulus( dft[ i ] );
                total += val;
                if ( ( nyq_y2 <= y && y < h - nyq_y2 )
                    || ( nyq_x2 <= x && x < w - nyq_x2 ) ) {
                    ++count;
                    accum += val;
                    if ( ! box ) {
                        j = ( (y % nyq_y) * nyq_x + (x % nyq_x) ) * d + c;
                        tmp[ j ][ 0 ] += dft[ i ][ 0 ] * scale;
                        tmp[ j ][ 1 ] += dft[ i ][ 1 ] * scale;
                    }
                } else {
                    j = ( (y % nyq_y) * nyq_x + (x % nyq_x) ) * d + c;
                    tmp[ j ][ 0 ] += dft[ i ][ 0 ] * scale;
                    tmp[ j ][ 1 ] += dft[ i ][ 1 ] * scale;
                }
            }
        }
    }
    memset( dft, 0, sizeof( fftw_complex ) * w * h * d );
    memcpy( dft, tmp, sizeof( fftw_complex ) * nyq_x * nyq_y * d );

    fftw_free( tmp );

    if ( accum > 0 )
        accum /= count;
    total /= w * h;

    printf( "Total energy: %g\n", total );
    printf( "Ratio: %g\n", accum / total );

    return accum;
}

float *
get_dft_raster(
    fftw_complex * dft,
    unsigned int w,
    unsigned int h,
    unsigned int d
) {
    float * raster;
    float scale, val;
    unsigned int x, y, i, j, c;

    scale = 0.0f;
    raster = (float*) malloc( sizeof( float ) * w * h * d );
    for ( i = 0; i < h; ++i ) {
        y = ( i + h/2 ) % h;
        for ( j = 0; j < w; ++j ) {
            x = ( j + w/2 ) % w;
            for ( c = 0; c < d; ++ c ) {
                val = log( modulus( dft[ ( i * w + j ) * d + c ] ) + 1 );
                scale = fmax( scale, val );
                raster[ ( y * w + x ) * d + c ] = val;
            }
        }
    }
    if ( scale > 0.0f ) {
        scale = 1.0f / scale;
        for ( i = 0; i < (w * h * d); ++i )
            raster[ i ] *= scale;
    }
    return raster;
}

/*
 * This program is freely available on the web page
 *
 * http://www.mi.parisdescartes.fr/~moisan/p+s
 *
 * I hope that you will find it useful.
 * If you use it for a publication, please mention 
 * this web page and the paper it makes reference to.
 * If you modify this program, please indicate in the
 * code that you did so and leave this message.
 * You can also report bugs or suggestions to 
 * lionel.moisan [AT] parisdescartes.fr
 *
 * 16 Jul 2015: Adapted by Jonathan Dorn
 */
void
p_plus_s_decomposition( float * u, unsigned w, unsigned h, unsigned d )
{
    float * v;
    double a, cx, cy;
    float b;
    unsigned x, y, c;
    
#define idx( x, y, c ) (((y)*w + (x))*d + (c))
    // 1. Compute the image v to capture intensity gaps of u across the border

    v = (float*) malloc( sizeof( float ) * w * h * d );
    memset( v, 0, sizeof( float ) * w * h * d );
    for ( x = 0; x < w; ++x ) {
        for ( c = 0; c < d; ++c ) {
            b = u[ idx( x, 0, c ) ] - u[ idx( x, h-1, c ) ];
            v[ idx( x,   0, c ) ] =  b;
            v[ idx( x, h-1, c ) ] = -b;
        }
    }
    for ( y = 0; y < h; ++y ) {
        for ( c = 0; c < d; ++c ) {
            b = u[ idx( 0, y, c ) ] - u[ idx( w-1, y, c ) ];
            v[ idx(   0, y, c ) ] += b;
            v[ idx( w-1, y, c ) ] -= b;
        }
    }

    // 2. compute the DFT of v

    fftw_complex * dft = get_dft( v, w, h, d );

    // 3. compute
    //                               v-hat(q, r)
    // s-hat(q, r) = ---------------------------------------------
    //               2 cos( 2 pi q / M ) + 2 cos( 2 pi r / N ) - 4
    //
    // s-hat(0, 0) = 0

    cx = 2.0 * M_PI / w;
    cy = 2.0 * M_PI / h;
    for ( y = 0; y < h; ++y ) {
        a = cos( cy * y );
        for ( x = 0; x < w; ++x ) {
            b = 0.5 / ( 2.0 - a - cos( cx * x ) );
            for ( c = 0; c < d; ++c ) {
                dft[ idx( x, y, c ) ][ 0 ] *= b;
                dft[ idx( x, y, c ) ][ 1 ] *= b;
            }
        }
    }
    for ( c = 0; c < d; ++ c ) {
        dft[ idx( 0, 0, c ) ][ 0 ] = 0.0f;
        dft[ idx( 0, 0, c ) ][ 1 ] = 0.0f;
    }

    // 4. compute u - per(u) = idft(s-hat)
    //     per(u) = u - idft(s-hat)

    free( v );
    v = invert_dft( dft, w, h, d, re );

    for ( c = 0; c < w * h * d; ++c )
        u[ c ] = u[ c ] - v[ c ];

    free( v );
    fftw_free( dft );
#undef idx
}

int
main( int argc, char ** argv )
{
    unsigned int i, j;
    const char *fname, *dftname, *outfile;
    float *raster, *tmp;
    fftw_complex *dft;
    float maxval;
    unsigned int rate, box, gray, p_s;
    unsigned int w, h, depth;
    float energy;

    dftname = NULL;
    outfile = NULL;
    rate = 1;
    box = 0;
    gray = 0;
    p_s = 0;

    i = parse_args( argc, argv,
        &rate, &box, &gray, &p_s, &dftname, &outfile, &fname
    );
    if ( i != 0 )
        return i;

    raster = load_ppm( fname, &w, &h, &depth );
    if ( gray && depth == 3 ) {
        tmp = (float*) malloc( sizeof( float ) * w * h );
        for ( i = 0; i < w * h; ++i )
            tmp[ i ] = lum( raster + i * depth, depth );
        free( raster );
        raster = tmp;
        depth = 1;
    }
    for ( i = 0; i < p_s; ++i ) 
        p_plus_s_decomposition( raster, w, h, depth );
    dft = get_dft( raster, w, h, depth );
    free( raster );

    energy = energy_above_rate( dft, w, h, depth, rate, box );
    w /= rate;
    h /= rate;

    if ( outfile != NULL ) {
        raster = invert_dft( dft, w, h, depth, re );
        save_ppm( outfile, raster, w, h, depth );
        free( raster );
    }
    if ( dftname != NULL ) {
        raster = get_dft_raster( dft, w, h, depth );
        save_ppm( dftname, raster, w, h, depth );
        free( raster );
    }
    fftw_free( dft );

    printf( "%g\n", energy );
    return 0;
}

