#!/usr/bin/python2

from collections import defaultdict, OrderedDict, namedtuple
from contextlib import contextmanager
from glob import glob
from optparse import OptionParser
import os
import re
import shutil
from subprocess import call, check_call, PIPE, Popen, STDOUT, CalledProcessError
import sys
import tempfile

parser = OptionParser( usage = "%prog [options] start stop frames" )
parser.add_option(
    "--shader", help = "shader to animate"
)
parser.add_option(
    "--variant", metavar = "name,file", action = "append", default = list(),
    help = "source code for shader variant to render"
)
parser.add_option(
    "-d", "--define", action = "append", default = list(),
    help = "macro definition to add to the compile command line"
)
parser.add_option(
    "--diff", metavar = "alg", choices = [ "l2", "ssim" ], default = "l2",
    help = "differencing algorithm to use"
)
parser.add_option(
    "-f", "--func", help = "name of shader function"
)
parser.add_option(
    "-o", "--output", default = "anim.mp4", help = "movie file to generate"
)
parser.add_option(
    "--fps", action = "store_true", help = "report FPS for each render mode"
)
parser.add_option(
    "--ground-truth", metavar = "N",
    help = "include super-sampled (with N samples) ground-truth image"
)
parser.add_option(
    "--inset", metavar = "alg", choices = [ "ground-truth" ],
    help = "image to inset in lower right corner of frames"
)
parser.add_option(
    "--msaa", metavar = "N",
    help = "include regular multisample antialiasing with NxN samples"
)
parser.add_option(
    "--perspective", action = "store_true",
    help = "render a perspective projection rather than orthographic"
)
parser.add_option(
    "--sampling", metavar = "alg", choices = [ "gauss", "hammersley" ],
    help = "select the MSAA sampling algorithm to use"
)
parser.add_option(
    "--size", metavar = "NxN", default = "150x150",
    help = "size of panels to render"
)
parser.add_option(
    "--skip", metavar = "mode", default = list(), action = "append",
    help = "do not render the selected filter mdoe; may be used many times"
)
parser.add_option(
    "--transform", action = "store_true",
    help = "automatically transform the shader to be band-limited"
)
options, args = parser.parse_args()

if len( args ) != 3:
    parser.print_help()
    exit()
if options.shader is None:
    print >>sys.stderr, "ERROR: --shader is a required flag\n"
    parser.print_help()
    exit()
if not hasattr( options, "inset" ):
    options.inset = None
if options.inset == "ground-truth":
    if options.ground_truth is None:
        print "ERROR: cannot inset ground truth: --ground-truth not defined"
        exit( 1 )
if options.func is None:
    options.func = os.path.splitext( os.path.basename( options.shader ) )[ 0 ]
if len( options.define ) == 0:
    options.define = [ None ]
options.variant = map( lambda ns: ns.split( ",", 1 ), options.variant )

start = float( args[ 0 ] )
stop  = float( args[ 1 ] )
n     = int( args[ 2 ] )
size  = ( stop - start ) / n

frame_height = int( options.size.split( "x" )[ 1 ] )

root = os.path.dirname( os.path.dirname( os.path.abspath( sys.argv[ 0 ] ) ) )
source = os.path.join( root, options.shader )
compile_sh = os.path.join( root, "tools", "compile.sh" )
renderer   = os.path.join( root, "tools", "renderer" )
transform  = os.path.join( root, "tools", "transform" )

for fname in [ source ] + map( lambda ns: ns[ 1 ], options.variant ):
    if not os.path.exists( fname ):
        print >>sys.stderr, "%s: no such file or directory" % fname
        exit( 1 )

Frame = namedtuple( "Frame", "fname inset fps error" )

@contextmanager
def mktemp( suffix = '' ):
    tmp = tempfile.NamedTemporaryFile( suffix = suffix, delete = False )
    tmp.close()
    try:
        yield tmp.name
    finally:
        if os.path.exists( tmp.name ):
            os.remove( tmp.name )

@contextmanager
def mkdir():
    tmp = tempfile.mkdtemp()
    try:
        yield tmp
    finally:
        if os.path.exists( tmp ):
            check_call( [ "rm", "-rf", tmp ] )

def write( msg ):
    sys.stdout.flush()
    print msg
    sys.stdout.flush()
    try:
        os.fsync( sys.stdout.fileno() )
    except OSError:
        pass

grp_size = 30
frames = map( lambda i: start + size * i, range( n ) )
fmt = "%%0%dd" % len( str( len( frames ) ) )
frames = map( lambda (i, z): (z, fmt % i), enumerate( frames ) )
frames = [ frames[ i:i + grp_size ] for i in range( 0, len(frames), grp_size ) ]

if os.path.exists( options.output ):
    os.remove( options.output )

modes = OrderedDict()
with open( options.shader ) as fh:
    pat = re.compile( "FILTER_MODE_([A-Za-z_]+)" )
    for line in fh:
        m = pat.search( line )
        if not m:
            continue
        mode = m.group( 1 ).lower()
        if not mode in modes:
            modes[ mode ] = source
for name, variant in options.variant:
    modes[ name ] = variant
if options.transform:
    modes[ "transform" ] = source
if options.msaa is not None:
    modes[ "msaa_%s" % options.msaa ] = source
if options.ground_truth is not None:
    modes[ "target" ] = source
for mode in modes.keys():
    if mode in options.skip:
        del modes[ mode ]

class Shader:
    def __init__( self, source, mode ):
        self.src     = source
        self.mode    = mode
        self.handles = list()

    def __enter__( self ):
        return self

    def build( self, *defines ):
        if os.path.splitext( self.src )[ 1 ] == ".cl":
            handle = mktemp( ".cl" )
            clfile = handle.__enter__()
            self.handles.append( ( handle, clfile ) )
            shutil.copyfile( self.src, clfile )
        else:
            do_transform = False
            mode = self.mode
            if self.mode == "target" or self.mode.startswith( "msaa_" ):
                mode = "none"
            if self.mode == "transform":
                do_transform = True
            mode = "tent"

            cmd = [ "gcc", "-E", self.src, "-DFILTER_MODE_" + mode.upper() ]
            for define in defines:
                cmd.append( "-D" + define )

            with mktemp( ".c" ) as srcpp:
                cmd += [ "-o", srcpp ]
                check_call( cmd )
                if do_transform:
                    with mktemp( ".c" ) as transformed:
                        check_call( [
                            transform, srcpp, options.func,
                                "--transform", "all",
                                "--change-indicator", "transformed",
                                "-o", transformed
                        ] )
                        os.rename( transformed, srcpp )
                handle = mktemp( ".cl" )
                clfile = handle.__enter__()
                self.handles.append( ( handle, clfile ) )

                check_call( [ compile_sh, options.func, clfile, srcpp ] )

    def __render_command( self, tmpdir, row, frames, golden ):
        tempfiles = list()
        cmd = [ renderer, self.handles[ -1 ][ 1 ] ]
        for i, ( ( zoom, num ), ref ) in enumerate( zip( frames, golden ) ):
            outfile = os.path.join( tmpdir,
                "frame-%d-%s-%s" % ( row, num, self.mode )
            )
            if ref is not None:
                outfile += "-diff"
            outfile += ".ppm"
            tempfiles.append( outfile )

            cmd += [ "-i", "--zoom", str( zoom ), "-o", outfile ]
            if ref is not None:
                cmd += [ "-g", ref ]
            if i == 0:
                cmd += [ "--size", options.size ]
                if mode == "target":
                    cmd += [ "--msaa", options.ground_truth ]
                if mode.startswith( "msaa_" ):
                    cmd += [ "--msaa", options.msaa, "--sampling", "hammersley" ]
                if options.fps:
                    cmd += [ "--fps" ]
                if options.perspective:
                    cmd += [ "--perspective" ]
        return cmd, tempfiles

    def render( self, tmpdir, row, frames, golden = None ):
        cmd, fnames = self.__render_command(
            tmpdir, row, frames, [ None ] * len( frames )
        )
        p = Popen( cmd, stdout = PIPE, stderr = STDOUT )
        lines = p.communicate()[ 0 ]
        if p.returncode != 0:
            print >>sys.stderr, lines
            raise CalledProcessError( returncode, cmd, None )

        result = list()
        if options.fps:
            for i, ( image, line ) in enumerate( zip( fnames, lines.splitlines() ) ):
                try:
                    fps = float( line.split()[ -1 ] )
                    result.append( Frame( image, None, fps, None ) )
                except:
                    result.append( Frame( image, None, None, None ) )
        else:
            for image in fnames:
                result.append( Frame( image, None, None, None ) )

        if golden is None:
            return result

        cmd, fnames = self.__render_command( tmpdir, row, frames, golden )
        p = Popen(
            cmd + [ "--diff", options.diff ],
            stdout = PIPE, stderr = STDOUT
        )
        lines = p.communicate()[ 0 ]
        if p.returncode != 0:
            print >>sys.stderr, lines
            raise CalledProcessError( p.returncode, cmd, None )

        for i, ( image, line ) in enumerate( zip( fnames, lines.splitlines() ) ):
            f = result[ i ]
            try:
                error = float( line.split()[ 0 ] )
                result[ i ] = Frame( f.fname, image, f.fps, error )
            except:
                result[ i ] = Frame( f.fname, image, f.fps, None )

        return result

    def __exit__( self, typ, val, trace ):
        for handle, _ in reversed( self.handles ):
            if handle.__exit__( typ, val, trace ):
                return True

frame_groups = frames
with mkdir() as framedir:
    # Generate individual frames for each mode in each row. Generate rows in
    # reverse order in case we need to generate diff insets against the target
    # images. Since target images are last, reverse order means the target will
    # exist by the time we need it.

    results = defaultdict( list )
    for mode in reversed( modes ):
        do_diff = options.inset is not None and mode != "target"
        golden = None
        for row, define in enumerate( options.define ):
            key = ( mode, row )
            if define is None:
                write( "filtering mode: %s" % mode )
                define = list()
            else:
                write( "filtering mode: %s   define: %s" % ( mode, define ) )
                define = [ define ]
            with Shader( modes[ mode ], mode ) as shader:
                shader.build( *define )
                for i, group in enumerate( frame_groups ):
                    if do_diff:
                        golden = map(
                            lambda f: f.fname,
                            results[ ( "target", row ) ][ i ]
                        )
                    frames = shader.render( framedir, row, group, golden )
                    results[ key ].append( frames )

    write( "applying insets" )
    fps = defaultdict( list )
    err = defaultdict( list )
    for key in results:
        for fs in results[ key ]:
            for f in fs:
                if f.fps is not None:
                    fps[ key ].append( f.fps )
                if f.error is not None:
                    err[ key ].append( f.error )
                if f.inset is not None:
                    with mktemp( ".ppm" ) as tmpimage:
                        check_call( [ "convert",
                            "-set", "colorspace", "RGB",
                            f.inset, "-resize", "25%",
                            f.fname, "-reverse",
                            "-flip", "-flop", "-mosaic", "-flop", "-flip",
                            tmpimage
                        ] )
                        os.rename( tmpimage, f.fname )

    write( "overlaying text" )
    frame_files = list()
    for key in results:
        mode = key[ 0 ]
        cmd = [ "mogrify", "-format", "png",
            "-set", "colorspace", "RGB",
            "-bordercolor", "black",
            "-border", "5",
            "-pointsize", "20",
            "-font", "Helvetica-Bold",
            "-fill", "white",
            "-stroke", "black",
            "-draw", "text 7,%d '%s'" % ( frame_height-2, mode )
        ]
        info = list()
        if key in fps:
            avg = sum( fps[ key ] ) / len( fps[ key ] )
            if avg > 99:
                avg = str( int( avg ) )
            else:
                avg = "%.3g" % avg
            info.append( "%s fps" % avg )
        if key in err:
            error = sum( err[ key ] ) / len( err[ key ] )
            info.append( "err: %4.3f" % error )
        if len( info ) > 0:
            info = " | ".join( info )
            cmd += [
                "-pointsize", "14",
                "-draw", "text 7,%d '%s'" % ( frame_height - 22, info )
            ]
        for fs in results[ key ]:
            for f in fs:
                cmd.append( f.fname )
        check_call( cmd )

    write( "concatenating frames" )
    frame_files = list()
    for i, group in enumerate( frame_groups ):
        for j, (_, n) in enumerate( group ):
            cmd = [ "montage",
                "-tile", "%dx%d" % ( len( modes ), len( options.define ) ),
                "-mode", "Concatenate",
            ]
            for row, define in enumerate( options.define ):
                for mode in modes:
                    key = ( mode, row )
                    cmd.append(
                        os.path.splitext( results[ key ][ i ][ j ].fname )[ 0 ] + ".png"
                    )
            cmd.append( os.path.join( framedir, "frame-%s.png" % n ) )
            check_call( cmd )
            frame_files.append( "frame-%s.png" % n )

    if options.output.endswith( ".tar.gz" ) \
            or options.output.endswith( ".tgz" ) \
            or options.output.endswith( ".tar.bz2" ):
        if options.output.endswith( ".tar.bz2" ):
            flags = "-cjf"
        else:
            flags = "-czf"
        output = os.path.abspath( options.output )
        os.chdir( framedir )
        with mktemp() as filelist:
            with open( filelist, 'w' ) as fh:
                for frame_file in frame_files:
                    print >>fh, frame_file
            check_call( [ "tar", flags, output, "-T", filelist ] )
    else:
        check_call( [ "avconv",
            "-i", os.path.join( framedir, "frame-" + fmt + ".png" ),
            "-c:v", "libx264",
            "-preset", "slow",
            "-crf", "0",
            options.output
        ] )

