#!/usr/bin/python2

from collections import defaultdict, OrderedDict, namedtuple
from contextlib import contextmanager
from glob import glob
from optparse import OptionParser
import os
import re
from subprocess import call, check_call, PIPE, Popen, STDOUT, CalledProcessError
import sys
import tempfile

parser = OptionParser(
    usage = "%prog [options] start stop frames name source"
)
parser.add_option(
    "-f", "--func", metavar = "func", help = "name of shader function"
)
parser.add_option(
    "-o", "--output", metavar = "dir", default = "anim",
    help = "directory to contain generated frames"
)
parser.add_option(
    "--coeff", metavar = "c1,c2",
    help = "comma-separated list of coefficients for the shader"
)
parser.add_option(
    "--fps", metavar = "side", choices = [ "left", "right" ],
    help = "side to write FPS on"
)
parser.add_option(
    "--samples", metavar = "N", type = int, default = 1,
    help = "use N samples per pixel"
)
parser.add_option(
    "--size", metavar = "NxN", default = "256x256",
    help = "size of panels to render"
)
options, args = parser.parse_args()

if len( args ) != 5:
    parser.print_help()
    exit()

start = float( args[ 0 ] )
stop  = float( args[ 1 ] )
n     = int( args[ 2 ] )
size  = ( stop - start ) / n

label  = args[ 3 ]
source = args[ 4 ]

if not os.path.exists( source ):
    print >>sys.stderr, "%s: no such file or directory" % source
    exit( 1 )
if options.func is None:
    options.func = os.path.splitext( os.path.basename( source ) )[ 0 ]

frame_height = int( options.size.split( "x" )[ 1 ] )

root = os.path.dirname( os.path.dirname( os.path.abspath( sys.argv[ 0 ] ) ) )
source = os.path.join( root, source )
compile_sh = os.path.join( root, "tools", "compile.sh" )
renderer   = os.path.join( root, "tools", "renderer" )

Frame = namedtuple( "Frame", "fname fps" )

@contextmanager
def mktemp( suffix = '' ):
    tmp = tempfile.NamedTemporaryFile( suffix = suffix, delete = False )
    tmp.close()
    try:
        yield tmp.name
    finally:
        if os.path.exists( tmp.name ):
            os.remove( tmp.name )

@contextmanager
def mkdir():
    tmp = tempfile.mkdtemp()
    try:
        yield tmp
    finally:
        if os.path.exists( tmp ):
            check_call( [ "rm", "-rf", tmp ] )

def write( msg ):
    sys.stdout.flush()
    print msg
    sys.stdout.flush()
    try:
        os.fsync( sys.stdout.fileno() )
    except OSError:
        pass

grp_size = 30
frames = map( lambda i: start + size * i, range( n ) )
fmt = "%%0%dd" % len( str( len( frames ) ) )
frames = map( lambda (i, z): (z, fmt % i), enumerate( frames ) )
frames = [ frames[ i:i + grp_size ] for i in range( 0, len(frames), grp_size ) ]

if os.path.exists( options.output ):
    check_call( [ "rm", "-rf", options.output ] )
os.makedirs( options.output )

class Shader:
    def __init__( self, source, mode ):
        self.src     = source
        self.mode    = mode
        self.handles = list()
        self.code    = None

    def __enter__( self ):
        return self

    def build( self, *defines ):
        mode = self.mode

        if os.path.splitext( self.src )[ 1 ] == ".c":
            cmd = [ "gcc", "-E", self.src, "-DFILTER_MODE_NONE" ]

            with mktemp( ".c" ) as srcpp:
                cmd += [ "-o", srcpp ]
                check_call( cmd )
                handle = mktemp( ".cl" )
                clfile = handle.__enter__()
                self.handles.append( handle )

                check_call( [ compile_sh, options.func, clfile, srcpp ] )

                self.code = clfile
        else:
            self.code = self.src

    def __render_command( self, frames ):
        tempfiles = list()
        cmd = [
            renderer, self.code,
                "--size", options.size,
                "--msaa", str( options.samples ),
                "--fps",
                "--perspective"
        ]
        if options.coeff is not None:
            cmd += [ "--coeff", options.coeff ]
        for i, ( zoom, num ) in enumerate( frames ):
            outfile = os.path.join(
                options.output, "frame-%s-%s" % ( num, self.mode )
            )
            outfile += ".ppm"
            tempfiles.append( outfile )

            if i > 0:
                cmd.append( "-i" )
            cmd += [ "--zoom", str( zoom ), "-o", outfile ]
        return cmd, tempfiles

    def render( self, frames ):
        cmd, fnames = self.__render_command( frames )
        p = Popen( cmd, stdout = PIPE, stderr = STDOUT )
        lines = p.communicate()[ 0 ]
        if p.returncode != 0:
            print >>sys.stderr, lines
            raise CalledProcessError( returncode, cmd, None )

        result = list()
        for i, ( image, line ) in enumerate( zip( fnames, lines.splitlines() ) ):
            try:
                fps = float( line.split()[ -1 ] )
                result.append( Frame( image, fps ) )
            except:
                result.append( Frame( image, None ) )

        return result

    def __exit__( self, typ, val, trace ):
        for handle in reversed( self.handles ):
            if handle.__exit__( typ, val, trace ):
                return True

frame_groups = frames
results = list()
with Shader( source, label ) as shader:
    shader.build()
    for i, group in enumerate( frame_groups ):
        frames = shader.render( group )
        results.append( frames )

fps = list()
for fs in results:
    for f in fs:
        if f.fps is not None:
            fps.append( f.fps )

write( "overlaying text" )
frame_files = list()
cmd = [ "mogrify", "-format", "png",
    "-set", "colorspace", "RGB",
    "-bordercolor", "black",
    "-border", "5",
    "-font", "Helvetica-Bold",
    "-fill", "white",
    "-stroke", "black",
    "-pointsize", "20",
]
if options.fps == "left":
    cmd += [ "-gravity", "West" ]
else:
    cmd += [ "-gravity", "East" ]
cmd += [ "-draw", "text 7,%d '%s'" % ( frame_height/2 - 7, label ) ]
#if len( fps ) > 0:
#    avg = sum( fps ) / len( fps )
#    if avg > 99:
#        avg = str( int( avg ) )
#    else:
#        avg = "%.3g" % avg
#    cmd += [
#        "-pointsize", "14",
#        "-draw", "text 7,%d '%s'" % ( frame_height/2 - 27, avg )
#    ]
for fs in results:
    for f in fs:
        cmd.append( f.fname )
check_call( cmd )

cmd = [ "rm" ]
for fs in results:
    for f in fs:
        cmd.append( f.fname )
check_call( cmd )

