#!/usr/bin/python2

from collections import defaultdict
from contextlib import contextmanager
from glob import glob
from optparse import OptionParser
import os
import re
from subprocess import call, check_call, Popen, PIPE, STDOUT
import sys
import tempfile

from utils import *

parser = OptionParser(
    usage = "%prog [options] shader size fitness-file outfile"
)
parser.add_option(
    "--diff", metavar = "func",
    help = "include image difference from golden using the named algorithm"
)
parser.add_option(
    "--dft", action = "store_true", help = "include DFT of image"
)
parser.add_option(
    "--gray", action = "store_true",
    help = "display grayscale images instead of RGB images"
)
parser.add_option(
    "--ortho", metavar = "x", default = "1",
    help = "comma-separated list of orthographic scales to render"
)
parser.add_option(
    "--persp", metavar = "x", default = "0",
    help = "comma-separated list of perspective rotations to render"
)
parser.add_option(
    "--timeout", metavar = "secs", type = int, default = 10,
    help = "max number of seconds to allow tested applications to run"
)
options, args = parser.parse_args()

if len( args ) < 4:
    parser.print_help()
    exit()
if options.diff is not None:
    options.diff = options.diff.strip( "-" ).lower()
options.ortho = map( int, options.ortho.split( "," ) )
options.persp = map( int, options.persp.split( "," ) )

def mykey( fname ):
    terms = list()
    for term in re.split( "(\d+)", fname ):
        try:
            terms.append( int( term ) )
        except:
            terms.append( term )
    return terms

def write( string, ctx_list ):
    tmpfile = ctx_list.append( mktemp( ".ppm" ) )
    check_call( [
        "convert",
            "-size", "%dx20" % w,
            "-draw", "scale 1.5,1.5 gravity Center text 1x1 '%s'" % string,
            "canvas:white",
            tmpfile
    ] )
    return tmpfile

root = os.path.abspath( os.path.dirname( sys.argv[ 0 ] ) )

compile_sh = os.path.join( root, "compile.sh" )
renderer   = os.path.join( root, "renderer" )

shader, size, fitnessfile, outfile = args[ :4 ]
w, h = map( int, size.split( "x", 1 ) )

check_executable( [ compile_sh ] )
check_executable( [ renderer, "-h" ] )

if os.path.isdir( fitnessfile ):
    fitnessfiles = glob( os.path.join( fitnessfile, "*.fitness" ) )
    fitnessfiles.sort()

    fitnesses = list()
    sources = list()
    for fitnessfile in fitnessfiles:
        with open( fitnessfile ) as fh:
            fitnesses.append( next( fh ) )
        sources.append( os.path.splitext( fitnessfile )[ 0 ] + ".c" )
else:
    with open( fitnessfile ) as fh:
        fitnesses = fh.readlines()
    sources = glob( os.path.splitext( fitnessfile )[ 0 ] + "-*.c" )
    sources.sort( key = mykey )

modes = [
    ( False, [], options.ortho ),
    ( True,  [ "--perspective" ], options.persp )
]
with ContextList() as ctx:
    blank = write( "", ctx )
    if options.diff is not None:
        diffhead = write( options.diff.upper(), ctx )
    if options.dft:
        dfthead = write( "DFT", ctx )

    matrix = defaultdict( list )
    matrix[ "golden" ].append( write( "golden", ctx ) )
    for persp, flags, zooms in modes:
        if len( zooms ) == 0:
            continue
        for ss in zooms:
            golden = get_golden( shader, size, ss, options.gray, persp )
            if not os.path.exists( golden ):
                print >>sys.stderr, "cannot find golden image:", golden
                exit( 1 )
            matrix[ "golden" ].append( golden )
    matrix[ "golden" ].append( blank )

    clfile = ctx.append( mktemp( ".cl" ) )

    for source in sources:
        print >>sys.stderr, "rendering", source
        diffkey = ( source, "diff" )
        dftkey  = ( source, "dft" )
        check_call( [ compile_sh, shader, clfile, source ] )

        matrix[ source ].append( write( os.path.basename( source ), ctx ) )
        if options.diff is not None:
            matrix[ diffkey ].append( diffhead )
        if options.dft:
            matrix[ dftkey ].append( dfthead )

        for persp, flags, zooms in modes:
            if len( zooms ) == 0:
                continue
            cmd = [ renderer, clfile ] + flags
            if options.gray:
                cmd.append( "--gray" )
            for ss in zooms:
                tmpfile = ctx.append( mktemp( ".ppm" ) )
                cmd += [
                    "-i", "-o", tmpfile, "--size", size, "--zoom", str( ss )
                ]
                matrix[ source ].append( tmpfile )
            check_call( [ "timeout", str( options.timeout ) ] + cmd )

        if options.diff is not None:
            for persp, flags, zooms in modes:
                if len( zooms ) == 0:
                    continue
                cmd = [ renderer, clfile ] + flags
                if options.diff.lower() == "similarity":
                    cmd += [ "--similarity" ]
                else:
                    cmd += [ "--diff", options.diff ]
                if options.gray:
                    cmd.append( "--gray" )
                for ss in zooms:
                    diffimg = ctx.append( mktemp( ".ppm" ) )
                    cmd += [ "-i",
                        "-o", diffimg,
                        "--size", size,
                        "--zoom", str( ss ),
                        "-g", matrix[ "golden" ][ len( matrix[ diffkey ] ) ],
                    ]
                    matrix[ diffkey ].append( diffimg )
                check_call( [ "timeout", str( options.timeout ) ] + cmd )

        if options.dft:
            for persp, flags, zooms in modes:
                if len( zooms ) == 0:
                    continue
                cmd = [ renderer, clfile, "--aliasing" ]
                if options.gray:
                    cmd.append( "--gray" )
                for ss in zooms:
                    dftimg = ctx.append( mktemp( ".ppm" ) )
                    cmd += [ "-i",
                        "-o", dftimg,
                        "--size", size,
                        "--zoom", str( ss )
                    ]
                    matrix[ dftkey ].append( dftimg )
            check_call( [ "timeout", str( options.timeout ) ] + cmd )

        fitness = map( float, fitnesses.pop( 0 ).strip().split( "," ) )
        fitness = ", ".join( map( lambda x: "%4.3f" % x, fitness ) )
        matrix[ source ].append( write( fitness, ctx ) )
        if options.diff is not None:
            matrix[ diffkey ].append( blank )
        if options.dft:
            matrix[ dftkey ].append( blank )

    rows = len( options.ortho ) + len( options.persp ) + 2
    cmd = [
        "montage",
            "-tile", "%dx%d" % ( len( matrix ), rows ),
            "-mode", "Concatenate",
            "-frame", "2x2"
    ]
    for i in range( rows ):
        cmd.append( matrix[ "golden" ][ i ] )
        for source in sources:
            cmd.append( matrix[ source ][ i ] )
            if options.diff is not None:
                cmd.append( matrix[ ( source, "diff" ) ][ i ] )
            if options.dft:
                cmd.append( matrix[ ( source, "dft" ) ][ i ] )
    cmd.append( outfile )
    check_call( cmd )

