kernel
void
set_buffers(
    global float * g_txRand_f,
    global float4 * g_txRand_f4,
    global global_buffers * g
) {
    unsigned w = g->sizex;

    g->g_txRand_f = g_txRand_f;
    g->g_txRand_f4 = g_txRand_f4;

    for ( unsigned j = 0; j < g->sizey; ++j ) {
        for ( unsigned i = 0; i < g->sizex; ++i ) {
            float4 accum = g_txRand_f4[ j * w + i ];
            if ( i > 0 )
                accum += g_txRand_f4[ j * w + (i-1) ];
            if ( j > 0 )
                accum += g_txRand_f4[ (j-1) * w + i ];
            if ( i > 0 && j > 0 )
                accum -= g_txRand_f4[ (j-1) * w + (i-1) ];
            g_txRand_f4[ j * w + i ] = accum;
        }
    }
}

float3
projectPoint( float3 p, float scalex )
{
    // apply projection

    float phi = radians( 59.0f );
    float denom = p.y * sin( phi ) + p.z * cos( phi );
    p = (float3)(
        p.z * p.x / denom,
        p.z * ( p.y * cos( phi ) - p.z * sin( phi ) ) / denom,
        p.z
    );

    // rotate plane

    float theta = radians( scalex );
    p = (float3)(
        p.x * cos( theta ) - p.y * sin( theta ),
        p.x * sin( theta ) + p.y * cos( theta ),
        p.z
    );

    return p;
}

#ifdef HAMMERSLEY
constant float hammersley_offsets[] = {
    0.0f,
    0.5f,
    0.25f, 0.75f,
    0.125f, 0.635f, 0.325f, 0.875f,
    0.0625f, 0.5625f, 0.3125f, 0.8125f,
    0.1875f, 0.6875f, 0.4375f, 0.9375f
};
#endif

kernel
void
plot(
    global global_buffers * g,
    global uint2 * state,
    float scale,
    float step,
    global float4 * result,
    global float * c
) {
    VS_OUTPUT myin;
    float x = get_global_id( 0 );
    float y = get_global_id( 1 );
    uint maxx = get_global_size( 0 );
    uint maxy = get_global_size( 1 );
    uint index = get_global_id( 1 ) * maxx + get_global_id( 0 );

    // jitter sample position

#ifdef HAMMERSLEY
    if ( state != 0 ) {
        x += (float)state[ index ].y / (float)state[ index ].x - 0.5f;
        y += hammersley_offsets[ state[ index ].y ] - 0.5f;
    }
#else
    if ( state != 0 ) {
        unsigned st = state[ index ].x;
        float2 jitter = convert_float2( prng_normal( &st, 0.0f, GAUSS_SIGMA_FACTOR ) );
        state[ index ].x = st;
        x += jitter.x;
        y += jitter.y;
    }
#endif

#ifdef PERSPECTIVE
    // position in screen coordinates

    float3 scrn = (float3)(
        x / maxx - 0.5f, y / maxy - 0.5f, sqrt( 3.0f ) * 0.5f
    );
    float3 pos = projectPoint( scrn, scale );

    // Assign to myin.dwdx and myin.dwdy field-by-field in case we're using
    // double precision. We can't auto-convert a float3, but we can auto-convert
    // individual floats...

    float3 tmp =
        projectPoint( scrn + (float3)(0.5f/maxx, 0.0f, 0.0f), scale ) -
        projectPoint( scrn - (float3)(0.5f/maxx, 0.0f, 0.0f), scale );
    myin.dwdx.x = tmp.x;
    myin.dwdx.y = tmp.y;
    myin.dwdx.z = tmp.z;
    tmp =
        projectPoint( scrn + (float3)(0.0f, 0.5f/maxy, 0.0f), scale ) -
        projectPoint( scrn - (float3)(0.0f, 0.5f/maxy, 0.0f), scale );
    myin.dwdy.x = tmp.x;
    myin.dwdy.y = tmp.y;
    myin.dwdy.z = tmp.z;
#else
    float3 pos = (float3)(
        ( ( x / maxx ) - 0.5f ) * scale,
        ( ( y / maxy ) - 0.5f ) * scale,
        sqrt( 3.0f ) * 0.5f
    );
    myin.dwdx = scale / maxx;
    myin.dwdy = scale / maxy;
#endif

    myin.worldPos.x  = pos.x;
    myin.worldPos.y  = pos.y;
    myin.worldPos.z  = pos.z;
    myin.objPos.x    = pos.x;
    myin.objPos.y    = pos.y;
    myin.objPos.z    = pos.z;
    myin.vLightDirVec.x = pos.x;
    myin.vLightDirVec.y = pos.y;
    myin.vLightDirVec.z = pos.z;
    myin.worldNorm.x = 0.0;
    myin.worldNorm.y = 0.0;
    myin.worldNorm.z = 1.0;

    float4 color = convert_float4( TEXPROC( g, &myin, myin.worldNorm.xy, c ) );
    /*
    float4 color = (float4)(
        myin.dwdx.x / length( pos.xy ) * maxx,
        myin.dwdy.y / length( pos.xy ) * maxy,
        0, 0
    );
    */
#ifdef GRAYSCALE
    color = dot( color, (float4)(0.2126, 0.7152, 0.0722, 0) );
#endif
    if ( state != 0 ) {
        uint n = state[ index ].y;
        if ( n == 0 )
            result[ index ] = color;
        else
            result[ index ] += ( color - result[ index ] ) / n;
        state[ index ].y = n + 1;
    } else {
        result[ index ] = color;
    }
}

