open Cil
open Pretty
open Utils

let opt v s = v := Some(s)

let doDouble = ref false
let outfile = ref None

let options = [
  ("--double", Arg.Set(doDouble), " Use double-precision floats");
  ("-o", Arg.String(opt outfile), "file Write OpenCL code to named file");
]

let floatType = TFloat(FFloat,[])

let flip f x y = f y x

(* pulled from genprog *)
let file_to_string fname =
  let b = Buffer.create 255 in
  begin
    try
      let fin = open_in fname in
      try
        while true do
          let line = input_line fin in
          Buffer.add_string b line;
          Buffer.add_char b '\n';
        done
      with _ ->
        close_in fin
    with _ -> ()
  end;
  Buffer.contents b
    
(* cribbed liberally from genprog's get_standard_headers and
   can_repair_location *)
let filter_standard_headers () =
  let tfile = Filename.temp_file "headers" ".txt" in
  try
    let command = "`gcc -print-prog-name=cc1` -v < /dev/null 2>" ^ tfile in
    match Unix.system command with
    | Unix.WEXITED(0) ->
      let path_file_str =
        try
          file_to_string tfile
        with _ -> ""
      in
      Sys.remove tfile;
      let regexp = Str.regexp "[:= \n\t]" in
      let parts = Str.split regexp path_file_str in
      let parts = List.filter (fun x ->
        x <> "" && x.[0] = '/'
      ) parts in
      fun g ->
        let prefix_match directory file =
          let minsize = min (String.length directory) (String.length file) in
          let suba = String.sub directory 0 minsize in
          let subb = String.sub file 0 minsize in
            ((String.length file) > (String.length directory)) && (suba = subb)
        in
          List.exists (fun d ->
            prefix_match d (get_globalLoc g).file
          ) parts
    | _ ->
      if Sys.file_exists tfile then
        Sys.remove tfile;
      fun _ -> false

  with e ->
    Sys.remove tfile;
    raise e

let filter_opencl_keywords g =
  let types =
    List.fold_left (flip SSet.add) SSet.empty [
      "bool"; "uchar"; "ushort"; "uint"; "ulong"; "half"; "size_t"; "ptrdiff_t";
      "intptr_t"; "uintptr_t"; "image2d_t"; "imag3d_t"; "sample_t"; "event_t";
      "char2"; "char3"; "char4"; "char8"; "char16";
      "uchar2"; "uchar3"; "uchar4"; "uchar8"; "uchar16";
      "short2"; "short3"; "short4"; "short8"; "short16";
      "ushort2"; "ushort3"; "ushort4"; "ushort8"; "ushort16";
      "int2"; "int3"; "int4"; "int8"; "int16";
      "uint2"; "uint3"; "uint4"; "uint8"; "uint16";
      "long2"; "long3"; "long4"; "long8"; "long16";
      "ulong2"; "ulong3"; "ulong4"; "ulong8"; "ulong16";
      "float2"; "float3"; "float4"; "float8"; "float16";
    ]
  in
  let functions =
    List.fold_left (flip SSet.add) SSet.empty [
      "get_work_dim"; "get_global_size"; "get_global_id"; "get_local_size";
      "get_local_id"; "get_num_groups"; "get_group_id";

      "abs"; "abs_diff"; "add_sat"; "hadd"; "rhadd"; "clz"; "mad_hi"; "mad24";
      "mad_sat"; "max"; "min"; "mul_hi"; "mul24"; "rotate"; "sub_sat";

      "clamp"; "degrees"; "mix"; "radians"; "step"; "smoothstep"; "sign";

      "acos"; "acosh"; "acospi"; "asin"; "asinh"; "asinpi"; "atan"; "atan2";
      "atanh"; "atanpi"; "cbrt"; "ceil"; "copysign"; "cos"; "cosh"; "cospi";
      "half_divide"; "native_divide"; "erfc"; "erf"; "exp"; "exp2"; "exp10"; 
      "expm1"; "fabs"; "fdim"; "floor"; "fma"; "fmax"; "fmin"; "fmod"; "fract";
      "frexp"; "hypot"; "ilogb"; "ldexp"; "lgamma"; "lgamma_r"; "log"; "log2";
      "log10"; "log1p"; "logb"; "mad"; "modf"; "nan"; "nextafter"; "pow";
      "pown"; "powr"; "half_recip"; "native_recip"; "remainder"; "remquo";
      "rint"; "rootn"; "round"; "rsqrt"; "sin"; "sincos"; "sinh"; "sinpi";
      "sqrt"; "tan"; "tanh"; "tanpi"; "tgamma"; "trunc";

      "cross"; "dot"; "distance"; "length"; "normalize"; "fast_distance";
      "fast_length"; "fast_normalize";

      "isequal"; "isnotequal"; "isgreater"; "isgreaterequal"; "isless";
      "islessequal"; "islessgreater"; "isfinite"; "isinf"; "isnan"; "isnormal";
      "isordered"; "isunordered"; "signbit"; "any"; "all"; "bitselect";
      "select";

      "barrier"; "mem_fence"; "read_mem_fence"; "write_mem_fence";
    ]
  in
  match g with
  | GType(ti,_)        -> SSet.mem ti.tname types
  | GCompTag(ci,_)     -> SSet.mem ci.cname types
  | GCompTagDecl(ci,_) -> SSet.mem ci.cname types
  | GEnumTag(ei,_)     -> SSet.mem ei.ename types
  | GEnumTagDecl(ei,_) -> SSet.mem ei.ename types
  | GVarDecl(vi,_)     -> SSet.mem vi.vname functions
  | GVar(vi,_,_)       -> SSet.mem vi.vname functions
  | GFun(fd,_) when SSet.mem fd.svar.vname functions ->
    Printf.fprintf stderr "warning: %s: %s: definition of OpenCL built-in"
      (sprint ~width:80 (d_loc () (get_globalLoc g))) fd.svar.vname;
    true
  | _ -> false

class makeGlobalVarsConst =
object (self)
  inherit nopCilVisitor

  method vglob g =
    match g with
    | GVar(vi, _, _)
    | GVarDecl(vi, _) ->
      if not (isFunctionType vi.vtype) then begin
        let t = visitCilType (self :> cilVisitor) vi.vtype in
        let a = visitCilAttributes (self :> cilVisitor) vi.vattr in
        vi.vtype <- t;
        vi.vattr <- a;
      end;
      SkipChildren
    | _ -> SkipChildren

  method vtype t =
    ChangeDoChildrenPost(typeRemoveAttributes ["const"] t, fun t -> t)

  method vattr (Attr(n, params)) =
    let map_filter f lst =
      List.rev (List.fold_left (fun lst x ->
        match f x with
        | Some(y) -> y :: lst
        | None -> lst
      ) [] lst)
    in
    let filter_params p =
      match p with
      | AStr("constant") | ACons("constant", _) -> None
      | _ -> Some(p)
    in
    if n = "opencl" then begin
      let params' = map_filter filter_params params in
      if (List.length params') > 0 then
        ChangeTo([Attr(n, params')])
      else
        ChangeTo([])
    end else
      SkipChildren
end

let usedDoubles = ref false

class allFloatsAreDouble =
object
  inherit nopCilVisitor

  method vtype t =
    match t with
    | TFloat(FDouble,_) -> DoChildren
    | TFloat(_,a) ->
      usedDoubles := true ;
      ChangeTo(TFloat(FDouble,a))
    | TNamed(ti,_) when startsWith "float" ti.tname ->
      usedDoubles := true ;
      let suffix = String.sub ti.tname 5 ((String.length ti.tname)-5) in
      ti.tname <- "double" ^ suffix;
      DoChildren
    | TComp(ci,_) when startsWith "float" ci.cname ->
      usedDoubles := true ;
      let suffix = String.sub ci.cname 5 ((String.length ci.cname)-5) in
      ci.cname <- "double" ^ suffix;
      DoChildren
    | _ -> DoChildren

  method vexpr e =
    match e with
    | Const(CReal(_, FDouble, _)) ->
      usedDoubles := true ;
      DoChildren
    | Const(CReal(f, _, _)) ->
      usedDoubles := true ;
      ChangeTo( Const(CReal(f, FDouble, None)) )
    | _ -> DoChildren
end

class allFloatsAreFloat =
object
  inherit nopCilVisitor

  method vtype t =
    match t with
    | TFloat(FFloat,_) -> DoChildren
    | TFloat(_,a) when List.exists (match_attr "opencl" "force_double") a ->
      usedDoubles := true ;
      SkipChildren
    | TFloat(_,a) -> ChangeTo(TFloat(FFloat,a))
    | _ -> DoChildren

  method vinst i =
    match i with
    | Set(_, e, _) ->
      begin match typeOf e with
      | TFloat(_,a) when List.exists (match_attr "opencl" "force_double") a ->
        let _ = visitCilInstr (new allFloatsAreDouble) i in
        SkipChildren
      | _ -> DoChildren
      end
    | _ -> DoChildren
end

class noMoreStatic =
object
  inherit nopCilVisitor

  method vvdec vi =
    vi.vstorage <- NoStorage;
    DoChildren
end

class dontCastArguments =
object
  inherit nopCilVisitor

  method vinst i =
    match i with
    | Call(lv, f, args, loc) ->
      let _, formals, _, _ = splitFunctionType (typeOf f) in
      let formals = argsToList formals in
      let args =
        try
          List.map2 (fun e (_, t, _) ->
            match e with
            | CastE(t', e') when (typeSig t) = (typeSig t') -> e'
            | _ -> e
          ) args formals
        with Invalid_argument _ ->
          failwith (
            (sprint ~width:80 (d_loc () loc))
            ^ ": call to " ^ (sprint ~width:80 (dn_exp () f)) ^ ": got "
            ^ (string_of_int (List.length args)) ^ " args ("
            ^ (string_of_int (List.length formals)) ^ " expected)"
          )
          failwith ("List.map2: " ^ (sprint ~width:80 (d_instr () i)))
      in
      ChangeTo([Call(lv, f, args, loc)])
    | _ -> DoChildren
end

class makeSwizzleTemps =
object
  inherit nopCilVisitor

  val mutable current_fd = None

  method vfunc fd =
    let old_fd = current_fd in
    current_fd <- Some(fd);
    ChangeDoChildrenPost(fd, fun fd -> current_fd <- old_fd; fd)

  method vinst i =
    match i with
    | Call(res, Lval(Var(vi), NoOffset), [arg], loc)
        when ((String.length vi.vname) > 7)
          && ((String.sub vi.vname 0 7) = "swizzle") ->
      begin match arg with
      | Lval(Var(_), _) -> SkipChildren
      | _ ->
        let Some(fd) = current_fd in
        let tmp = makeTempVar fd (typeOf arg) in
        ChangeTo([
          Set(var tmp, arg, locUnknown);
          Call(res, Lval(Var(vi), NoOffset), [Lval(var tmp)], loc);
        ])
      end
    | Call(res, Lval(Var(vi), NoOffset), [dst; src], loc)
        when ((String.length vi.vname) > 10)
          && ((String.sub vi.vname 0 10) = "swizzleset") ->
      begin match src with
      | Lval(Var(_), _) -> SkipChildren
      | _ ->
        let Some(fd) = current_fd in
        let tmp = makeTempVar fd (typeOf src) in
        ChangeTo([
          Set(var tmp, src, locUnknown);
          Call(res, Lval(Var(vi), NoOffset), [dst; Lval(var tmp)], loc);
        ])
      end
    | _ -> SkipChildren
end

class rewriteApi =
object
  inherit nopCilVisitor

  val mutable current_fd = dummyFunDec

  val functions = Hashtbl.create 257

  method vglob g =
    begin match g with
    | GFun(fd,_) ->
      Hashtbl.replace functions fd.svar.vname fd.svar
    | GVarDecl(vi,_) when isFunctionType vi.vtype ->
      Hashtbl.replace functions vi.vname vi
    | _ -> ()
    end;
    DoChildren

  method vfunc fd =
    let old_fd = current_fd in
    current_fd <- fd;
    ChangeDoChildrenPost(fd, fun fd -> current_fd <- old_fd; fd)

  method vinst i =
    let suffix name =
      let last = (String.length name) - 1 in
      let c = String.get name last in
      if ('0' <= c) && (c <= '9') then String.sub name last 1 else ""
    in
    let resolve_vectors rv op lhs rhs loc =
      let get_op_fun n1 n2 =
        let funname =
          match op with
          | PlusA  -> Printf.sprintf "add_f%sf%s" n1 n2
          | MinusA -> Printf.sprintf "sub_f%sf%s" n1 n2
          | Mult   -> Printf.sprintf "mul_f%sf%s" n1 n2
          | Div    -> Printf.sprintf "div_f%sf%s" n1 n2
          | _ -> failwith "logic error"
        in
        Lval(var (Hashtbl.find functions funname))
      in
      match unrollType (typeOf lhs), unrollType (typeOf rhs) with
      | TComp(ci1,_), TComp(ci2,_) ->
        let f = get_op_fun (suffix ci1.cname) (suffix ci2.cname) in
        Call(Some(rv), f, [lhs; rhs], loc)
      | TComp(ci1,_), TFloat(_) ->
        let f = get_op_fun (suffix ci1.cname) "" in
        Call(Some(rv), f, [lhs; rhs], loc)
      | TFloat(_), TFloat(_) ->
        Set(rv, BinOp(op, lhs, rhs, floatType), loc)
      | _ -> failwith "logic error"
    in
    match i with
    | Call(Some(rv), Lval(Var(vi), NoOffset), args, loc)
        when startsWith "fract_f" vi.vname ->
      let tmp = var (makeTempVar current_fd (typeOfLval rv)) in
      let s = suffix vi.vname in
      let floor = Lval(var (Hashtbl.find functions ("floor_f"^s))) in
      ChangeTo( [
        Call(Some(tmp), floor, args, loc);
        resolve_vectors rv MinusA (List.hd args) (Lval tmp) loc;
      ] )
    | Call(Some(rv), Lval(Var(vi), NoOffset), args, loc)
        when startsWith "sq_f" vi.vname ->
      ChangeTo( [ resolve_vectors rv Mult (List.hd args) (List.hd args) loc ] )
    | _ -> DoChildren
end

class openclPrinterClass =
object (self)
  inherit defaultCilPrinterClass as super

  val blacklist = [
    filter_standard_headers();
    filter_builtin;
    filter_api3d ();
    filter_opencl_keywords
  ]

  val whitelist =
    [ retain_marked_functions ]

  val swizzle_ci =
    mkCompInfo true "swizzle" (fun _ ->
      List.map (fun n -> n, floatType, None, [], locUnknown) [
        "xx"; "xy"; "xz"; "xw";
        "yx"; "yy"; "yz"; "yw";
        "zx"; "zy"; "zz"; "zw";
        "wx"; "wy"; "wz"; "ww";
        "xxx"; "xxy"; "xxz"; "xxw";
        "xyx"; "xyy"; "xyz"; "xyw";
        "xzx"; "xzy"; "xzz"; "xzw";
        "xwx"; "xwy"; "xwz"; "xww";
        "yxx"; "yxy"; "yxz"; "yxw";
        "yyx"; "yyy"; "yyz"; "yyw";
        "yzx"; "yzy"; "yzz"; "yzw";
        "ywx"; "ywy"; "ywz"; "yww";
        "zxx"; "zxy"; "zxz"; "zxw";
        "zyx"; "zyy"; "zyz"; "zyw";
        "zzx"; "zzy"; "zzz"; "zzw";
        "zwx"; "zwy"; "zwz"; "zww";
        "wxx"; "wxy"; "wxz"; "wxw";
        "wyx"; "wyy"; "wyz"; "wyw";
        "wzx"; "wzy"; "wzz"; "wzw";
        "wwx"; "wwy"; "wwz"; "www";
        "zyxw";
      ]
    ) []

  val mutable in_kernel = false

  method dGlobal out g =
    let blacklisted = List.fold_left (fun b f -> b || f g) false blacklist in
    let whitelisted = List.fold_left (fun b f -> b || f g) false whitelist in
    if whitelisted || (not blacklisted) then begin
      let post = ref [] in
      begin match g with
        | GVar(vi, _, _)
        | GVarDecl(vi, _) when not (isFunctionType vi.vtype) ->
          Printf.fprintf out "constant "
        | GFun(fd, _)
            when List.exists (match_attr "opencl" "kernel") fd.svar.vattr ->
          Printf.fprintf out "kernel ";
          in_kernel <- true;
          let attrs =
            extract_function_attributes (match_attr "opencl" "kernel") fd
          in
          post := (fun () ->
            in_kernel <- false;
            fd.svar.vattr <- attrs @ fd.svar.vattr
          ) :: !post
        | _ -> ()
      end;
      super#dGlobal out g;
      List.iter (fun f -> f()) !post
    end

  method pAttr (Attr(n, _) as a) =
    if n = "opencl" then begin
      List.fold_left (fun ((doc,_) as accum) addrspace ->
        if match_attr "opencl" addrspace a then
          (doc ++ text " " ++ text addrspace), false
        else
          accum
      ) (nil, false) [ "local"; "global"; "constant"; "private" ]
    end else super#pAttr a

  method pInstr () i =
    let real = if !doDouble then "double" else "float" in
    let rkind = if !doDouble then FDouble else FFloat in
    let zerof = Const(CReal(0.0, rkind, None)) in
    let onef  = Const(CReal(1.0, rkind, None)) in

    match i with
    | Call(result, Lval(Var(f), NoOffset), args, loc) ->
      let rename_f ?(args=args) newname =
        let oldname = f.vname in
        f.vname <- newname;
        let res = super#pInstr () (Call(result, Lval(var f), args, loc)) in
        f.vname <- oldname;
        res
      in
      let to_binop op =
        match result with
        | Some(result') ->
          let a, b =
            match args with
            | [a; b] -> a, b
            | _ -> failwith "wrong number of args to binary operator"
          in
          let e = BinOp(op, a, b, typeOf a) in
          let i' = 
            (* Negate result of comparison operators for consistency with
               results in C. OpenCL specifies they return -1 instead of 1.
               Also cast to appropriate int vectory type, since comparisons of
               doubles may return longs for some reason. *)
            match op with
            | Lt|Gt|Le|Ge|Eq|Ne ->
              let e' = UnOp(Neg, e, typeOf e) in
              begin match typeOfLval result' with
              | TComp(ci,_) ->
                let castfun = makeGlobalVar ("convert_"^ci.cname) voidType in
                Call(result, Lval(var castfun), [e], loc)
              | TNamed(ti,_) ->
                let castfun = makeGlobalVar ("convert_"^ti.tname) voidType in
                Call(result, Lval(var castfun), [e], loc)
              | _ -> Set(result', UnOp(Neg, e, typeOf e), loc)
              end
            | _ -> Set(result', e, loc)
          in
          super#pInstr () i'
        | _ -> nil
      in
      let to_lval swiz =
        let off = Field(getCompField swizzle_ci swiz, NoOffset) in
        match result with
        | Some(result') ->
          let arg =
            match List.hd args with
            | Lval(lv) -> Lval(addOffsetLval off lv)
            | _ -> failwith "cannot swizzle non-lval"
          in
          let i' = Set(result', arg, loc) in
          super#pInstr () i'
        | _ -> nil
      in
      let swiz_assignment swiz =
        let off = Field(getCompField swizzle_ci swiz, NoOffset) in
        let result', b =
          match args with
          | [a; b] -> mkMem a off, b
          | _ -> failwith "wrong number of args to swizzleset"
        in
        super#pInstr () (Set(result', b, loc))
      in
      begin match f.vname with
      | "abs_f"         -> rename_f "fabs"
      | "abs_f2"        -> rename_f "fabs"
      | "abs_f3"        -> rename_f "fabs"
      | "abs_f4"        -> rename_f "fabs"
      | "add_i2i"       -> to_binop PlusA
      | "add_i2i2"      -> to_binop PlusA
      | "add_i3i"       -> to_binop PlusA
      | "add_i3i3"      -> to_binop PlusA
      | "add_i4i"       -> to_binop PlusA
      | "add_i4i4"      -> to_binop PlusA
      | "add_f2f"       -> to_binop PlusA
      | "add_f2f2"      -> to_binop PlusA
      | "add_f3f"       -> to_binop PlusA
      | "add_f3f3"      -> to_binop PlusA
      | "add_f4f"       -> to_binop PlusA
      | "add_f4f4"      -> to_binop PlusA
      | "and_i2i"       -> to_binop BAnd
      | "and_i2i2"      -> to_binop BAnd
      | "and_i3i"       -> to_binop BAnd
      | "and_i3i3"      -> to_binop BAnd
      | "and_i4i"       -> to_binop BAnd
      | "and_i4i4"      -> to_binop BAnd
      | "ceil_f"        -> rename_f "ceil"
      | "ceil_f2"       -> rename_f "ceil"
      | "ceil_f3"       -> rename_f "ceil"
      | "ceil_f4"       -> rename_f "ceil"
      | "div_i2i"       -> to_binop Div
      | "div_i2i2"      -> to_binop Div
      | "div_i3i"       -> to_binop Div
      | "div_i3i3"      -> to_binop Div
      | "div_i4i"       -> to_binop Div
      | "div_i4i4"      -> to_binop Div
      | "div_f2f"       -> to_binop Div
      | "div_f2f2"      -> to_binop Div
      | "div_f3f"       -> to_binop Div
      | "div_f3f3"      -> to_binop Div
      | "div_f4f"       -> to_binop Div
      | "div_f4f4"      -> to_binop Div
      | "dot_f"         -> rename_f "dot"
      | "dot_f2"        -> rename_f "dot"
      | "dot_f3"        -> rename_f "dot"
      | "dot_f4"        -> rename_f "dot"
      | "float2_i2"     -> rename_f ("convert_"^real^"2")
      | "float3_i3"     -> rename_f ("convert_"^real^"3")
      | "float4_i4"     -> rename_f ("convert_"^real^"4")
      | "floor_f"       -> rename_f "floor"
      | "floor_f2"      -> rename_f "floor"
      | "floor_f3"      -> rename_f "floor"
      | "floor_f4"      -> rename_f "floor"
      | "ge_i2i2"       -> to_binop Ge
      | "gt_i2i2"       -> to_binop Gt
      | "ge_i3i3"       -> to_binop Ge
      | "gt_i3i3"       -> to_binop Gt
      | "ge_i4i4"       -> to_binop Ge
      | "gt_i4i4"       -> to_binop Gt
      | "ge_f2f2"       -> to_binop Ge
      | "gt_f2f2"       -> to_binop Gt
      | "ge_f3f3"       -> to_binop Ge
      | "gt_f3f3"       -> to_binop Gt
      | "ge_f4f4"       -> to_binop Ge
      | "gt_f4f4"       -> to_binop Gt
      | "int2_f2"       -> rename_f "convert_int2"
      | "int3_f3"       -> rename_f "convert_int3"
      | "int4_f4"       -> rename_f "convert_int4"
      | "length_f"      -> rename_f "length"
      | "length_f2"     -> rename_f "length"
      | "length_f3"     -> rename_f "length"
      | "length_f4"     -> rename_f "length"
      | "makei2_i"      -> rename_f "(int2)"
      | "makei2_ii"     -> rename_f "(int2)"
      | "makei3_i"      -> rename_f "(int3)"
      | "makei3_i2i"    -> rename_f "(int3)"
      | "makei3_iii"    -> rename_f "(int3)"
      | "makei4_i"      -> rename_f "(int4)"
      | "makei4_i3i"    -> rename_f "(int4)"
      | "makei4_iii"    -> rename_f "(int4)"
      | "makef2_f"      -> rename_f ("("^real^"2)")
      | "makef2_ff"     -> rename_f ("("^real^"2)")
      | "makef3_f"      -> rename_f ("("^real^"3)")
      | "makef3_f2f"    -> rename_f ("("^real^"3)")
      | "makef3_fff"    -> rename_f ("("^real^"3)")
      | "makef4_f"      -> rename_f ("("^real^"4)")
      | "makef4_f3f"    -> rename_f ("("^real^"4)")
      | "makef4_ffff"   -> rename_f ("("^real^"4)")
      | "mul_i2i"       -> to_binop Mult
      | "mul_i2i2"      -> to_binop Mult
      | "mul_i3i"       -> to_binop Mult
      | "mul_i3i3"      -> to_binop Mult
      | "mul_i4i"       -> to_binop Mult
      | "mul_i4i4"      -> to_binop Mult
      | "mul_f2f"       -> to_binop Mult
      | "mul_f2f2"      -> to_binop Mult
      | "mul_f3f"       -> to_binop Mult
      | "mul_f3f3"      -> to_binop Mult
      | "mul_f4f"       -> to_binop Mult
      | "mul_f4f4"      -> to_binop Mult
      | "normalize_f2"  -> rename_f "normalize" 
      | "normalize_f3"  -> rename_f "normalize" 
      | "normalize_f4"  -> rename_f "normalize" 
      | "or_i2i"        -> to_binop BOr
      | "or_i2i2"       -> to_binop BOr
      | "or_i3i"        -> to_binop BOr
      | "or_i3i3"       -> to_binop BOr
      | "or_i4i"        -> to_binop BOr
      | "or_i4i4"       -> to_binop BOr
      | "saturate_f"    -> rename_f ~args:(args @ [ zerof; onef ]) "clamp"
      | "saturate_f2"   -> rename_f ~args:(args @ [ zerof; onef ]) "clamp"
      | "saturate_f3"   -> rename_f ~args:(args @ [ zerof; onef ]) "clamp"
      | "saturate_f4"   -> rename_f ~args:(args @ [ zerof; onef ]) "clamp"
      | "smoothstep_f"  -> rename_f "smoothstep"
      | "smoothstep_f2" -> rename_f "smoothstep"
      | "smoothstep_f3" -> rename_f "smoothstep"
      | "smoothstep_f4" -> rename_f "smoothstep"
      | "step_f"        -> rename_f "step"
      | "step_f2"       -> rename_f "step"
      | "step_f3"       -> rename_f "step"
      | "step_f4"       -> rename_f "step"
      | "sub_ii2"       -> to_binop MinusA
      | "sub_i2i2"      -> to_binop MinusA
      | "sub_ii3"       -> to_binop MinusA
      | "sub_i3i3"      -> to_binop MinusA
      | "sub_ii4"       -> to_binop MinusA
      | "sub_i4i4"      -> to_binop MinusA
      | "sub_ff2"       -> to_binop MinusA
      | "sub_f2f2"      -> to_binop MinusA
      | "sub_ff3"       -> to_binop MinusA
      | "sub_f3f3"      -> to_binop MinusA
      | "sub_ff4"       -> to_binop MinusA
      | "sub_f4f4"      -> to_binop MinusA
      | "swizzle_i2xx"  -> to_lval "xx"
      | "swizzle_i2xy"  -> to_lval "xy"
      | "swizzle_i2yx"  -> to_lval "yx"
      | "swizzle_i2yy"  -> to_lval "yy"
      | "swizzle_i3xx"  -> to_lval "xx"
      | "swizzle_i3xy"  -> to_lval "xy"
      | "swizzle_i3xz"  -> to_lval "xz"
      | "swizzle_i3yx"  -> to_lval "yx"
      | "swizzle_i3yy"  -> to_lval "yy"
      | "swizzle_i3yz"  -> to_lval "yz"
      | "swizzle_i3zx"  -> to_lval "zx"
      | "swizzle_i3zy"  -> to_lval "zy"
      | "swizzle_i3zz"  -> to_lval "zz"
      | "swizzle_f3xx"  -> to_lval "xx"
      | "swizzle_f3xy"  -> to_lval "xy"
      | "swizzle_f3xz"  -> to_lval "xz"
      | "swizzle_f3yx"  -> to_lval "yx"
      | "swizzle_f3yy"  -> to_lval "yy"
      | "swizzle_f3yz"  -> to_lval "yz"
      | "swizzle_f3zx"  -> to_lval "zx"
      | "swizzle_f3zy"  -> to_lval "zy"
      | "swizzle_f3zz"  -> to_lval "zz"
      | "swizzle_f3xxx" -> to_lval "xxx"
      | "swizzle_f3xxy" -> to_lval "xxy"
      | "swizzle_f3xxz" -> to_lval "xxz"
      | "swizzle_f3xyx" -> to_lval "xyx"
      | "swizzle_f3xyy" -> to_lval "xyy"
      | "swizzle_f3xyz" -> to_lval "xyz"
      | "swizzle_f3xzx" -> to_lval "xzx"
      | "swizzle_f3xzy" -> to_lval "xzy"
      | "swizzle_f3xzz" -> to_lval "xzz"
      | "swizzle_f3yxx" -> to_lval "yxx"
      | "swizzle_f3yxy" -> to_lval "yxy"
      | "swizzle_f3yxz" -> to_lval "yxz"
      | "swizzle_f3yyx" -> to_lval "yyx"
      | "swizzle_f3yyy" -> to_lval "yyy"
      | "swizzle_f3yyz" -> to_lval "yyz"
      | "swizzle_f3yzx" -> to_lval "yzx"
      | "swizzle_f3yzy" -> to_lval "yzy"
      | "swizzle_f3yzz" -> to_lval "yzz"
      | "swizzle_f3zxx" -> to_lval "zxx"
      | "swizzle_f3zxy" -> to_lval "zxy"
      | "swizzle_f3zxz" -> to_lval "zxz"
      | "swizzle_f3zyx" -> to_lval "zyx"
      | "swizzle_f3zyy" -> to_lval "zyy"
      | "swizzle_f3zyz" -> to_lval "zyz"
      | "swizzle_f3zzx" -> to_lval "zzx"
      | "swizzle_f3zzy" -> to_lval "zzy"
      | "swizzle_f3zzz" -> to_lval "zzz"
      | "swizzle_f4xx"  -> to_lval "xx"
      | "swizzle_f4xy"  -> to_lval "xy"
      | "swizzle_f4xz"  -> to_lval "xz"
      | "swizzle_f4xw"  -> to_lval "xw"
      | "swizzle_f4yx"  -> to_lval "yx"
      | "swizzle_f4yy"  -> to_lval "yy"
      | "swizzle_f4yz"  -> to_lval "yz"
      | "swizzle_f4yw"  -> to_lval "yw"
      | "swizzle_f4zx"  -> to_lval "zx"
      | "swizzle_f4zy"  -> to_lval "zy"
      | "swizzle_f4zz"  -> to_lval "zz"
      | "swizzle_f4zw"  -> to_lval "zw"
      | "swizzle_f4wx"  -> to_lval "wx"
      | "swizzle_f4wy"  -> to_lval "wy"
      | "swizzle_f4wz"  -> to_lval "wz"
      | "swizzle_f4ww"  -> to_lval "ww"
      | "swizzle_f4xxx" -> to_lval "xxx"
      | "swizzle_f4xxy" -> to_lval "xxy"
      | "swizzle_f4xxz" -> to_lval "xxz"
      | "swizzle_f4xxw" -> to_lval "xxw"
      | "swizzle_f4xyx" -> to_lval "xyx"
      | "swizzle_f4xyy" -> to_lval "xyy"
      | "swizzle_f4xyz" -> to_lval "xyz"
      | "swizzle_f4xyw" -> to_lval "xyw"
      | "swizzle_f4xzx" -> to_lval "xzx"
      | "swizzle_f4xzy" -> to_lval "xzy"
      | "swizzle_f4xzz" -> to_lval "xzz"
      | "swizzle_f4xzw" -> to_lval "xzw"
      | "swizzle_f4xwx" -> to_lval "xwx"
      | "swizzle_f4xwy" -> to_lval "xwy"
      | "swizzle_f4xwz" -> to_lval "xwz"
      | "swizzle_f4xww" -> to_lval "xww"
      | "swizzle_f4yxx" -> to_lval "yxx"
      | "swizzle_f4yxy" -> to_lval "yxy"
      | "swizzle_f4yxz" -> to_lval "yxz"
      | "swizzle_f4yxw" -> to_lval "yxw"
      | "swizzle_f4yyx" -> to_lval "yyx"
      | "swizzle_f4yyy" -> to_lval "yyy"
      | "swizzle_f4yyz" -> to_lval "yyz"
      | "swizzle_f4yyw" -> to_lval "yyw"
      | "swizzle_f4yzx" -> to_lval "yzx"
      | "swizzle_f4yzy" -> to_lval "yzy"
      | "swizzle_f4yzz" -> to_lval "yzz"
      | "swizzle_f4yzw" -> to_lval "yzw"
      | "swizzle_f4ywx" -> to_lval "ywx"
      | "swizzle_f4ywy" -> to_lval "ywy"
      | "swizzle_f4ywz" -> to_lval "ywz"
      | "swizzle_f4yww" -> to_lval "yww"
      | "swizzle_f4zxx" -> to_lval "zxx"
      | "swizzle_f4zxy" -> to_lval "zxy"
      | "swizzle_f4zxz" -> to_lval "zxz"
      | "swizzle_f4zxw" -> to_lval "zxw"
      | "swizzle_f4zyx" -> to_lval "zyx"
      | "swizzle_f4zyy" -> to_lval "zyy"
      | "swizzle_f4zyz" -> to_lval "zyz"
      | "swizzle_f4zyw" -> to_lval "zyw"
      | "swizzle_f4zzx" -> to_lval "zzx"
      | "swizzle_f4zzy" -> to_lval "zzy"
      | "swizzle_f4zzz" -> to_lval "zzz"
      | "swizzle_f4zzw" -> to_lval "zzw"
      | "swizzle_f4zwx" -> to_lval "zwx"
      | "swizzle_f4zwy" -> to_lval "zwy"
      | "swizzle_f4zwz" -> to_lval "zwz"
      | "swizzle_f4zww" -> to_lval "zww"
      | "swizzle_f4zyxw" -> to_lval "zyxw"
      | "swizzleset_f4zw"  -> swiz_assignment "zw"
      | "swizzleset_f4xyz" -> swiz_assignment "xyz"
      | "swizzleset_f4yzw" -> swiz_assignment "yzw"
      | "xor_i2i"       -> to_binop BXor
      | "xor_i2i2"      -> to_binop BXor
      | "xor_i3i"       -> to_binop BXor
      | "xor_i3i3"      -> to_binop BXor
      | "xor_i4i"       -> to_binop BXor
      | "xor_i4i4"      -> to_binop BXor
      | _ -> super#pInstr () i
      end
    | _ -> super#pInstr () i
      
end
let openclPrinter = new openclPrinterClass

let main () = begin
  let usage   = "Usage: " ^ Sys.argv.(0) ^ " [options] filename procname" in
  let options = Arg.align options in
  let args = ref [] in
  Arg.parse options (fun x -> args := x :: !args) usage;

  let fname, proc =
    match List.rev !args with
    | [ fname; proc ] -> fname, proc
    | _ ->
      prerr_string "ERROR: you must provide a filename and procedure name\n\n";
      Arg.usage options usage;
      exit 1
  in

  let boilerplate =
    Filename.concat (Filename.dirname Sys.argv.(0)) "boilerplate.cl"
  in
  let cilfile = parse fname in
  Rmtmps.removeUnusedTemps cilfile;

  visitCilFile (new makeGlobalVarsConst) cilfile;
  visitCilFileSameGlobals (new globalIsForPointers) cilfile;
  visitCilFileSameGlobals (new allFloatsAreFloat) cilfile;
  visitCilFileSameGlobals (new noMoreStatic) cilfile;
  visitCilFileSameGlobals (new dontCastArguments) cilfile;
  visitCilFileSameGlobals (new makeSwizzleTemps) cilfile;
  visitCilFileSameGlobals (new rewriteApi) cilfile;

  if !doDouble then
    visitCilFileSameGlobals (new allFloatsAreDouble) cilfile;
  if !usedDoubles then
    cilfile.globals <-
      GText("#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n") :: cilfile.globals;

  let chan, close =
    match !outfile with
    | Some(fname) ->
      let chan = open_out fname in
      chan, fun () -> close_out chan
    | None ->
      stdout, fun () -> ()
  in
  lineDirectiveStyle := None;
  dumpFile openclPrinter chan (fname ^ ".cl") cilfile;

  let content = file_to_string boilerplate in
  let content = Str.global_replace (Str.regexp "TEXPROC") proc content in
  Printf.fprintf chan "%s" content;

  close ()
end ;;

main () ;;

