#include "hooke.h"
#include <algorithm>
#include <cmath>

OptimizerHooke::OptimizerHooke(ErrorFunk funk, void* funk_param, float* start_pos, int ndim, float rho, float epsilon)
{
  // copy
  this->funk = funk;
  this->funk_param = funk_param;
  this->ndim = ndim;
  this->current_pos = new float[ndim];
  this->end_pos = new float[ndim];    
  this->rho = rho;
  this->epsilon = epsilon;

  // alloc
  delta = new float[ndim];

  // reset
  reset(start_pos);
  
}

OptimizerHooke::~OptimizerHooke(void)
{
  delete[] current_pos;
  delete[] end_pos;
  delete[] delta;
}

void OptimizerHooke::reset(float* start_pos)
{
  // copy start pos
  std::copy(start_pos, &(start_pos[ndim]), this->current_pos);
}


float* OptimizerHooke::pos(void) const
{
  float* result = new float[ndim];
  std::copy(end_pos, &(end_pos[ndim]), result);
  return result;
}


inline void OptimizerHooke::bestNearby(void)
{
  // update current_point based on end point
  std::copy(end_pos, &(end_pos[ndim]), current_pos);
  current_error = error;

  for(unsigned int d=0; d < ndim; d++)
  {
    float tmp = current_pos[d];
    current_pos[d] = tmp + delta[d];
    float posErr = funk(current_pos, funk_param);

    if(posErr < current_error) current_error = posErr;
    else {
      current_pos[d] = tmp + delta[d];
      float negErr = funk(current_pos, funk_param);
      
      if(negErr < current_error) 
      {
	current_error = negErr;
	delta[d] *= -1.0f;   // optimization: test good directions first.
      }

      else current_pos[d] = tmp;
    }
  }

  // Done.
}


bool OptimizerHooke::step(void)
{
  // update current_pos with best nearby point from end_pos
  bestNearby();

  // pursue direction if improvement was made
  bool done = false;
  while(current_error < error && !done)
  {
    // move current_point again
    std::copy(current_pos, &(current_pos[ndim]), end_pos);
    error = current_error;

    // get error
    bestNearby();

    // bail if larger
    if(current_error >= error) done = true;

    // check if not due round-off
    else {
      done=true;
      for(unsigned d=0; d < ndim; d++)
	done &= ( fabs(current_pos[d] - end_pos[d]) < 0.5f * fabs(delta[d]) );
    }
  }

  if( steplength >= epsilon && current_error >= error )
  {
    steplength *= rho;
    for(unsigned int d=0; d < ndim; d++)
      delta[d] *= rho;
  }

  // Done.
  return (error < epsilon);
}


int OptimizerHooke::solve(int max_iter)
{
  // set initial delta
  for(int i =0; i < ndim; i++)
  {
    delta[i] = fabs(current_pos[i] * rho);
    if(delta[i] == 0.0f)
      delta[i] = rho;
  }

  // init
  steplength = rho;
  error = funk(current_pos, funk_param);
  std::copy(current_pos, &(current_pos[ndim]), end_pos);

  // iterate
  bool stepResult = true;
  unsigned long iter=0;
  while(iter < max_iter && steplength > epsilon)
  {
    step();
    iter++;
  }

  // Done.
  return iter;
}
  
