#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "simplex.h"
#include "math.h"

OptimizerSimplex::OptimizerSimplex(ErrorFunk _funk,
				   void* _funk_params, 
				   float* start_pos,
				   int _ndim)  
{
    timer.start();
    num_steps = 0;
    num_evals = 0;
    last_reset_step = 0;
    //    logfile = fopen("optimizer.log", "a");

    funk        = _funk;
    funk_params = _funk_params;
    ndim        = _ndim;

    mutate_position = new float[ndim];
    sum_positions   = new float[ndim];

    // Count how many times function generated infinities
    inf_count = 0;

    // initialize vertices of simplex
    positions = new float*[ndim+1];
    for (int i = 0; i < ndim+1; i++)
        positions[i]  = arraydup(start_pos, ndim);

    for (int i = 1; i < ndim+1; i++)
        positions[i][i-1] += 1.0f;

    // init errors for the simplex verts
    errors = new float[ndim+1];
    float besterr = 10.0;  
    for (int i = 0; i < ndim+1; i++)
      {
	errors[i] = funk(positions[i], funk_params);
        if (besterr > errors[i])
	  besterr = errors[i];  
	//        logEval(errors[i], besterr);
      }

    // init vector sum
    sumPositions();

    findBestWorst();
}

void OptimizerSimplex::reset(float *start_pos)
{
    num_steps = 0;
    num_evals = 0;
    last_reset_step = 0;
    visited.clear();

    // initialize vertices of simplex
    for (int i = 0; i < ndim+1; i++)
      memcpy(positions[i],start_pos,sizeof(float)*ndim);

    for (int i = 1; i < ndim+1; i++)
      positions[i][i-1] += 1.0f;

    // init errors for the simplex verts
    float besterr = 10.0;  
    for (int i = 0; i < ndim+1; i++)
      {
        errors[i] = funk(positions[i], funk_params);
        if (besterr > errors[i])
	  besterr = errors[i];  
	//        logEval(errors[i], besterr);
      }

    // init vector sum
    sumPositions();
}

void OptimizerSimplex::logEval(float error, float besterr)
{
    num_evals++;
    double time = timer.stop();
//    timer.start();
    
    //    fprintf(logfile, "%i, %f, %f, %f,\n", num_evals, error, besterr, time);
}

OptimizerSimplex::OptimizerSimplex(ErrorFunk _funk, void* _funk_params, 
        float* start_pos, int _ndim, float* steps)  
{
    timer.start();

    num_steps = 0;
    num_evals = 0;
    last_reset_step = 0;
    funk        = _funk;
    funk_params = _funk_params;
    ndim        = _ndim;
    mutate_position = new float[ndim];
    sum_positions   = new float[ndim];
    //    logfile = fopen("optimizer.log", "a");

    // initialize vertices of simplex
    positions = new float*[ndim+1];
    for (int i = 0; i < ndim+1; i++)
        positions[i]  = arraydup(start_pos, ndim);

    for (int i = 1; i < ndim+1; i++)
        positions[i][i-1] += steps[i-1];

    // init errors for the simplex verts
    errors = new float[ndim+1];
    float besterr = 10.0;  
    for (int i = 0; i < ndim+1; i++)
    {
        errors[i] = funk(positions[i], funk_params);
        if (besterr > errors[i])
            besterr = errors[i];  
        logEval(errors[i], besterr);
    }

    // init vector sum
    sumPositions();
}

OptimizerSimplex::OptimizerSimplex(ErrorFunk _funk, void* _funk_params, 
        float* start_pos, int _ndim, float stepSize)  
{
    timer.start();
    num_steps = 0;
    num_evals = 0;
    last_reset_step = 0;
    funk        = _funk;
    funk_params = _funk_params;
    ndim        = _ndim;
    mutate_position = new float[ndim];
    sum_positions   = new float[ndim];
    //    logfile = fopen("optimizer.log", "a");

    // initialize vertices of simplex
    positions = new float*[ndim+1];
    for (int i = 0; i < ndim+1; i++)
        positions[i]  = arraydup(start_pos, ndim);

    for (int i = 1; i < ndim+1; i++)
        positions[i][i-1] += stepSize;


    // init errors for the simplex verts
    errors = new float[ndim+1];
    float besterr = 10.0;  
    for (int i = 0; i < ndim+1; i++)
    {
        errors[i] = funk(positions[i], funk_params);
        if (besterr > errors[i])
            besterr = errors[i];  
        logEval(errors[i], besterr);
    }

    // init vector sum
    sumPositions();
}

OptimizerSimplex::~OptimizerSimplex() 
{
    for (int i = 0; i < ndim+1; i++)
        delete [] positions[i];
    delete [] positions;

    delete [] errors;
    delete [] mutate_position;
    delete [] sum_positions;
}

float* 
OptimizerSimplex::pos()const
{
    return arraydup(positions[best], ndim);
}

float old_range = 0.0f;

bool
OptimizerSimplex::step()
{
   
    num_steps++;

    findBestWorst();

    if (timer.stop() > 360.0)
    {
      //printf("timed out\n");
      return true;
    }

    float range;

    float delta=errors[worst]+errors[best];

    if(delta < 1e-10f && delta > -1e-10f){
      range=20000.0f;
    }else{
      range = 2.0f*(errors[worst] - errors[best]) / 
              (errors[worst] + errors[best]);
    }

//    printf(" range: %f worst: %f best: %f\n", range, errors[worst], errors[best]);

    //ABORT if function is generating infinities
    if (isinf(errors[best]) || isnan(errors[best]))
    {
      inf_count++;
//      printf("encountered infinite error\n");
    }

    if (num_steps % 5 == 0) {
      if (range == old_range) {
//        printf("got stuck\n\n");
        return true;
      }
      old_range = range;
    }
    if ( visited.count( range ) > 0 )
        return true;
    visited.insert( range );

    if (range < SIMPLEX_TOL || errors [best] < SIMPLEX_TOL)
    {
		//printf("Optimizer converged - range: %f worst: %f best: %f\n", range, errors[worst], errors[best]);
      return true;
    } 

    // try reflecting the simplex from the worst vert
    float mutate_error = mutateSimplex(-1.0f);
    
    // if reflection created new best, try additional extrapolation 
    if (mutate_error <= errors[best])
        mutate_error = mutateSimplex(2.0f);
    
    // if reflection worse than old next to worst point
    else if (mutate_error >= errors[nworst])
    {
        float save_error = errors[worst];
        mutate_error = mutateSimplex(0.5);

        if (mutate_error >= save_error)
        {
            
            for (int i = 0; i < ndim+1; i++)
            {
                if (i == best) continue;
                for (int j = 0; j < ndim; j++)
                    positions[i][j] = sum_positions[j] = 0.5f *
                        (positions[i][j]+positions[best][j]);
                errors[i] = funk(sum_positions, funk_params);
                float min = (errors[best] < errors[i]) ? errors[best] : errors[i];
                logEval(errors[i], min);

            }
            sumPositions();
        }
    }
    return false;
}

int 
OptimizerSimplex::solve(int max_iter)
{
  int max_infs = 10;
  for (int i = 0; i < max_iter; i++)
  {
    num_steps++;
    if (step()) 
    { 
#if 0
	    printf(" Converged after %d iterations with error : %f.\n", i+1, errors[best]);
      for (int j = 0; j < ndim+1; j++)
      {
	      printf(" pos[%i] = %f, %f: error = %f\n", j, positions[j][0], positions[j][1], errors[j]);
      }
	    printf(" best   : %i\n", best);
	    printf(" worst  : %i\n", worst);
	    printf(" nworst : %i\n\n", nworst);
#endif
      return i+1;
    }
    if (i == max_iter-1)
    {
#if 0
	    printf(" Failed to converge after %d iterations with error : %f.\n", i+1, errors[best]);
      for (int j = 0; j < ndim+1; j++)
      {
        printf(" pos[%i] = %f, %f: error = %f\n", j, positions[j][0], positions[j][1], errors[j]);
      }
      printf(" best   : %i\n", best);
      printf(" worst  : %i\n", worst);
      printf(" nworst : %i\n\n", nworst);
#endif
      return max_iter;
    }
    if (inf_count >= max_infs)
    {
#if 0
      printf(" Encountered too many infinite best errors (giving up)\n");
#endif
      return i+1;
    }
  }
    return -1;
}

void OptimizerSimplex::findBestWorst()
{
    best = 0;
    if (errors[1] > errors[2])
    {
        worst = 1;
        nworst = 2;
    }
    else
    {
        worst = 2;
        nworst = 1;
    }
    for (int i = 0; i < ndim+1; i++)
    {
        best = errors[i] <= errors[best] ? i : best;
        if (errors[i] > errors[worst])
        { 
            nworst = worst;
            worst  = i;
        }
        else if (errors[i] > errors[nworst] && i != worst) nworst = i;  
    }
    /*  
    for (int i = 0; i < ndim+1; i++)
    {
        printf(" pos[%i] = %f, %f: error = %f\n", i, positions[i][0],
                positions[i][1], errors[i]);
    }
    printf(" best   : %i\n", best);
    printf(" worst  : %i\n", worst);
    printf(" nworst : %i\n\n", nworst);
    */ 
}

float OptimizerSimplex::mutateSimplex(float factor)
{
    float mutate_error;
    float factor1, factor2;

    factor1 = (1.0f - factor) / (float)ndim;
    factor2 = factor1 - factor;
   
    // calculate new vertex position for worst vert 
    for (int j=0; j < ndim; j++) 
        mutate_position[j] = sum_positions[j]*factor1 - 
                             positions[worst][j]*factor2;
    mutate_error = funk(mutate_position, funk_params);

    float min = (mutate_error < errors[best]) ? mutate_error : errors[best];
    logEval(mutate_error, min);

    // check if we improved
    if (mutate_error < errors[worst]) 
    {
        errors[worst] = mutate_error;
        for (int j = 0; j < ndim; j++) 
        {
            sum_positions[j] += mutate_position[j] - positions[worst][j];
            positions[worst][j] = mutate_position[j];
        }
    }
    return mutate_error;
}



