#!/usr/bin/python2

from collections import defaultdict
import csv
import math
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from numpy import matrix
from optparse import OptionParser
import os
import re
import sys
import Tkinter as tk
import tkFileDialog
import tkMessageBox

parser = OptionParser( usage = "%prog [options] repair-log" )
parser.add_option(
    "--csv", metavar = "file",
    help = "save data to a CSV file for web visualization"
)
parser.add_option(
    "--final-frontier", action = "store_true",
    help = "do not draw generational pareto frontiers"
)
parser.add_option(
    "--point", metavar = "ifit,afit", action = "append", default = [],
    help = "highlight a selected point"
)
parser.add_option(
    "--pdf", metavar = "filename", help = "save figure as PDF"
)
parser.add_option(
    "--png", metavar = "filename", help = "save figure as PNG"
)
parser.add_option(
    "--svg", metavar = "filename", help = "save figure as SVG"
)
options, args = parser.parse_args()

if len( args ) < 1:
    parser.print_help()
    exit()
count = 0
for target in [ options.pdf, options.png, options.svg ]:
    if target is not None:
        count += 1
if count > 1:
    print >>sys.stderr, "ERROR: --pdf, --png, and --svg are mutually exclusive"
    parser.print_help()
    exit( 1 )

class Point:
    def __init__( self, *args ):
        self.x = list()
        for arg in args:
            try:
                v = float( arg )
                if v == 0:
                    self.x.append( float( "inf" ) )
                else:
                    self.x.append( 1 / v )
            except Exception as e:
                self.x.append( float( "nan" ) )
        self.x = tuple( self.x )

    def __cmp__( self, other ):
        return cmp( self.x, other.x )

    def __eq__( self, other ):
        if not isinstance( other, Point ):
            return False
        return self.x == other.x

    def __ne__( self, other ):
        return not self == other

    def __hash__( self ):
        return hash( self.x )

    def __iter__( self ):
        return iter( self.x )

    def __str__( self ):
        return str( map( lambda x: 1/x, self.x ) )

    def csv( self ):
        fields = list( self.x )
        return fields

    def csv_hdr( self ):
        if len( self.x ) <= 3:
            hdr = [ "x", "y", "z" ][ 0:len( self.x ) ]
        else:
            hdr = map( "x%d".__mod__, range( len( self.x ) ) )
        return hdr

    def dim( self ):
        return len( self.x )

    def distsq( self, other ):
        d = 0.0
        for a, b in zip( self.x, other.x ):
            d += abs( a - b ) * abs( a - b )
        return d

    def dominates( self, other ):
        better = False
        for a, b in zip( self.x, other.x ):
            if a < b:
                better = True
            elif a > b:
                return False
        return better

    def exists( self, predicate ):
        return any( map( predicate, self.x ) )

    def forall( self, predicate ):
        return all( map( predicate, self.x ) )

class KDTree:
    def __init__( self, points, depth = 0 ):
        dim = max( map( Point.dim, points ) )
        axis = depth % dim
        points.sort( key = lambda p: p.x[ axis ] )
        mid = len( points ) // 2
        while mid > 0 and points[ mid-1 ].x[ axis ] == points[ mid ].x[ axis ]:
            mid -= 1

        self.p = points[ mid ]
        self.axis = axis
        self.left = None
        self.right = None

        left = points[ :mid ]
        if len( left ) > 0:
            self.left = KDTree( left, depth + 1 )
        right = points[ mid+1: ]
        if len( right ) > 0:
            self.right = KDTree( right, depth + 1 )

    def nearest( self, pt ):
        dist = self.p.distsq( pt )
        d_to_plane = pt.x[ self.axis ] - self.p.x[ self.axis ]
        if d_to_plane < 0:
            a, b = self.left, self.right
        else:
            a, b = self.right, self.left

        if a is not None:
            candidate, distance = a.nearest( pt )
        if a is None or dist < distance:
            candidate, distance = self.p, dist
        if abs( d_to_plane ) < distance and b is not None:
            candidate2, distance2 = b.nearest( pt )
            if distance2 < distance:
                candidate, distance = candidate2, distance2
        return candidate, distance

def data_frame( points ):
    columns = dict()
    names = list()
    for p in points:
        if len( names ) == 0:
            names = p.csv_hdr()
            for name in names:
                columns[ name ] = list()
        vals = list( p )
        if len( vals ) != len( names ):
            continue
        for name, v in zip( names, vals ):
            columns[ name ].append( v )
    return columns

def nth( i ):
    return lambda x, i = i: x[ i ]

fst = nth( 0 )
snd = nth( 1 )

gen_pat = re.compile( r"multiopt: ngsa_ii generation (\d+) begins" )
fit_pat = re.compile(
    r"((\s*([-+]?(\d*\.)?\d+(e[-+]\d+)?|nan|[-+]?inf))+)\s*(.*)"
)

def points( fname ):
    current_gen = 0
    with open( fname ) as fh:
        for line in fh:
            m = gen_pat.match( line )
            if m:
                current_gen = int( m.group( 1 ) )
                continue

            m = fit_pat.match( line )
            if not m:
                continue

            fitness = m.group( 1 ).split()
            if all( map( lambda x: x == "-inf", fitness ) ):
                continue
            if any( map( lambda x: x == "nan", fitness ) ):
                continue

            p = Point( *fitness )
            yield current_gen, p, m.group( 6 )

original = None
generations = list()
ndim = 0
for gen, p, history in points( args[ 0 ] ):
    while len( generations ) < gen:
        generations.append( set() )
    ndim = max( ndim, p.dim() )
    generations[ gen - 1 ].add( p )
    if history.strip() == "original":
        original = p

for i in range( len( generations ) ):
    generations[ i ] = filter( lambda p: p.dim() == ndim, generations[ i ] )

frontiers = list()
frontier = set()
for i, gen in enumerate( generations ):
    for p in gen:
        new_frontier = set()
        for old in frontier:
            if old.dominates( p ):
                frontier = set( frontier )
                break
            elif not p.dominates( old ):
                new_frontier.add( old )
        else:
            new_frontier.add( p )
            frontier = new_frontier
    frontiers.append( frontier )

all_points = set()
for gen in generations:
    for p in gen:
        if p.exists( math.isnan ) or p.exists( math.isinf ):
            continue
        all_points.add( p )

kdtree = KDTree( list( all_points ) )

class MultioptPlot( tk.Frame ):
    def __init__( self, master = None ):
        tk.Frame.__init__( self, master )

        self.dpi = 72

        ########
        # create the menu bar
        ########

        menubar = tk.Menu( master )
        filemenu = tk.Menu( menubar, tearoff = 0 )
        filemenu.add_command(
            label = "Save...", accelerator = "Ctrl+S", command = self.save
        )
        self.bind_all( "<Control-s>", self.save )
        filemenu.add_command(
            label = "Quit", accelerator = "Ctrl+Q", command = self.quit
        )
        self.bind_all( "<Control-q>", self.quit )
        self.bind_all( "<Escape>", self.quit )
        menubar.add_cascade( label = "File", menu = filemenu )

        master.config( menu = menubar )

        ########
        # initialize the scatter plot
        ########

        self.fig = Figure( figsize = ( 10, 8 ), dpi = self.dpi )
        self.plot = self.fig.add_subplot( 1, 1, 1 )

        data = data_frame( all_points )
        self.plot.plot(
            data[ 'x' ], data[ 'y' ],
            c = "k", alpha = 0.5,
            linestyle = "None",
            marker = "o",
        )
        if original is not None:
            self.target( original.x[ 0 ], original.x[ 1 ], "r" )

        if options.final_frontier:
            self.draw_frontier( frontiers[ -1 ], "b" )
        else:
            n = float( len( frontiers ) )
            colors = cm.rainbow( map( lambda i: i/n, range( int( n ) ) ) )
            for frontier, c in zip( frontiers, reversed( colors ) ):
                self.draw_frontier( frontier, c )

        for i, pt in enumerate( options.point ):
            pt = map( float, pt.split( "," ) )
            pt = Point( *pt )
            self.target( pt.x[ 0 ], pt.x[ 1 ], "g", str( i + 1 ) )

        ax = self.fig.gca()
        ax.spines[ "right" ].set_color( "none" )
        ax.spines[ "top" ].set_color( "none" )
#        ax.spines[ "bottom" ].set_position( ("data", 0) )
#        ax.spines[ "left" ].set_position( ("data", 0) )

        self.plot.set_xlabel( "image error" )
        self.plot.set_ylabel( "aliasing" )

        xlim = self.plot.get_xlim()
        ylim = self.plot.get_ylim()
        dx = xlim[ 1 ] - xlim[ 0 ]
        dy = ylim[ 1 ] - ylim[ 0 ]
        self.plot.set_xlim( ( xlim[ 0 ], xlim[ 0 ] + max( dx, dy ) ) )
        self.plot.set_ylim( ( ylim[ 0 ], ylim[ 0 ] + max( dx, dy ) ) )

        ########
        # initialize the listbox
        ########

        sb = tk.Scrollbar( self )
        sb.pack( side = "right", fill = "y" )

        self.box = tk.Listbox( self )
        self.box.pack( side = "right", fill = "y" )
        for i in range( len( generations ) ):
            row = "generation %d" % ( i + 1 )
            self.box.insert( "end", row )

        self.selection = None
        self.box.bind( "<<ListboxSelect>>", self.onselect )

        self.box[ "yscrollcommand" ] = sb.set
        sb[ "command" ] = self.box.yview

        ########
        # add the plot to our widget
        ########

        canvas = FigureCanvasTkAgg( self.fig, master = self )
        canvas.show()
        canvas.get_tk_widget().pack( side = "left", fill = "both", expand = 1 )

        self.fig.tight_layout()

        ########
        # set up data structures for scrolling
        ########

        self.transform = None
        self.start     = None
        self.pressed   = False

        self.plot.figure.canvas.mpl_connect( "button_press_event", self.ondown )
        self.plot.figure.canvas.mpl_connect( "button_release_event", self.onup )
        self.plot.figure.canvas.mpl_connect( "motion_notify_event", self.onmove )
        self.plot.figure.canvas.mpl_connect( "scroll_event", self.onscroll )

    def target( self, x, y, c, label = None ):
        if label is None:
            self.plot.plot(
                x, y, ms = 10, c = c, linestyle = "None", marker = "o"
            )
        else:
            self.plot.annotate(
                s = label, xy = ( x, y ), size = "large",
                horizontalalignment = "center",
                verticalalignment = "center",
                color = 'w', backgroundcolor = c
            )

    def draw_frontier( self, frontier, c ):
        frontier = list( frontier )
        frontier.sort()
        data = data_frame( frontier )
        self.plot.plot(
            data[ "x" ], data[ "y" ],
            drawstyle = "steps-post",
            marker = "s",
            lw = 2, c = c
        )

    def save( self, *args ):
        name = tkFileDialog.asksaveasfilename( parent = self )
        if name == "":
            return
        try:
            self.fig.savefig( name, bbox_inches = "tight" )
        except ValueError as e:
            tkMessageBox.showerror( "Save file", e.message )

    def quit( self, *args ):
        tk.Frame.quit( self )

    def onselect( self, event ):
        sel = self.box.curselection()[ 0 ]
        if self.selection is not None:
            pts = self.plot.lines[ -1 ]
            self.plot.lines.remove( pts )
        if self.selection != sel:
            self.selection = sel

            i = int( self.box.get( sel ).split()[ 1 ] ) - 1
            data = data_frame( generations[ i ] )
            self.plot.plot(
                data[ 'x' ], data[ 'y' ],
                c = "g", alpha = 1,
                linestyle = "None",
                marker = 'o'
            )
            self.plot.figure.canvas.draw()
        elif self.selection is not None:
            self.selection = None
            self.box.selection_clear( 0, "end" )
            self.plot.figure.canvas.draw()

    def ondown( self, event ):
        if event.inaxes != self.plot.axes:
            return
        if event.button == 1:
            self.transform = self.plot.transData.inverted()
            self.start     = self.transform.transform( ( event.x, event.y ) )
            self.pressed   = True
        elif event.button == 3:
            p = Point( 1/event.xdata, 1/event.ydata )
            p = kdtree.nearest( p )[ 0 ]
            print p

    def onmove( self, event ):
        if not self.pressed:
            return
        current = self.transform.transform( ( event.x, event.y ) )
        xlim = self.plot.get_xlim()
        ylim = self.plot.get_ylim()

        dx, dy = current - self.start
        self.start = current
        self.plot.set_xlim( xlim - dx )
        self.plot.set_ylim( ylim - dy )
        self.plot.figure.canvas.draw()

    def onup( self, event ):
        if not self.pressed:
            return
        self.transform = None
        self.start     = None
        self.pressed   = False

    def onscroll( self, event ):
        if event.inaxes == self.plot.axes:
            factor = 1.1 if event.button == "down" else 0.9
            xlim = self.plot.get_xlim()
            self.plot.set_xlim( ( xlim - event.xdata ) * factor + event.xdata )
            if event.key != "control":
                ylim = self.plot.get_ylim()
                self.plot.set_ylim( ( ylim - event.ydata ) * factor + event.ydata )
            self.plot.figure.canvas.draw()

def plot_with_matplotlib():
    root = tk.Tk()
    root.wm_title( os.path.basename( args[ 0 ] ) )

    app = MultioptPlot( root )
    app.pack( expand = 1, fill = "both" )

    tk.mainloop()

if options.csv is not None:
    header = False
    with open( options.csv, 'w' ) as fh:
        writer = csv.writer( fh )
        for gen, p, _ in points( args[ 0 ] ):
            if p.dim() != ndim:
                continue
            if not header:
                writer.writerow( [ "generation" ] + p.csv_hdr() )
                header = True
            writer.writerow( map( str, [ gen ] + p.csv() ) )
else:
    plot_with_matplotlib()

