float4
heat( float lvl )
{
    float hue = 0;
    float sat = 1;
    float val = 1;
    if ( lvl < 0 )
        return (float4)( 0, 0, 0, 1 );
    else if ( 0 <= lvl && lvl < 0.25 ) {
        hue = 240;
        val = lvl * 4;
    } else if ( 0.25 <= lvl && lvl < 0.75 )
        hue = ( 0.75 - lvl ) * 2 * 240;
    else
        sat = ( 1 - lvl ) * 4;

    float chroma = sat * val;
    float h = hue / 60;
    float x = chroma * ( 1 - fabs( fmod( h, 2 ) - 1 ) );
    float m = val - chroma;
    switch ( (int) h ) {
        case 0: return (float4)( chroma + m, x + m, m, 1 );
        case 1: return (float4)( x + m, chroma + m, m, 1 );
        case 2: return (float4)( m, chroma + m, x + m, 1 );
        case 3: return (float4)( m, x + m, chroma + m, 1 );
        case 4: return (float4)( x + m, m, chroma + m, 1 );
        case 5: return (float4)( chroma + m, m, x + m, 1 );
        default: return (float4)( 0, 1, 1, 1 );
    }
}

void
gaussian_weight( local float * weight, float sigma, uint size )
{
    float sum = 0.0f;
    for ( uint i = 0; i < size; ++i ) {
        for ( uint j = 0; j < size; ++j ) {
            float x = ( (float) i - (float) size / 2 ) / size;
            float y = ( (float) i - (float) size / 2 ) / size;
            float w = exp( -( x*x + y*y ) / sigma );
            weight[ i * size + j ] = w;
            sum += w;
        }
    }
    mem_fence( CLK_LOCAL_MEM_FENCE );
    for ( uint i = 0; i < size; ++i )
        for ( uint j = 0; j < size; ++j )
            weight[ i * size + j ] /= sum;
}

float
luminance( float4 c )
{
    float lum = dot( c, (float4)( 0.2126, 0.7152, 0.0722, 0 ) );
    return pow( clamp( lum, 0.0f, 1.0f ) / 1.25f, 1.0f / 2.2f );
}

float
floatSSIM(
    global float4 * img1,
    global float4 * img2,
    uint w, uint h,
    uint x, uint y,
    uint size,
    float c1,
    float c2,
    local float * weight
) {
    float alpha = 1.0f;
    float beta = 1.0f;
    float gamma = 1.0f;

    float ux, uy, sigmasqx, sigmasqy, sigmaxy;
    ux = uy = sigmasqx = sigmasqy = sigmaxy = 0.0f;
    for ( uint i = 0; i < size; ++i ) {
        for ( uint j = 0; j < size; ++j ) {
            uint pos = ( y + i ) * w + ( x + j );
            ux += weight[ i * size + j ] * luminance( img1[ pos ] ) * 255;
            uy += weight[ i * size + j ] * luminance( img2[ pos ] ) * 255;
        }
    }
    for ( uint i = 0; i < size; ++i ) {
        for ( uint j = 0; j < size; ++j ) {
            uint pos = ( y + i ) * w + ( x + j );
            float dx = luminance( img1[ pos ] ) * 255 - ux;
            float dy = luminance( img2[ pos ] ) * 255 - uy;
            sigmasqx += weight[ i * size + j ] * dx * dx;
            sigmasqy += weight[ i * size + j ] * dy * dy;
            sigmaxy  += weight[ i * size + j ] * dx * dy;
        }
    }
    float l = ( 2.0f * ux * uy + c1 ) / ( ux*ux + uy*uy + c1 );
    float c = ( 2.0f * sqrt( sigmasqx ) * sqrt( sigmasqy ) + c2 ) / ( sigmasqx + sigmasqy + c2 );
    float s = ( sigmaxy + c2 ) / ( sqrt( sigmasqx ) * sqrt( sigmasqy ) + c2 );
    return pow( l, alpha ) * pow( c, alpha ) * pow( s, alpha );
}

kernel
void
heatmap( global float * delta, global float4 * img )
{
    uint i = get_global_id( 1 ) * get_global_size( 0 ) + get_global_id( 0 );
    img[ i ] = heat( delta[ i ] );
}

kernel
void
l2_diff(
    global float4 * img1,
    global float4 * img2,
    global float * delta
) {
    uint i = get_global_id( 1 ) * get_global_size( 0 ) + get_global_id( 0 );
    float3 c1 = img1[ i ].xyz;
    float3 c2 = img2[ i ].xyz;
    delta[ i ] = distance( c1, c2 ) / sqrt( 3.0f );
}

kernel
void
ssim_diff(
    global float4 * img1,
    global float4 * img2,
    global float * delta,
    local float * weight,
    uint winsize
) {
    uint x = get_global_id( 0 );
    uint y = get_global_id( 1 );
    uint w = get_global_size( 0 );
    uint h = get_global_size( 1 );

    if ( get_local_id( 0 ) == 0 && get_local_id( 1 ) == 0 )
        gaussian_weight( weight, 0.1, winsize );
    barrier( CLK_LOCAL_MEM_FENCE );

    if ( x + winsize < w && y + winsize < h ) {
        float val = floatSSIM(
            img1, img2,
            w, h,
            x, y,
            winsize,
            0.01 * 255 * 0.01 * 255,
            0.03 * 255 * 0.03 * 255,
            weight
        );
        delta[ y * w + x ] = ( 1.0f - val ) * 0.5;
    } else {
        delta[ y * w + x ] = 0;
    }
}

kernel
void
supersample(
    global float4 * src,
    global float4 * dst,
    uint window
) {
    uint x = get_global_id( 0 ) * window;
    uint y = get_global_id( 1 ) * window;
    uint raster_w = get_global_size( 0 ) * window;
    uint raster_h = get_global_size( 1 ) * window;

    float count = 0.0f;
    float4 accum = (float4) 0;
    for ( uint i = 0; i < window; ++i ) {
        for ( uint j = 0; j < window; ++j ) {
            if ( x + j < raster_w && y + i < raster_h ) {
                accum += src[  ( y + i ) * raster_w + x + j ];
                count += 1.0f;
            }
        }
    }
    dst[ get_global_id( 1 ) * get_global_size( 0 ) + get_global_id( 0 ) ] =
        accum / count;
}

kernel
void
upsample_img(
    global float4 * src,
    global float4 * dst,
    uint f
) {
    uint x = get_global_id( 0 );
    uint y = get_global_id( 1 );
    uint w = get_global_size( 0 );
    for ( uint i = 0; i < f; ++i )
        for ( uint j = 0; j < f; ++j )
            dst[ ( y * f + i ) * w * f + ( x * f + j ) ] = src[ y * w + x ];
}

kernel
void
flatten_image( global float4 * img )
{
    uint i = get_global_id( 1 ) * get_global_size( 0 ) + get_global_id( 0 );
    float4 c = clamp( img[ i ], (float4)(0, 0, 0, 1), (float4)(1, 1, 1, 1) );
    img[ i ] = floor( c * 255.0f ) / 255.0f;
}

kernel
void
make_complex( global float4 * img, global float2 * raster )
{
    uint i = get_global_id( 1 ) * get_global_size( 0 ) + get_global_id( 0 );
    raster[ i ] = (float2)( luminance( img[ i ] ), 0.0f );
}

kernel
void
dft_image(
    global float2 * raster,
    global float4 * img,
    float peak,
    uint width,
    uint height )
{
    uint x = get_global_id( 0 );
    uint y = get_global_id( 1 );
    uint w = get_global_size( 0 );
    uint h = get_global_size( 1 );

    uint w2 = width >> 1;
    uint h2 = height >> 1;

    if ( ( ( 0 <= y && y < h2 ) || ( h - h2 <= y && y < h ) ) &&
         ( ( 0 <= x && x < w2 ) || ( w - w2 <= x && x < w ) ) ) {
        float amp = log( length( raster[ y * w + x ] ) + 1.0f ) / peak;
        img[ ( ( y + h2 ) % height ) * width + ( ( x + w2 ) % width ) ] =
            (float4)( amp, amp, amp, 1 );
    }
}

float
dissimilarity(
    global float4 * img1,
    global float4 * img2,
    uint x1, uint y1,
    uint x2, uint y2,
    uint window
) {
    uint w = get_global_size( 0 );
    float dist = 0.0f;
    for ( uint i = 0; i < window; ++i ) {
        for ( uint j = 0; j < window; ++j ) {
            uint src = ( y1 + i ) * w + ( x1 + j );
            uint dst = ( y2 + i ) * w + ( x2 + j );
            float d = distance( img1[ src ].xyz, img2[ dst ].xyz );
            dist += d * d;
        }
    }
    return dist;
}

kernel
void
init_bidir_sim(
    global float * d_complete,
    global float * d_cohere,
    uint window
) {
    uint x = get_global_id( 0 );
    uint y = get_global_id( 1 );
    uint w = get_global_size( 0 );
    uint h = get_global_size( 1 );

    uint i = y * w + x;
    if ( x < w - window && y < h - window ) {
        d_complete[ i ] = MAXFLOAT;
        d_cohere[ i ]   = MAXFLOAT;
    } else {
        d_complete[ i ] = 0.0f;
        d_cohere[ i ]   = 0.0f;
    }
}

kernel
void
merge_bidir_sim(
    global float * d_complete,
    global float * d_cohere,
    uint window
) {
    uint i = get_global_id( 1 ) * get_global_size( 0 ) + get_global_id( 0 );
    d_complete[ i ] =
        ( d_complete[ i ] + d_cohere[ i ] ) / window / window / sqrt( 3.0f );
}

kernel
void
bidirectional_similarity(
    global float4 * img1,
    global float4 * img2,
    global float * d_complete,
    global float * d_cohere,
    uint i,
    uint j,
    uint window
) {
    uint x = get_global_id( 0 );
    uint y = get_global_id( 1 );
    uint w = get_global_size( 0 );
    uint h = get_global_size( 1 );

    uint index = y * w + x;

    if ( x < w - window && y < h - window ) {
        float d1 = dissimilarity( img1, img2, x, y, j, i, window );
        d_complete[ index ] = min( d_complete[ index ], d1 );
        float d2 = dissimilarity( img2, img1, x, y, j, i, window );
        d_cohere[ index ] = min( d_cohere[ index ], d2 );
    }
}

