#!/usr/bin/python2

from collections import defaultdict
from optparse import OptionParser
import os
from subprocess import call, check_call, Popen, PIPE, STDOUT
import sys

from utils import *

parser = OptionParser( usage = "%prog [options] outdir size shader source..." )
parser.add_option(
    "--ortho", metavar = "z1,z2",
    help = "comma-separated list of orthographic scales to render"
)
parser.add_option(
    "--persp", metavar = "r1,r2",
    help = "comma-separated list of perspective rotations to render"
)
parser.add_option(
    "--crop", metavar = "WxH",
    help = "crop images after rendering, before assembling"
)
parser.add_option(
    "--diff", metavar = "alg", help = "compute different heatmaps"
)
options, args = parser.parse_args()

if len( args ) < 4:
    parser.print_help()
    exit()

outdir  = args[ 0 ]
size    = args[ 1 ]
shader  = args[ 2 ]
sources = args[ 3: ]

if options.ortho is None:
    options.ortho = list()
else:
    options.ortho = map( float, options.ortho.split( "," ) )
if options.persp is None:
    options.persp = list()
else:
    options.persp = map( float, options.persp.split( "," ) )

root = os.path.abspath( os.path.dirname( sys.argv[ 0 ] ) )

compilesh = os.path.join( root, "compile.sh" )
testpy    = os.path.join( root, "test.py" )
renderer  = os.path.join( root, "renderer" )

check_executable( [ compilesh ] )
check_executable( [ testpy, "--help" ] )

def get_original( shader ):
    return os.path.join( root, "../src", shader + ".c" )

compiled = dict()
def build( ctx, shader, source, filtered ):
    global compiled

    if not ( source, filtered ) in compiled:
        with mktemp( ".c" ) as ppfile:
            cmd = [ "gcc", "-E", source, "-o", ppfile ]
            if filtered:
                cmd.append( "-DFILTER_MODE_GAUSS" )
            check_call( cmd )

            exe = ctx.append( mktemp( "" ) )
            check_call( [ compilesh, shader, exe, ppfile ] )
            compiled[ ( source, filtered ) ] = exe
    return compiled[ ( source, filtered ) ]

def get_render_flags( ctx, persp, zooms, do_diff = False ):
    flags = list()
    ppms  = list()
    for zoom in zooms:
        if len( flags ) > 0:
            flags.append( "-i" )
        ppm = ctx.append( mktemp( ".ppm" ) )
        ppms.append( ppm )
        flags += [ "-o", ppm, "--zoom", "%g" % zoom ]
        if do_diff:
            flags += [ "-g", get_golden( shader, size, zoom, False, persp ) ]
    if persp:
        flags.append( "--perspective" )
    if do_diff:
        flags += [ "--diff", options.diff ]
    return flags, ppms

modes = [
    ( False, options.ortho ),
    ( True,  options.persp ),
]

if not os.path.exists( outdir ):
    os.makedirs( outdir )

with ContextList() as ctx:
    # check that we have all the golden images...

    for persp, zooms in modes:
        tests = list()
        for zoom in zooms:
            golden = get_golden( shader, size, zoom, False, persp )
            if not os.path.exists( golden ):
                tests.append( zoom )
        if len( tests ) == 0:
            continue
        print >>sys.stderr, "generating golden images"
        exe = build( ctx, shader, get_original( shader ), False )
        cmd = [
            testpy, shader, exe, "/dev/null", size,
                "--imagew", "1",
                "--zoom", ",".join( map( "%g".__mod__, tests ) )
        ]
        if persp:
            cmd.append( "--perspective" )
        with open( "/dev/null", 'w' ) as fh:
            check_call( cmd, stdout = fh )

    # collect images for the requested variants

    matrix = defaultdict( list )

    for source in sources:
        if source == "golden":
            for persp, zooms in modes:
                for zoom in zooms:
                    matrix[ source ].append(
                        get_golden( shader, size, zoom, False, persp )
                    )

        elif source == "msaa":
            print >>sys.stderr, "rendering MSAA for", shader
            for persp, zooms in modes:
                if len( zooms ) == 0:
                    continue
                flags, ppms = get_render_flags( ctx, persp, zooms )
                matrix[ source ].extend( ppms )
                cmd = [
                    renderer,
                        build( ctx, shader, get_original( shader ), False ),
                        "--sampling", "gaussian",
                        "--size", size,
                        "--msaa", "16"
                ]
                check_call( cmd + flags )
                if options.diff is not None:
                    flags, ppms = get_render_flags( ctx, persp, zooms, True )
                    for i in range( -len( ppms ), 0 ):
                        matrix[ source ][ i ] = \
                            ( matrix[ source ][ i ], ppms[ i ] )
                    check_call( cmd + flags )

        else:
            print >>sys.stderr, "rendering", source
            for persp, zooms in modes:
                if len( zooms ) == 0:
                    continue
                flags, ppms = get_render_flags( ctx, persp, zooms )
                matrix[ source ].extend( ppms )

                cmd = [ renderer ]
                if os.path.splitext( source )[ 1 ] == ".cl":
                    cmd += [ source ]
                else:
                    cmd += [ build( ctx, shader, source, True ) ]
                cmd += [ "--size", size ]
                check_call( cmd + flags )
                if options.diff is not None:
                    flags, ppms = get_render_flags( ctx, persp, zooms, True )
                    for i in range( -len( ppms ), 0 ):
                        matrix[ source ][ i ] = \
                            ( matrix[ source ][ i ], ppms[ i ] )
                    check_call( cmd + flags )

    if options.crop is not None:
        def crop( ctx, fname ):
            ppm = ctx.append( mktemp( ".ppm" ) )
            cmd = [
                "convert", fname,
                    "-crop", options.crop + "+0+0", "+repage",
                    ppm
            ]
            check_call( cmd )
            return ppm

        for key in matrix:
            for i in range( len( matrix[ key ] ) ):
                if isinstance( matrix[ key ][ i ], tuple ):
                    crops = map( lambda f: crop( ctx, f ), matrix[ key ][ i ] )
                    matrix[ key ][ i ] = tuple( crops )
                else:
                    matrix[ key ][ i ] = crop( ctx, matrix[ key ][ i ] )

    print >>sys.stderr, "preprocessing cells for", shader
    for key in matrix:
        for i in range( len( matrix[ key ] ) ):
            png = ctx.append( mktemp( ".png" ) )
#             check_call( [
#                 "convert", diffimg, "-resize", "25%",
#                     image, "-reverse",
#                     "-flip", "-flop", "-mosaic", "-flop", "-flip",
#                     tmpimage
#             ] ) 
            cmd = [ "convert" ]
            if isinstance( matrix[ key ][ i ], tuple ):
                cmd += [ matrix[ key ][ i ][ 1 ],
                    "-resize", "33%",
                    matrix[ key ][ i ][ 0 ],
                    "-reverse",
                    "-flip", "-flop", "-mosaic", "-flop", "-flip",
                ]
            else:
                cmd += [ matrix[ key ][ i ] ]
            cmd += [
                "-set", "colorspace", "RGB",
                "-bordercolor", "white",
                "-border", "5",
                png
            ]
            check_call( cmd )
            matrix[ key ][ i ] = png

    print >>sys.stderr, "building montage for", shader
    cmd = [ "montage",
        "-tile", "%dx%d" % ( len( matrix ), len( matrix[ sources[ 0 ] ] ) ),
        "-mode", "Concatenate"
    ]
    for i in range( len( matrix[ sources[ 0 ] ] ) ):
        for source in sources:
            cmd.append( matrix[ source ][ i ] )
    cmd.append( os.path.join( outdir, shader + ".png" ) )
    check_call( cmd )

