module Make : functor (Ord : Map.OrderedType) -> 
sig
  (** The type of members of the list to use as keys. *)
  type key

  (** The type of tries using lists of type [key] to type ['a]. *)
  type 'a t

  (** The empty trie. *)
  val empty : 'a t

  (** [mem x t] returns {i true} if [t] contains a value for [x], and {i false}
      otherwise. *)
  val mem : key list -> 'a t -> bool

  (** [find x t] returns the value associated with [x] in [t] or raises
      {!Not_found} if no such association exists. *)
  val find : key list -> 'a t -> 'a

  (** [find_prefix x t] returns the list of key-value pairs from [t] where the
      key has [x] as a prefix. *)
  val find_prefix : key list -> 'a t -> (key list * 'a) list
      
  (** [add x y t] returns a trie containing the same associations as [t], plus
      the association of [x] with [y]. If [x] is already associated with a
      value, the previous association is lost. *)
  val add : key list -> 'a -> 'a t -> 'a t

  (** [remove x t] returns a trie containing the same associations as [t],
      except that [x] is not associated with a value in the returned trie. *)
  val remove : key list -> 'a t -> 'a t

  (** [singleton x y] returns a one-element trie that associates [x] with [y]. *)
  val singleton : key list -> 'a -> 'a t

  (** [fold f t a] computes [(f kN dN ... (f k1 d1 a)...)], where [k1 ... kN]
      are the keys with associated data and [d1 ... dN] are the associated
      data. *)
  val fold : (key list -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b

  (** [iter f t] applies [f] to all bindings in trie [t]. [f] receives the key
      as first argument and the associated data as second argument. *)
  val iter : (key list -> 'a -> unit) -> 'a t -> unit

  (** [size t] returns the number of values stored in [t]. *)
  val size : 'a t -> int
end 
with type key = Ord.t ;;
