open Cil

module ISet = Set.Make(struct type t = int let compare = compare end)
module SMap = Map.Make(String)
module SSet = Set.Make(String)

(** Enables debugging messages *)
let debug = ref false

(** Prints a debugging message to standard error if [ !debug] is true. Uses
    Cil.Pretty.printf style format strings. *)
let debugf fmt =
  let display d =
    if !debug then
      Printf.eprintf "DEBUG: %s\n" (Pretty.sprint ~width:80 d)
  in
  Pretty.gprintf display fmt

(** Constructs a Failure exception with the given format. Uses Cil.Pretty.printf
    style format strings. *)
let failure fmt =
  let display d =
    failwith (Pretty.sprint ~width:80 d)
  in
  Pretty.gprintf display fmt

(**
 * [filter_map f lst] applies [f] to every element of [lst], constructing a new
 * list of the non-[None] results.
 *)
let filter_map f lst =
  List.rev (List.fold_left (fun vs l ->
    match f l with
    | Some(v) -> v::vs
    | None    -> vs
  ) [] lst)

(**
 * Parses the named file and returns the resulting {!Cil.file} object. If there
 * were any parsing errors, raises {!Failure} with some (hopefully) human
 * readable information indicating the error.
 *)
let parse fname =
  Errormsg.hadErrors := false;
  let cilfile =
    try
      Frontc.parse fname ()
    with Parsing.Parse_error ->
      failure "parse error: %s" fname
  in
  if !Errormsg.hadErrors then
    failure "parse error: %s" fname;
  cilfile

(**
 * Returns an lval representing the named field of the given lval. For example
 * if [lv] represents [x] and [fname] represents [foo], then [addField lv fname]
 * returns an lval represening [x.foo].
 *)
let addField lv fname =
  match unrollType (typeOfLval lv) with
  | TComp(ci,_) ->
    let fi =
      try
        getCompField ci fname
      with Not_found ->
        failwith (compFullName ci ^ " has no field name " ^ fname)
    in
      addOffsetLval (Field(fi, NoOffset)) lv
  | t ->
    failure "%a (%a) is not a struct" dn_lval lv dn_type t

(**
 * [match_attr name param attr] matches attributes that have the given name and
 * contain a string or constructed parameter with the given param name.
 *)
let match_attr name param (Attr(n, params)) =
  (n = name) && (List.fold_left (fun matched p ->
    match p with
    | AStr(n') | ACons(n', _) when n' = param -> true
    | _ -> matched
  ) false params)

(**
 * Removes attributes matching the predicate from the variable's attribute list
 * and returns the removed attributes.
 *)
let extract_variable_attributes predicate vi =
  let extracted, attrs = List.partition predicate vi.vattr in
  vi.vattr <- attrs;
  extracted

(**
 * Removes attributes matching the predicate from the function's attribute list
 * and returns the removed attributes.
 *)
let extract_function_attributes predicate fd =
  let extracted, attrs = List.partition predicate fd.svar.vattr in
  fd.svar.vattr <- attrs;
  match fd.svar.vtype with
  | TFun(ret, args, variadic, attrs) ->
    let extracted2, attrs = List.partition predicate attrs in
    fd.svar.vtype <- TFun(ret, args, variadic, attrs);
    extracted @ extracted2
  | _ -> extracted

(**
 * Filter that accepts function declarations or definitions that have the
 * attribute [__attribute__(opencl(keep))] and rejects all other globals. The
 * [opencl] attribute is removed by this filter, so it only accepts the first
 * such declaration or definition it encounters.
 *)
let retain_marked_functions g =
  let check_function vi =
    let keepers = extract_variable_attributes (match_attr "opencl" "keep") vi in
    (List.length keepers) > 0
  in
  match g with
  | GVarDecl(vi,_) when isFunctionType vi.vtype -> check_function vi
  | GFun(fd,_) -> check_function fd.svar
  | _ -> false

(**
 * Matches globals that appear in api3d.h and are not marked with
 * [__attribute__(opencl(keep))].
 *)
let filter_api3d () =
  let apifile =
    Filename.concat (Filename.dirname Sys.argv.(0)) "../src/api3d.h"
  in
  let api = Frontc.parse apifile () in

  let cache = Hashtbl.create 127 in
  let retained = Hashtbl.create 127 in

  let key_of g =
    match g with
    | GType(ti,_)        -> Some("GType", ti.tname)
    | GCompTag(ci,_)     -> Some("GComp", ci.cname)
    | GCompTagDecl(ci,_) -> Some("GComp", ci.cname)
    | GEnumTag(ei,_)     -> Some("GEnum", ei.ename)
    | GEnumTagDecl(ei,_) -> Some("GEnum", ei.ename)
    | GVarDecl(vi,_)     -> Some("GVar", vi.vname)
    | GVar(vi,_,_)       -> Some("GVar", vi.vname)
    | GFun(fd,_)         -> Some("GVar", fd.svar.vname)
    | _ -> None
  in
  iterGlobals api (fun g ->
    match retain_marked_functions g, key_of g with
    | true, Some(key) -> Hashtbl.replace retained key ()
    | _ -> ()
  );
  iterGlobals api (fun g ->
    match key_of g with
    | Some(key) when not (Hashtbl.mem retained key) ->
      Hashtbl.replace cache key ()
    | _ -> ()
  );
  fun g ->
    match key_of g with
    | Some(key) -> Hashtbl.mem cache key
    | _ -> false
  
(** Matches globals that represent builtin functions. *)
let filter_builtin g =
  match g with
  | GVarDecl(vi,_) -> Hashtbl.mem builtinFunctions vi.vname
  | GFun(fd,_) -> Hashtbl.mem builtinFunctions fd.svar.vname
  | _ -> false

(**
 * Visitor to remove [__global__] markers from non-pointer types. This is
 * useful when transformations infer types from pointers that are thereby
 * incorrectly tainted with __global__.
 *)
class globalIsForPointers =
object
  inherit nopCilVisitor

  method vtype t =
    let filt a = not (match_attr "opencl" "global" a) in
    let clean attrs = List.filter filt attrs in
    let return t' = ChangeDoChildrenPost(t', fun t -> t) in
    match t with
    | TVoid(attrs)             -> return (TVoid(clean attrs))
    | TInt(ik, attrs)          -> return (TInt(ik, clean attrs))
    | TFloat(fk, attrs)        -> return (TFloat(fk, clean attrs))
    | TArray(t', s, attrs)     -> return (TArray(t', s, clean attrs))
    | TFun(r, args, va, attrs) -> return (TFun(r, args, va, clean attrs))
    | TNamed(ti, attrs)        -> return (TNamed(ti, clean attrs))
    | TComp(ci, attrs)         -> return (TComp(ci, clean attrs))
    | TEnum(ei, attrs)         -> return (TEnum(ei, clean attrs))
    | TBuiltin_va_list(attrs)  -> return (TBuiltin_va_list(clean attrs))
    | TPtr(_) -> SkipChildren
end

