open Cil
open Pretty
open Utils

(**
 * Implements a dataflow analysis to compute width expressions for each node
 * in the syntax tree.
 *)

module IntTrie = Trie.Make(struct type t = int let compare = compare end)
module DimSetMap = Map.Make(struct
  type t = int list SMap.t
  let compare = SMap.compare compare
end)

(** Simplified expression tree for width expressions. *)
type widthexp =
| DVector of widthexp SMap.t * typ
| DLinExp of float DimSetMap.t * typ

exception HigherOrderTerm of string * int list SMap.t * int list SMap.t

(**
 * Lookup table to map vids back to their varinfos so we can convert keys back
 * to lvals.
 *)
let varinfos = Hashtbl.create 257

(** Lookup table to map function names to the varinfos for those functions. *)
let functions = Hashtbl.create 257

(** Returns a canonicalized key corresponding to the given lval. *)
let rec key_of_lval lv =
  let rec field_indices accum off =
    match off with
    | NoOffset -> accum
    | Field(fi, off') ->
      let index, _ =
        List.fold_left (fun (index,i) field ->
          if (index = None) && (field.fname = fi.fname) then
            Some(i), i+1
          else
            index, i+1
        ) (None,0) fi.fcomp.cfields
      in
      begin match index with
      | Some(index) -> field_indices (index::accum) off'
      | None ->
        let name = compFullName fi.fcomp in
        failure "could not find field %s in %s" fi.fname name
      end
    | Index(_) -> failwith "array indexing not supported yet"
  in
  match lv with
  | Var(vi), off ->
    Hashtbl.replace varinfos vi.vid vi;
    vi.vid :: List.rev (field_indices [] off)
  | Mem(Lval(lv')), off ->
    let prefix = key_of_lval lv' in
    List.rev_append (List.rev prefix) ((-1) :: List.rev (field_indices [] off))
  | Mem(_), _ ->
    failure "memory references %a not supported yet" dn_lval lv

(** Returns the lval corresponding to the given key. *)
let lval_of_key key =
  let vi = Hashtbl.find varinfos (List.hd key) in 
  let rec helper lv indices =
    match indices with
    | i::is ->
      if i < 0 then
        helper (Mem(Lval(lv)), NoOffset) is
      else
        let fname = 
          match unrollType (typeOfLval lv) with
          | TComp(ci,_) -> (List.nth ci.cfields i).fname
          | _ -> failure "%a is not a struct" dn_lval lv
        in      
        helper (addField lv fname) is
    | [] -> lv
  in
  helper (var vi) (List.tl key)

(** Returns the dimensions in which a width expression operates *)
let dims_of_widthexp we =
  let rec helper accum we =
    match we with
    | DVector(fields, _) ->
      SMap.fold (fun _ we' accum -> helper accum we') fields accum
    | DLinExp(terms, _) ->
      DimSetMap.fold (fun ds _ set ->
        SMap.fold (fun d _ set -> SSet.add d set) ds set
      ) terms accum
  in
    helper SSet.empty we

let type_of_widthexp = function
  | DVector(_, t)      -> t
  | DLinExp(_, t)      -> t

let floatType = TFloat(FFloat, [])

let mk_const_widthexp f = DLinExp(DimSetMap.singleton SMap.empty f, floatType)

let is_const_widthexp we = (SSet.cardinal (dims_of_widthexp we)) = 0

let nowidth = mk_const_widthexp 0.0

let mk_DLeaf name key =
  DLinExp(DimSetMap.singleton (SMap.singleton name key) 1.0, floatType)

let rec mk_DBinOp op we1 we2 =
  match we1, we2 with
  | DVector(map1, t1), DVector(map2, t2) ->
    if (typeSig t1) <> (typeSig t2) then
      failure "%t: cannot apply '%a' to types %a and %a"
        d_thisloc d_binop op dn_type t1 dn_type t2;
    let keys = SMap.fold (fun n _ ns -> SSet.add n ns) map1 SSet.empty in
    let keys = SMap.fold (fun n _ ns -> SSet.add n ns) map2 keys in
    let find_default default n map =
      if SMap.mem n map then SMap.find n map else default
    in
    let map =
      SSet.fold (fun n map ->
        let we1' = find_default nowidth n map1 in
        let we2' = find_default nowidth n map2 in
        SMap.add n (mk_DBinOp op we1' we2') map
      ) keys SMap.empty
    in
    DVector(map, t1)
  | DVector(map, t), _ ->
    let map =
      SMap.fold (fun n we map ->
        SMap.add n (mk_DBinOp op we we2) map
      ) map SMap.empty
    in
    DVector(map, t)
  | _, DVector(map, t) ->
    let map =
      SMap.fold (fun n we map ->
        SMap.add n (mk_DBinOp op we1 we) map
      ) map SMap.empty
    in
    DVector(map, t)

  | DLinExp(ts1, t1), DLinExp(ts2, t2) ->
    let terms name apply =
      try
        DimSetMap.fold (fun ds1 c1 ts ->
          DimSetMap.fold (fun ds2 c2 ts ->
            let overlap =
              SMap.fold (fun d _ n -> if SMap.mem d ds2 then n+1 else n) ds1 0
            in
            if overlap > 0 then 
              raise (HigherOrderTerm(name, ds1, ds2));
            let ds = SMap.fold SMap.add ds1 ds2 in
            let c = if DimSetMap.mem ds ts then DimSetMap.find ds ts else 0.0 in
            DimSetMap.add ds (c +. (apply c1 c2)) ts
          ) ts2 ts
        ) ts1 DimSetMap.empty
      with HigherOrderTerm(name, ds1, ds2) ->
        let keys m = List.rev (SMap.fold (fun k _ ks -> k::ks) m []) in
        eprintf "%t: attempt to %s %s by %s\n" d_thisloc name
          (String.concat "*" (keys ds1)) (String.concat "*" (keys ds2));
        DimSetMap.fold (fun ds1 c1 ts ->
          let c = try DimSetMap.find ds1 ts with Not_found -> c1 *. 2.0 in
          DimSetMap.add ds1 (c *. 0.5) ts
        ) ts1 ts2
    in
    let ts =
      match op with
      | PlusA | MinusA ->
        DimSetMap.fold (fun ds c ts ->
          let c' =
            try
              DimSetMap.find ds ts
            with Not_found -> 0.0
          in
          DimSetMap.add ds (c +. c') ts
        ) ts1 ts2
      | Mult -> terms "multiply" ( *. )
      | Div  -> terms "divide" ( /. )
      | _ -> failure "cannot apply %a to width expressions" d_binop op
    in
    DLinExp(ts, t1)

(** Returns a CIL expression to compute the width expression. *)
let exp_of_widthexp ?(noconst=true) we = 
  match we with 
  | DVector(_) -> failwith "width expressions from DVector not implemented"
  | DLinExp(terms, t) ->
    let e =
      DimSetMap.fold (fun ds c e ->
        if noconst && SMap.is_empty ds then e
        else
          let e' =
            SMap.fold (fun _ key e' ->
              BinOp(Mult, e', Lval(lval_of_key key), t)
            ) ds (Const(CReal(c, FFloat, None)))
          in
          BinOp(PlusA, e, e', t)
      ) terms zero
    in
    visitCilExpr (object
      inherit nopCilVisitor

      method vexpr e =
        ChangeDoChildrenPost(e, fun e ->
          match e with
          | BinOp(PlusA, Const(CInt64(i,_,_)), e', _)
              when (Int64.compare i Int64.zero) = 0 -> e'
          | BinOp(Mult, Const(CReal(f,_,_)), e', _) when f = 1.0 -> e'
          | _ -> e
        )
    end) e

let rec doc_of_widthexp () we =
  match we with
  | DVector(map, _) ->
    let pairs, width =
      SMap.fold (fun n we' (pairs, width) ->
        (n, doc_of_widthexp () we') :: pairs, max width (String.length n)
      ) map ([], 0)
    in
    List.fold_left (fun doc (n, d) ->
      doc ++ text n ++ text ": " ++ indent (width - (String.length n)) d ++ line
    ) nil (List.rev pairs)
  | _ ->
    try
      dn_exp () (exp_of_widthexp ~noconst:false we)
    with Failure msg ->
      text ("error(" ^ msg ^ ")")

let string_of_widthexp we = Pretty.sprint ~width:80 (doc_of_widthexp () we)

(**
 * [widthexp_of_exp lookup e] returns a width expression corresponding to the
 * CIL expression [e]. The [lookup] trie allows retrieval of the width
 * expressions for lvals (when known).
 *)
let rec widthexp_of_exp lookup e =
  match e with
  | Const(CInt64(i,_,_)) -> mk_const_widthexp (Int64.to_float i)
  | Const(CReal(f,_,_))  -> mk_const_widthexp f
  | Const(CStr(_))       -> nowidth
  | Const(CWStr(_))      -> nowidth
  | Const(CChr(c))       -> widthexp_of_exp lookup (Const(charConstToInt c))
  | Const(CEnum(_))      -> widthexp_of_exp lookup (constFold true e) 
  | SizeOf(_)            -> widthexp_of_exp lookup (constFold true e) 
  | SizeOfE(_)           -> widthexp_of_exp lookup (constFold true e) 
  | SizeOfStr(_)         -> widthexp_of_exp lookup (constFold true e) 
  | AlignOf(_)           -> widthexp_of_exp lookup (constFold true e) 
  | AlignOfE(_)          -> widthexp_of_exp lookup (constFold true e) 
  | AddrOf(_)            -> nowidth
  | StartOf(_)           -> nowidth
  | CastE(_,e')          -> widthexp_of_exp lookup e' (* FIXME: handle truncating casts *)
  | UnOp(_,e',_)         -> widthexp_of_exp lookup e'
  | BinOp(op,e1,e2,t)    ->    
    mk_DBinOp op (widthexp_of_exp lookup e1) (widthexp_of_exp lookup e2)
  | Lval(lv) ->
    let rec helper lv =
      try
        IntTrie.find (key_of_lval lv) lookup
      with Not_found ->
        match unrollType (typeOfLval lv) with
        | TComp(ci,attrs) ->
          let map =
            List.fold_left (fun map fi ->
              SMap.add fi.fname (helper (addField lv fi.fname)) map
            ) SMap.empty ci.cfields
          in
          DVector(map, (TComp(ci,attrs)))
        | _ -> nowidth
    in
    helper lv

(**
 * [widthexp_of_function_call lookup name args] returns a width expression
 * representing the result of the named function called with the given
 * arguments. The [lookup] trie allows retrieval of the width expressions for
 * lvals (when known). *)
let widthexp_of_function_call lookup name args =
  let wargs = List.map (widthexp_of_exp lookup) args in
  let return_type () =
    let rt, _, _, _ = splitFunctionTypeVI (Hashtbl.find functions name) in
    rt
  in
  let unknown wargs =
    let _ =
      eprintf "Warning: do not know how to transform width expression for %s\n"
        name
    in
    let n, we =
      List.fold_left (fun (n, we1) we2 ->
        if is_const_widthexp we2 then
          n, we1
        else
          n+1, mk_DBinOp PlusA we1 we2
      ) (0,nowidth) wargs
    in
    if n < 2 then
      we
    else
      mk_DBinOp Div we (mk_const_widthexp (float_of_int n))
  in
  let assign_to_fields xs =
    let map =
      List.fold_left (fun map (n,we) -> SMap.add n we map) SMap.empty xs
    in
    DVector(map, return_type ())
  in
  let do_binop op =
    match wargs with
    | [we1; we2] -> mk_DBinOp op we1 we2
    | _ ->
      failure "%t: unexpected width expression arguments to %s (%a) %d"
        d_thisloc name (d_list " | " doc_of_widthexp) wargs (List.length wargs)
  in
  let collapse = function
    | DVector(fields, _) ->
      let n, we =
        SMap.fold (fun _ we2 (n, we1) ->
          if is_const_widthexp we2 then
            n, we1
          else
            n+1, mk_DBinOp PlusA we1 we2
        ) fields (0, nowidth)
      in
        if n < 2 then
          we
        else
          mk_DBinOp Div we (mk_const_widthexp (float_of_int n))
    | we -> we
  in
  let swizzle mapping =
    match List.hd wargs with
    | DVector(map, _) ->
      assign_to_fields (List.fold_left (fun fields (src,dst) ->
        if SMap.mem src map then
          (dst, SMap.find src map) :: fields
        else
          fields
      ) [] mapping)
    | we ->
      failure "%t: attempting to swizzle non-DVector (%s)!" d_thisloc (string_of_widthexp we)
      (*
      eprintf "attempting to swizzle non-DVector (%s)!\n" (string_of_widthexp we);
      unknown wargs
      *)
  in
  match name with
  | "add_f2f"  -> do_binop PlusA
  | "add_f2f2" -> do_binop PlusA
  | "add_f3f"  -> do_binop PlusA
  | "add_f3f3" -> do_binop PlusA
  | "add_f4f"  -> do_binop PlusA
  | "add_f4f4" -> do_binop PlusA
  | "mul_f2f2" -> do_binop Mult
  | "mul_f3f"  -> do_binop Mult
  | "mul_f3f3" -> do_binop Mult
  | "mul_f4f"  -> do_binop Mult
  | "mul_f4f4" -> do_binop Mult
  | "sub_f2f2" -> do_binop PlusA
  | "sub_f3f3" -> do_binop PlusA
  | "sub_f4f4" -> do_binop PlusA

  | "div_f2f"
  | "div_f2f2"
  | "div_f3f"
  | "div_f3f3"
  | "div_f4f"
  | "div_f4f4" ->
    if (SSet.cardinal (dims_of_widthexp (List.nth wargs 1))) > 0 then begin
      debugf "Warning division by varying expressions not implemented";
      unknown wargs
    end else
      do_binop Div
  | "dot_f3" -> collapse (do_binop Mult)

  | "length_f2" ->
    begin match List.hd wargs with
    | DVector(map, _) -> unknown (SMap.fold (fun _ we wes -> we::wes) map [])
    | _ -> unknown wargs
    end
  | "lerp_ff"
  | "lerp_f4f" ->
    begin match wargs with
    | [da; db; dx] ->
      begin try
        let dy = mk_DBinOp MinusA (mk_const_widthexp 1.0) dx in
        let lhs = mk_DBinOp Mult da dx in
        let rhs = mk_DBinOp Mult db dy in
        mk_DBinOp PlusA lhs rhs
      with Failure msg ->
        ignore (eprintf "Warning: %t: %s\n" d_thisloc msg);
        unknown wargs
      end
    | _ -> unknown wargs
    end
  | "makef2_ff" ->
    assign_to_fields (List.map2 (fun n we -> n, we) ["x"; "y"] wargs)
  | "makef3_f" ->
    let we = List.hd wargs in
    assign_to_fields [("x", we); ("y", we); ("z", we)]
  | "makef3_fff" ->
    assign_to_fields (List.map2 (fun n we -> n, we) ["x"; "y"; "z"] wargs)
  | "makef4_f" ->
    let we = List.hd wargs in
    assign_to_fields [("x", we); ("y", we); ("z", we); ("w", we)]
  | "makef4_f3f" ->
    begin match List.hd wargs with
    | DVector(map, _) ->
      assign_to_fields (SMap.fold (fun n we fields ->
        (n, we) :: fields
      ) map ["w", List.nth wargs 1])
    | _ -> unknown wargs
    end
  | "makef4_ffff" ->
    assign_to_fields (List.map2 (fun n we -> n, we) ["x"; "y"; "z"; "w"] wargs)

  | name when startsWith "samplef" name ->
    if name.[7] == '_' then
      unknown wargs
    else
      let we = unknown wargs in
      let ndims = int_of_string (String.sub name 7 1) in
      let fields, _ =
        List.fold_left (fun (fields, ndims) name ->
          if ndims = 0 then
            fields, ndims
          else
            (name, we) :: fields, ndims - 1
        ) ([], ndims) ["x"; "y"; "z"; "w"]
      in
        assign_to_fields fields
      
  | name when startsWith "swizzle_f" name ->
    let mapping, _ =
      List.fold_left (fun (mapping, i) slot ->
        if (String.length name) <= i then
          mapping, i
        else
          (((String.make 1 name.[i]), slot) :: mapping), i+1
      ) ([], 10) ["x"; "y"; "z"; "w"]
    in
      swizzle mapping
  | _ -> unknown wargs

module DimensionTransfer =
struct
  let name = "Dimensional Analysis"

  let debug = ref false

  (* trie from lval-keys to width expressions *)
  type t = widthexp IntTrie.t

  let copy data = data

  let stmtStartData = Inthash.create 257

  let initStartData sid formals =
    let get_attr_dims kind attrs =
      List.fold_left (fun dims (Attr(name, aps)) ->
        if kind = name then
          let args =
            List.fold_left (fun args ap ->
              match ap with
              | AStr(s) -> s :: args
              | _ ->
                failure "unexpected argument to attribute '%s': %a"
                  name d_attrparam ap
            ) [] aps
          in
          List.rev args :: dims
        else
          dims
      ) [] attrs
    in
    let rec get_dims kind accum lv =
      let t = typeOfLval lv in
      match unrollType t with
      | TInt(_, attrs)
      | TFloat(_, attrs) ->
        let dims = get_attr_dims kind attrs in
        begin match List.length dims with
        | 0 -> accum
        | 1 ->
          let dims = List.hd dims in
          if (List.length dims) <> 1 then
            failure "cannot attach dimensions to fields of %a (%a)"
              dn_lval lv dn_type t;
          (List.hd dims, lv) :: accum
        | _ ->
          failure "cannot attach multiple dimensions to %a (%a)"
            dn_lval lv dn_type t
        end
      | TPtr(_, attrs) ->
        if (List.length (get_attr_dims kind attrs)) > 0 then
          failure "cannot attach dimensions to ptr type: %a" dn_lval lv;
        get_dims kind accum (mkMem (Lval(lv)) NoOffset)
      | TArray(_, _, attrs) ->
        if (List.length (get_attr_dims kind attrs)) > 0 then
          failure "cannot attach dimensions to array type: %a" dn_lval lv;
        get_dims kind accum (mkMem (StartOf(lv)) NoOffset)
      | TNamed(_, attrs) -> failwith "unrollType failure"
      | TComp(ci, attrs) ->
        let dims = get_attr_dims kind attrs in
        if (List.length dims) > 0 then
          List.fold_left (fun accum dims ->
            match dims with
            | [ field; dim ] -> (dim, addField lv field) :: accum
            | _ -> failure "invalid dimension assignment for %a" dn_lval lv
          ) accum dims
        else
          List.fold_left (fun accum fi ->
            get_dims kind accum (addField lv fi.fname)
          ) accum ci.cfields

      | TVoid(attrs)
      | TFun(_, _, _, attrs)
      | TEnum(_, attrs)
      | TBuiltin_va_list(attrs) ->
        if (List.length (get_attr_dims kind attrs)) > 0 then
          failure "cannot attach dimensions to %a (%a)" dn_lval lv dn_type t;
        accum
    in
    let base_exprs =
      List.fold_left (fun map vi ->
        List.fold_left (fun map (dim, lv) ->
          SMap.add dim (mk_DLeaf dim (key_of_lval lv)) map
        ) map (get_dims "width" [] (var vi))
      ) SMap.empty formals
    in
    let init =
      List.fold_left (fun init vi ->
        List.fold_left (fun init (dim, lv) ->
          try
            let we = SMap.find dim base_exprs in
            IntTrie.add (key_of_lval lv) we init
          with Not_found ->
            Printf.eprintf
              "Warning: could not find width expression for dimension %s\n"
              dim;
            init
        ) init (get_dims "dim" [] (var vi))
      ) IntTrie.empty formals
    in
    Inthash.clear stmtStartData;
    Inthash.add   stmtStartData sid init

  let pretty () data =
    let pairs =
      IntTrie.fold (fun key we pairs ->
        (sprint ~width:80 (dn_lval () (lval_of_key key)), we) :: pairs
      ) data []
    in
    let pairs = List.sort (fun a b -> compare (fst a) (fst b)) pairs in
    let width =
      List.fold_left (fun w (n,_) -> max (String.length n) w) 0 pairs
    in
    List.fold_left (fun doc (name,we) ->
      let width_doc = doc_of_widthexp () we in
      let width_doc = indent (width - (String.length name)) width_doc in
      doc ++ text name ++ text ": " ++ width_doc ++ line
    ) line pairs

  let computeFirstPredecessor _ data = data

  let combinePredecessors s ~old (data: widthexp IntTrie.t) =
    let updated = ref false in
    let join =
      IntTrie.fold (fun key we1 join ->
        try
          let we2 = IntTrie.find key join in
          if we1 = we2 then
            join
          else
            failure
              "same variable, different width expressions: %a\nold: %a new: %a"
              dn_lval (lval_of_key key)
              doc_of_widthexp we1
              doc_of_widthexp we2
        with Not_found ->
          updated := true;
          IntTrie.add key we1 join
      ) data old
    in
    if !updated then Some(join) else None

  let doInstr i data =
    debugf "DEBUG: processing instr: %a\n" dn_instr i;
    let rec helper data lv we =
      match we with
      | DVector(fields, _) ->
        SMap.fold (fun n we' data ->
          helper data (addField lv n) we'
        ) fields data
      | _ -> IntTrie.add (key_of_lval lv) we data
    in
    match i with
    | Set(lv, e, _) -> 
      Dataflow.Done(helper data lv (widthexp_of_exp data e))
    | Call(Some(rv), Lval(Var(vi), NoOffset), args, _) ->
      let we = widthexp_of_function_call data vi.vname args in
      Dataflow.Done(helper data rv we)
    | Call(Some(rv), _, args, _) ->
      failwith "indirect function calls not supported"
    | _ -> Dataflow.Default

  let doStmt _ _ = Dataflow.SDefault

  let doGuard _ _ = Dataflow.GDefault

  let filterStmt _ = true
end

module TransferAnalysis = Dataflow.ForwardsDataFlow(DimensionTransfer)

let mk_nop_fun data rv args = None

let mk_unary namef namew data rv args =
  let wargs = List.map (fun e -> widthexp_of_exp data e) args in
  let wargs = List.filter (fun we -> not (is_const_widthexp we)) wargs in
  let wargs = List.map exp_of_widthexp wargs in

  let check_formals name args =
    let f = Lval(var (Hashtbl.find functions name)) in
    let _, formals, _, _ = splitFunctionType (typeOf f) in
    let formals = argsToList formals in
    if (List.length formals) <> (List.length args) then
      None
    else
      List.fold_left2 (fun f (_,tf,_) arg ->
        if f = None then f
        else
          let ta = typeOf arg in
          match unrollType tf, unrollType ta with
          | TVoid(_), TVoid(_)
          | TInt(_), TInt(_)
          | TInt(_), TFloat(_)
          | TInt(_), TEnum(_)
          | TFloat(_), TInt(_)
          | TFloat(_), TFloat(_)
          | TEnum(_), TInt(_)
          | TPtr(_), TPtr(_)
          | TBuiltin_va_list(_), TBuiltin_va_list(_) -> f
          | TComp(cif, _), TComp(cia, _) when cif.cname = cia.cname -> f
          | _ -> None
      ) (Some([Call(rv, f, args, locUnknown)])) formals args
  in

  let repl = check_formals namew (args @ wargs) in
  if repl = None then
    if (check_formals namef args) = None then
      failure "%t: %s() called with wrong number of arguments" d_thisloc namef
    else
      None
  else
    repl

let mk_piecewise namef namew data rv args =
  let Some(rv) = rv in
  let wargs = List.map (widthexp_of_exp data) args in
  let fieldnames =
    List.fold_left (fun fns we ->
      match we with
      | DVector(map, _) -> SMap.fold (fun n _ fns ->  SSet.add n fns) map fns
      | DLinExp(_) -> fns
    ) SSet.empty wargs
  in
  if SSet.is_empty fieldnames then
    mk_unary namef namew data (Some(rv)) args
  else begin
    let args  = List.map (fun (Lval(lv)) -> lv) args in
    let changed, ins =
      SSet.fold (fun fn (changed, ins) ->
        let args' =
          List.rev (List.fold_left2 (fun args' lv we ->
            match we with
            | DVector(_) -> Lval(addField lv fn) :: args'
            | DLinExp(_) -> Lval(lv) :: args'
          ) [] args wargs)
        in
        match mk_unary namef namew data (Some(addField rv fn)) args' with
        | Some(ins') -> true, List.rev_append ins' ins
        | None ->
          let f = Lval(var (Hashtbl.find functions namef)) in
          changed, (Call(Some(addField rv fn), f, args', locUnknown) :: ins)
      ) fieldnames (false,[])
    in
    if changed then
      Some(List.rev ins)
    else
      None
  end

let mk_custom limited_f field_nums data rv args =
  let is_const, wargs =
    List.fold_left (fun (is_const, wargs) e ->
      let we = widthexp_of_exp data e in
      is_const && is_const_widthexp we, we :: wargs
    ) (true, []) (List.map (List.nth args) field_nums)
  in
  if is_const then
    None
  else
    let wargs = List.map exp_of_widthexp wargs in
    let f = Lval(var (Hashtbl.find functions limited_f)) in
    Some([Call(rv, f, args @ wargs, locUnknown)])

(* Maps non-band-limited C functions to an OCaml function that can generate
   band-limited calls *)
let band_limited_mapping = Hashtbl.create 11
let _ =
  List.iter (fun (n,f) -> Hashtbl.add band_limited_mapping n f) [
    "abs_f",             mk_unary "abs_f" "abs_w";
    "abs_f2",            mk_piecewise "abs_f" "abs_w";
    "abs_f3",            mk_piecewise "abs_f" "abs_w";
    "abs_f4",            mk_piecewise "abs_f" "abs_w";
    "add_f2f",           mk_nop_fun;
    "add_f2f2",          mk_nop_fun;
    "add_f3f",           mk_nop_fun;
    "add_f3f3",          mk_nop_fun;
    "add_f4f",           mk_nop_fun;
    "add_f4f4",          mk_nop_fun;
    "axis_align_widths", mk_nop_fun;
    "div_f2f",           mk_nop_fun;
    "div_f2f2",          mk_nop_fun;
    "div_f3f",           mk_nop_fun;
    "div_f3f3",          mk_nop_fun;
    "div_f4f",           mk_nop_fun;
    "div_f4f4",          mk_nop_fun;
    "dot_f2",            mk_nop_fun;
    "dot_f3",            mk_nop_fun;
    "dot_f4",            mk_nop_fun;
    "ceil_f",            mk_unary "ceil_f" "ceil_w";
    "ceil_f2",           mk_piecewise "ceil_f" "ceil_w";
    "ceil_f3",           mk_piecewise "ceil_f" "ceil_w";
    "ceil_f4",           mk_piecewise "ceil_f" "ceil_w";
    "cos",               mk_unary "cos" "cos_w";
    "floor_f",           mk_unary "floor_f" "floor_w";
    "floor_f2",          mk_piecewise "floor_f" "floor_w";
    "floor_f3",          mk_piecewise "floor_f" "floor_w";
    "floor_f4",          mk_piecewise "floor_f" "floor_w";
    "fract_f",           mk_unary "fract_f" "fract_w";
    "fract_f2",          mk_piecewise "fract_f" "fract_w";
    "fract_f3",          mk_piecewise "fract_f" "fract_w";
    "fract_f4",          mk_piecewise "fract_f" "fract_w";
    "gabor_noise",       mk_custom "filtered_gabor_noise" [7; 6];
    "lerp_ff",           mk_nop_fun;
    "lerp_f4f",          mk_nop_fun;
    "makef2_ff",         mk_nop_fun;
    "makef3_f",          mk_nop_fun;
    "makef3_fff",        mk_nop_fun;
    "makef4_f",          mk_nop_fun;
    "makef4_f3f",        mk_nop_fun;
    "makef4_ffff",       mk_nop_fun;
    "mul_f2f",           mk_nop_fun;
    "mul_f2f2",          mk_nop_fun;
    "mul_f3f",           mk_nop_fun;
    "mul_f3f3",          mk_nop_fun;
    "mul_f4f",           mk_nop_fun;
    "mul_f4f4",          mk_nop_fun;
    "samplef_ff",        mk_custom "samplef_ff_w" [4; 3];
    "samplef4_ff",       mk_custom "samplef4_ff_w" [4; 3];
    "saturate_f",        mk_unary "saturate_f" "saturate_w";
    "saturate_f4",       mk_piecewise "saturate_f" "saturate_w";
    "sin",               mk_unary "sin" "sin_w";
    "sq_f",              mk_unary "sq_f" "sq_w";
    "sq_f2",             mk_piecewise "sq_f" "sq_w";
    "sq_f3",             mk_piecewise "sq_f" "sq_w";
    "sq_f4",             mk_piecewise "sq_f" "sq_w";
    "step_f",            mk_unary "step_f" "step_w";
    "step_f2",           mk_piecewise "step_f" "step_w";
    "step_f3",           mk_piecewise "step_f" "step_w";
    "step_f4",           mk_piecewise "step_f" "step_w";
    "sub_f2f2",          mk_nop_fun;
    "sub_ff3",           mk_nop_fun;
    "sub_f3f3",          mk_nop_fun;
    "sub_f4f4",          mk_nop_fun;
    "swizzle_f3xy",      mk_nop_fun;
    "swizzle_f3xz",      mk_nop_fun;
    "swizzle_f3yxz",     mk_nop_fun;
    "swizzle_f3zxy",     mk_nop_fun;
    "swizzle_f3xzy",     mk_nop_fun;
    "swizzle_f4xyz",     mk_nop_fun;
    "swizzle_f4zyxw",    mk_nop_fun;
  ]

let init debug cilfile =
  DimensionTransfer.debug := debug;

  iterGlobals cilfile (fun g ->
    match g with
    | GVarDecl(vi,_) when isFunctionType vi.vtype ->
      Hashtbl.add functions vi.vname vi
    | GFun(fd,_) ->
      Hashtbl.add functions fd.svar.vname fd.svar
    | _ -> ()
  );

  Cfg.clearFileCFG cilfile;
  Cfg.computeFileCFG cilfile

type t = DimensionTransfer.t

let initStartData = DimensionTransfer.initStartData
let compute       = TransferAnalysis.compute
let getData sid   = Inthash.find DimensionTransfer.stmtStartData sid

let transformExp data e =
  match e with
  (*
  | BinOp((Mult|Div|Mod), e1, e2, t) ->
    let ds1 = dims_of_widthexp (widthexp_of_exp data e1) in
    let ds2 = dims_of_widthexp (widthexp_of_exp data e2) in
    if (SSet.cardinal (SSet.inter ds1 ds2)) > 0 then
      (* FIXME: can we handle simple cases like x*x? *)
      failwith "cannot take product of common dimensions";
    None
  *)
  | BinOp(
      (Shiftlt|Shiftrt|Lt|Gt|Le|Ge|Eq|Ne|BAnd|BXor|BOr|LAnd|LOr),
      _, _, _) ->
    failure "band-limited expression for (%a) unknown" dn_exp e
  | _ -> None

let transformFunc data name rv args =
  if Hashtbl.mem band_limited_mapping name then
    let gen = Hashtbl.find band_limited_mapping name in
    gen data rv args
  else
    failure "don't know how to bandlimit: %s(%a)" name (d_list ", " dn_exp) args

