#include <algorithm>
#include <cfloat>
#include <cmath>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <map>
#include <random>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#define DO_TIMING 0
#define USE_CLFFT 0

#include <time.h>

#if !USE_CLFFT
#include <fftw3.h>
#endif

#include <CL/cl.h>

extern "C" {
#include <pam.h>

#if USE_CLFFT
#include <clFFT.h>
#endif

#include "globals.h"
}

#include "fit/simplex.h"

#define quote_str(s) quote_str2(s)
#define quote_str2(s) #s

/*
 * Evaluates a shader variant on several images and determines
 *   1) image difference relative to a golden image
 *   2) fraction of energy outside the Nyquist limit
 *
 * Steps:
 *   1) initialize rendering framework (GPU)
 *      a) create a context and command queue
 *      b) compile FFT, SSIM, L2, and supersampling reducer (may be cached)
 *      c) [optional] compile golden shader
 *      d) compile candidate shader
 *      e) initialize randomness buffers
 *   2) for each requested image (difference metric)
 *      a) create GPU buffers
 *      b) render image
 *      c) get golden (from disk or regenerated)
 *      d) compute image difference
 *      e) copy difference value to CPU
 *      f) release GPU buffers
 *   3) for each requested image (aliasing metric)
 *      a) create GPU buffers
 *      b) render image
 *      c) take DFT of image
 *      d) compute enery outside Nyquist
 *      e) copy value to CPU
 *      f) release GPU buffers
 *   4) release rendering framework
 *
 * Note: We will want timing information for each of these tasks so that
 * alternate implementations may be evaluated and selected. We also need hooks
 * to dump out intermediate images to validate them against the "ground truth"
 * from earlier implementations.
 */

static std::map< std::string, unsigned long > timings;

unsigned long nsec() {
    struct timespec t;
    int status;

    status = clock_gettime( CLOCK_MONOTONIC, &t );
    if ( status != 0 ) {
        perror( "cannot read clock\n" );
        exit( 2 );
    }
    return t.tv_sec * 1000000000 + t.tv_nsec;
}

#if DO_TIMING
std::vector< std::string > timing_stack;

void
start( const std::string & key )
{
    if ( timing_stack.empty() ) {
        timing_stack.push_back( key );
        timings[ key ] = nsec() - timings[ key ];
    } else if ( timing_stack.back() != key ) {
        const std::string & old = timing_stack.back();
        timings[ old ] = nsec() - timings[ old ];
        timing_stack.push_back( key );
        timings[ key ] = nsec() - timings[ key ];
    }
}

void
stop( const std::string & key )
{
    if ( timing_stack.empty() ) {
        std::cerr << "ERROR: stopping timing for " << key
                  << " but not currently timing anything!\n";
        exit( 7 );
    } else if ( timing_stack.back() != key ) {
        std::cerr << "ERROR: stopping timing for " << key
                  << " but was timing " << timing_stack.back() << '\n';
        exit( 7 );
    } else {
        timings[ key ] = nsec() - timings[ key ];
        timing_stack.pop_back();
        if ( ! timing_stack.empty() ) {
            const std::string & old = timing_stack.back();
            timings[ old ] = nsec() - timings[ old ];
        }
    }
}
#else
#define start( key ) /**/
#define stop( key )  /**/
#endif

#ifdef PAM_STRUCT_SIZE
#  define LENGTH_OF_PAM PAM_STRUCT_SIZE( tuple_type )
#else
#  define LENGTH_OF_PAM sizeof( struct pam )
#endif

enum diff_t { NO_DIFF, DIFF_L2, DIFF_SSIM };
enum msaa_t { MSAA_GAUSSIAN, MSAA_HAMMERSLEY };

struct simplex_config {
    unsigned terms;

    simplex_config() : terms( 0 ) {}
};

struct image_options {
    std::string outfile;
    std::string golden;
    unsigned    width;
    unsigned    height;
    unsigned    subsamples;
    unsigned    upsample;
    unsigned    msaa;
    float       zoom;

    image_options()
        : outfile(),
          golden(),
          width( 512 ),
          height( 512 ),
          subsamples( 1 ),
          upsample( 1 ),
          msaa( 1 ),
          zoom( 1.0f )
    {}
};

struct options {
    std::vector< image_options > images;
    std::vector< float > coefficients;
    std::string shader;
    simplex_config simplex;
    float    seed;
    unsigned gpu;
    diff_t   diff_alg;
    msaa_t   msaa_alg;
    bool     grayscale;
    bool     aliasing;
    bool     perspective;
    bool     similarity;
    bool     fps;

    options()
        : images(),
          coefficients(),
          shader(),
          simplex(),
          seed( 0.0f ),
          gpu( 0 ),
          diff_alg( NO_DIFF ),
          msaa_alg( MSAA_GAUSSIAN ),
          grayscale( false ),
          aliasing( false ),
          perspective( false ),
          similarity( false ),
          fps( false )
    {}
};

void
check_args( int argc, char ** argv, const char * option, int n )
{
    if ( argc < n ) {
        std::cerr << "ERROR: missing " << option << " argument\n";
        exit( 1 );
    }
}

void
get_flag_arg(
    int argc, char ** argv, const char * option, int n, const char * format, ...
) {
    if ( argc <= 0 ) {
        std::cerr << "ERROR: missing " << option << " argument\n";
        exit( 1 );
    }

    va_list args;
    va_start( args, format );
    int status = vsscanf( *argv, format, args );
    va_end( args );
    if ( status != n ) {
        std::cerr << "ERROR: could not parse '" << *argv << "'\n";
        exit( 1 );
    }
}

void
print_usage( const std::string & prog )
{
    std::cout << "Usage: " << prog << " [options] shader.cl\n";
    std::cout << "\n";
    std::cout << "Global Options:\n";
    std::cout << "  -h, --help      show this help and exit\n";
    std::cout << "  --aliasing      compute energy outside nyquist limit\n";
    std::cout << "  --coeff c1,c2   comma-separated list of coefficients\n";
    std::cout << "  --diff fun      compute the difference vs a golden image\n";
    std::cout << "  --fps           report average frames per second\n";
    std::cout << "  --gpu N         select the GPU device\n";
    std::cout << "  --gray          render a grayscale image\n";
    std::cout << "  --perspective   render a perspective image\n";
    std::cout << "  --sampling alg  algorithm for generating sampling points\n";
    std::cout << "  --seed X        set the random seed\n";
    std::cout << "  --similarity    compute the bidirectional similarity\n";
    std::cout << "  --simplex N     number of dimensions for simplex search\n";
    std::cout << "  -i              introduces the next set of image options\n";
    std::cout << "\n";
    std::cout << "Image-Specific Options:\n";
    std::cout << "  -g, --golden  file  golden image to compare against\n";
    std::cout << "  -o, --output  file  write the rendered image to a file\n";
    std::cout << "  --msaa N            number of multi-samples to take\n";
    std::cout << "  --size NxN          size of the image in pixels\n";
    std::cout << "  --subsamples N      number of subsamples per pixel\n";
    std::cout << "  --upsample N        upsample final image by NxN\n";
    std::cout << "  --zoom f            zoom-out factor\n";
}

int
parse_args( int argc, char ** argv, options & opts )
{
    const std::string & prog = argv[ 0 ];
    image_options img;
    for ( ++argv, --argc; argc > 0; ++argv, --argc ) {
        const std::string & arg = *argv;
        if ( arg == "--help" || arg == "-h" ) {
            print_usage( prog );
            exit( 0 );
        } else if ( arg == "--aliasing" ) {
            opts.aliasing = true;
        } else if ( arg == "--coeff" ) {
            if ( --argc <= 0 ) {
                std::cerr << "ERROR: missing --coeff argument\n";
                exit( 1 );
            }
            opts.coefficients.clear();
            const char * opt = *( ++argv );
            int remaining = strlen( opt );
            float c;
            int size;
            while ( opt[ 0 ] != '\0' && remaining > 0 ) {
                int status = sscanf( opt, "%g%*[,]%n", &c, &size );
                if ( status == 0 ) {
                    std::cerr << "ERROR: could not parse '" << *opt << "'\n";
                    exit( 1 );
                }
                opts.coefficients.push_back( c );
                opt += size;
                remaining -= size;
            }
        } else if ( arg == "--diff" ) {
            check_args( --argc, ++argv, "--diff", 1 );
            const std::string & fun = *argv;
            if ( fun == "l2" )
                opts.diff_alg = DIFF_L2;
            else if ( fun == "ssim" )
                opts.diff_alg = DIFF_SSIM;
            else {
                std::cerr << "unknown difference algorithm " << fun << '\n';
                exit( 1 );
            }
        } else if ( arg == "--fps" ) {
            opts.fps = true;
        } else if ( arg == "--gpu" ) {
            get_flag_arg( --argc, ++argv, "--gpu", 1, "%u", &opts.gpu );
        } else if ( arg == "--gray" ) {
            opts.grayscale = true;
        } else if ( arg == "--perspective" ) {
            opts.perspective = true;
        } else if ( arg == "--sampling" ) {
            check_args( --argc, ++argv, "--sampling", 1 );
            const std::string & fun = *argv;
            if ( fun == "gaussian" )
                opts.msaa_alg = MSAA_GAUSSIAN;
            else if ( fun == "hammersley" )
                opts.msaa_alg = MSAA_HAMMERSLEY;
            else {
                std::cerr << "unknown sampling algorithm " << fun << '\n';
                exit( 1 );
            }
        } else if ( arg == "--seed" ) {
            get_flag_arg( --argc, ++argv, "--seed", 1, "%g", &opts.seed );
        } else if ( arg == "--similarity" ) {
            opts.similarity = true;
        } else if ( arg == "--simplex" ) {
            get_flag_arg(
                --argc, ++argv, "--simplex", 1, "%u", &opts.simplex.terms
            );
        } else if ( arg == "-i" ) {
            if ( opts.images.empty() ) {
                opts.images.push_back( image_options() );
            } else {
                opts.images.push_back( opts.images.back() );
                opts.images.back().outfile = "";
                opts.images.back().golden  = "";
            }

        } else if ( arg == "-g" || arg == "--golden" ) {
            check_args( --argc, ++argv, arg.c_str(), 1 );
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            opts.images.back().golden = *argv;
        } else if ( arg == "-o" || arg == "--output" ) {
            check_args( --argc, ++argv, arg.c_str(), 1 );
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            opts.images.back().outfile = *argv;
        } else if ( arg == "--msaa" ) {
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            get_flag_arg(
                --argc, ++argv, "--msaa", 1, "%u",
                &opts.images.back().msaa
            );
        } else if ( arg == "--size" ) {
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            get_flag_arg(
                --argc, ++argv, "--size", 2, "%ux%u",
                &opts.images.back().width, &opts.images.back().height
            );
        } else if ( arg == "--subsamples" ) {
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            get_flag_arg(
                --argc, ++argv, "--subsamples", 1, "%u",
                &opts.images.back().subsamples
            );
        } else if ( arg == "--upsample" ) {
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            get_flag_arg(
                --argc, ++argv, "--upsample", 1, "%u",
                &opts.images.back().upsample
            );
        } else if ( arg == "--zoom" ) {
            if ( opts.images.empty() )
                opts.images.push_back( image_options() );
            get_flag_arg( --argc, ++argv, "--zoom", 1, "%f",
            &opts.images.back().zoom );
        } else if ( **argv == '-' ) {
            std::cerr << "ERROR: unrecognized flag '" << arg << "'\n";
            print_usage( prog );
            exit( 1 );
        } else if ( opts.shader == "" ) {
            opts.shader = arg;
        } else {
            std::cerr << "ERROR: unrecognized argument '" << arg << "'\n";
            print_usage( prog );
            exit( 1 );
        }
    }

    if ( opts.images.empty() ) {
        std::cerr << "ERROR: you must specify at least one image to render\n";
        exit( 1 );
    }

    if ( opts.shader == "" ) {
        std::cerr << "ERROR: you must specify a --shader argument\n";
        exit( 1 );
    }

    if ( opts.diff_alg != NO_DIFF || opts.similarity || opts.simplex.terms > 0 ) {
        for ( size_t i = 0; i < opts.images.size(); ++i ) {
            if ( opts.images[ i ].golden.size() == 0 ) {
                std::cerr <<
                    "ERROR: diff, similarity, and simplex algorithms require "
                    "a golden reference for image " << i << '\n';
                exit( 1 );
            }
        }
    }

    return 0;
}

static bool pam_init = false;

void
writePPM(
    const char * fname,
    const float4 * raster, 
    unsigned int width,
    unsigned int height
) {
    struct pam metadata;
    FILE * fh;
    tuple * row;
    unsigned int x, y, i;

    if ( !pam_init ) {
        pm_init( "pnm", 0 );
        pam_init = true;
    }

    fh = fopen( fname, "wb" );

    metadata.size   = sizeof( metadata );
    metadata.len    = LENGTH_OF_PAM;
    metadata.height = height;
    metadata.width  = width;
    metadata.maxval = 255;
    metadata.format = RPPM_FORMAT;
    metadata.depth  = 3;
    metadata.file = fh;

    pnm_writepaminit( &metadata );
    row = pnm_allocpamrow( &metadata );
    for ( y = 0, i = 0; y < height; ++y ) {
        for ( x = 0; x < width; ++x, ++i ) {
            row[ x ][ 0 ] = saturate_f( raster[ i ].x ) * 255.0f;
            row[ x ][ 1 ] = saturate_f( raster[ i ].y ) * 255.0f;
            row[ x ][ 2 ] = saturate_f( raster[ i ].z ) * 255.0f;
        }
        pnm_writepamrow( &metadata, row );
    }
    pnm_freepamrow( row );

    fclose( fh );
}

float4 *
readPPM(
    const char * fname,
    unsigned int & width,
    unsigned int & height
) {
    struct pam metadata;
    FILE * fh;
    tuple * row;
    float4 * raster;
    float scale;
    unsigned int x, y, c, i;

    if ( ! pam_init ) {
        pm_init( "pnm", 0 );
        pam_init = true;
    }

    fh = fopen( fname, "rb" );
    pnm_readpaminit( fh, &metadata, LENGTH_OF_PAM );

    scale = 1.0f / (float) metadata.maxval;
    width  = metadata.width;
    height = metadata.height;

    raster = new float4[ width * height ];
    row = pnm_allocpamrow( &metadata );
    for ( y = 0, i = 0; y < height; ++y ) {
        pnm_readpamrow( &metadata, row );
        for ( x = 0; x < width; ++x, ++i ) {
            raster[ i ].x = row[ x ][ 0 ] * scale;
            raster[ i ].y = row[ x ][ 1 ] * scale;
            raster[ i ].z = row[ x ][ 2 ] * scale;
        }
    }
    pnm_freepamrow( row );

    fclose( fh );

    return raster;
}

void
clSuccess( cl_int status, const char * msg )
{
    if ( status == CL_SUCCESS )
        return;

    std::cerr << "OpenCL ERROR: (" << status << ") " << msg << '\n';
    exit( 2 );
}

void
clSuccess( cl_int status, const std::string & msg )
{
    if ( status == CL_SUCCESS )
        return;

    std::cerr << "OpenCL ERROR: (" << status << ") " << msg << '\n';
    exit( 2 );
}

struct cl_env {
    cl_env( unsigned int device_num )
        : dev( 0 ), context( 0 ), queue( 0 ),
          kern(), buffer(), progs()
#if USE_CLFFT
          , init_clfft( false )
#endif
    {
        cl_int status;

        start( "platform" );
        cl_uint nplat = 0;
        status = clGetPlatformIDs( 0, NULL, &nplat );
        clSuccess( status, "getting number of platforms" );
        cl_platform_id * plats = new cl_platform_id[ nplat ];
        status = clGetPlatformIDs( nplat, plats, NULL );
        clSuccess( status, "getting available platforms" );
        stop( "platform" );

        cl_uint ndev = 0;
        for ( cl_uint i = 0; i < nplat; ++i ) {
            status = clGetDeviceIDs(
                plats[ i ], CL_DEVICE_TYPE_GPU, 0, NULL, &ndev
            );
            if ( status == CL_DEVICE_NOT_FOUND || ndev == 0 )
                continue;
            if ( ndev < device_num ) {
                device_num -= ndev;
                ndev = 0;
                continue;
            }

            cl_device_id * devs = new cl_device_id[ ndev ];
            status = clGetDeviceIDs(
                plats[ i ], CL_DEVICE_TYPE_GPU, ndev, devs, NULL
            );
            clSuccess( status, "getting devices" );

            dev = devs[ device_num ];
            delete[] devs;

            start( "context" );
            cl_context_properties props[] = {
                (cl_context_properties) CL_CONTEXT_PLATFORM,
                (cl_context_properties) plats[ i ],
                0
            };
            context = clCreateContext( props, 1, &dev, NULL, NULL, &status );
            clSuccess( status, "creating context" );
            stop( "context" );
            break;
        }

        delete[] plats;
        if ( ndev == 0 ) {
            std::cerr << "ERROR: no compatible devices found\n";
            exit( 2 );
        }

        start( "queue" );
        queue = clCreateCommandQueue(
            context, dev, CL_QUEUE_PROFILING_ENABLE, &status
        );
        clSuccess( status, "creating command queue" );
        stop( "queue" );
    }

    ~cl_env() {
        cl_int status;

#if USE_CLFFT
        if ( init_clfft ) {
            status = clfftTeardown();
            clSuccess( status, "releasing clFFT resources" );
        }
#endif

        std::map< std::string, cl_kernel >::iterator i;
        for ( i = kern.begin(); i != kern.end(); ++i ) {
            status = clReleaseKernel( i->second );
            if ( status != CL_SUCCESS )
                clSuccess( status, "releasing kernel " + i->first );
        }

        std::vector< cl_program >::reverse_iterator j;
        for ( j = progs.rbegin(); j != progs.rend(); ++j )
            clSuccess( clReleaseProgram( *j ), "releasing program" );

        std::map< std::string, std::pair< cl_mem, size_t > >::iterator k;
        for ( k = buffer.begin(); k != buffer.end(); ++k ) {
            status = clReleaseMemObject( k->second.first );
            if ( status != CL_SUCCESS )
                clSuccess( status, "releasing buffer " + k->first );
        }

        clSuccess( clReleaseContext( context ), "releasing context" );
    }

    void load_program(
        const char * fname, const char * options, const char ** names
    ) {
        std::ifstream f( fname );
        std::string str;

        f.seekg( 0, std::ios::end );
        str.resize( f.tellg() );
        f.seekg( 0, std::ios::beg );
        f.read( &str[ 0 ], str.size() );
        f.close();

        const char * source = str.c_str();

        cl_int status;

        start( "compile" );
        size_t length = strlen( source );
        cl_program p =
            clCreateProgramWithSource( context, 1, &source, &length, &status );
        clSuccess( status, "creating program" );
        progs.push_back( p );

        status = clBuildProgram( p, 1, &dev, options, NULL, NULL );
        if ( status == CL_BUILD_PROGRAM_FAILURE || status == CL_INVALID_BINARY ) {
            status = clGetProgramBuildInfo(
                p, dev, CL_PROGRAM_BUILD_LOG, 0, NULL, &length
            );
            clSuccess( status, "getting build log size" );

            char * log = new char[ length + 1 ];
            status = clGetProgramBuildInfo(
                p, dev, CL_PROGRAM_BUILD_LOG, length, log, NULL
            );
            if ( status != CL_SUCCESS ) {
                delete[] log;
                clSuccess( status, "getting build log" );
            }
            log[ length ] = '\0';
            std::cerr << "=== build log ===\n";
            std::cerr << log << '\n';
            delete[] log;

            exit( 1 );
        }
        clSuccess( status, "building program" );

        for ( ; *names != NULL; ++names ) {
            cl_kernel k = clCreateKernel( p, *names, &status );
            if ( status != CL_SUCCESS ) {
                std::string msg( "getting " );
                clSuccess( status, msg + (*names) + " kernel" );
            }
            kern[ *names ] = k;
        }
        stop( "compile" );
    }

    void add_buffer( const char * name, cl_mem_flags flags, size_t size ) {
        cl_int status;
        cl_mem buf = clCreateBuffer( context, flags, size, NULL, &status );
        if ( status != CL_SUCCESS ) {
            std::stringstream msg;
            msg << "creating buffer " << name << " (" << size << " bytes)";
            clSuccess( status, msg.str() );
        }
        buffer[ name ] = std::make_pair( buf, size );
    }

    void free_buffer( const char * name ) {
        cl_int status;
        std::map< std::string, std::pair< cl_mem, size_t > >::iterator i;
        i = buffer.find( name );
        if ( i != buffer.end() ) {
            status = clReleaseMemObject( i->second.first );
            clSuccess( status, "releasing buffer" );
            buffer.erase( i );
        }
    }

    void rename_buffer( const char * src, const char * dst ) {
        auto i = buffer.find( dst );
        if ( i != buffer.end() )
            free_buffer( dst );
        buffer[ dst ] = buffer.at( src );
        buffer.erase( src );
    }

    template< typename T >
    cl_event write_buffer( const char * name, T * data ) {
        cl_event event;

        std::pair< cl_mem, size_t > entry = buffer[ name ];
        cl_int status = clEnqueueWriteBuffer(
            queue, entry.first, CL_FALSE,
            0, entry.second,
            data,
            0, NULL, &event
        );
        if ( status != CL_SUCCESS )
            clSuccess( status, std::string( "enqueing write to " ) + name );
        return event;
    }

    template< typename T >
    T * read_buffer( const char * name, cl_event * in, cl_event * out ) {
        std::pair< cl_mem, size_t > entry = buffer[ name ];
        T * result = new T[ entry.second / sizeof( T ) ];
        cl_int status = clEnqueueReadBuffer(
            queue, entry.first, CL_FALSE,
            0, entry.second,
            result,
            in != NULL, in, out
        );
        if ( status != CL_SUCCESS )
            clSuccess( status, std::string( "enqueing read from " ) + name );
        return result;
    }

    cl_mem * buf( const char * name ) {
        return &( buffer[ name ].first );
    }

#if USE_CLFFT
    void
    init_clFFT()
    {
        if ( init_clfft )
            return;

        cl_int status;

        clfftSetupData setup;
        status = clfftInitSetupData( &setup );
        clSuccess( status, "initializing clFFT setup data" );
        status = clfftSetup( &setup );
        clSuccess( status, "setting up clFFT" );

        init_clfft = true;
    }
#endif

    cl_device_id     dev;
    cl_context       context;
    cl_command_queue queue;
    std::map< std::string, cl_kernel > kern;
    std::map< std::string, std::pair< cl_mem, size_t > > buffer;
    std::vector< cl_program > progs;
#if USE_CLFFT
    bool             init_clfft;
#endif
};

global_buffers *
populate_global_buffers()
{
    global_buffers * g = new global_buffers;

    g->sizex = 512;
    g->sizey = 512;

    const unsigned int size = g->sizex * g->sizey;
    const float rand_max = (float) RAND_MAX;

    g->g_txRand_f = new float[ size ];
    for ( unsigned int i = 0; i < size; ++i )
        g->g_txRand_f[ i ] = rand() / rand_max;

    g->g_txRand_f4 = new float4[ size ];
    for ( unsigned int i = 0; i < size; ++i )
        g->g_txRand_f4[ i ] = makef4_ffff(
            rand() / rand_max,
            rand() / rand_max,
            rand() / rand_max,
            rand() / rand_max
        );

    return g;
}

void
init_opencl( cl_env & env, const options & opts )
{
    cl_int status;
    cl_event events[ 4 ];

    const char * std_kernels[] = {
        "heatmap", "l2_diff", "ssim_diff",
        "init_bidir_sim", "bidirectional_similarity", "merge_bidir_sim",
        "make_complex", "dft_image",
        "supersample", "upsample_img", "flatten_image", 0
    };
    const char * variant_kernels[] = { "plot", "set_buffers", 0 };

    std::string defines;
    if ( opts.grayscale )
        defines += "-DGRAYSCALE ";
    if ( opts.msaa_alg == MSAA_HAMMERSLEY )
        defines += "-DHAMMERSLEY ";
    if ( opts.perspective )
        defines += "-DPERSPECTIVE";

    env.load_program( quote_str(KERNELS_CL), "", std_kernels );
    env.load_program( opts.shader.c_str(), defines.c_str(), variant_kernels );

    start( "buffers" );
    global_buffers * g = populate_global_buffers();
    env.add_buffer(
        "txRand_f", CL_MEM_READ_ONLY, sizeof( cl_float ) * g->sizex * g->sizey
    );
    env.add_buffer(
        "txRand_f4", CL_MEM_READ_ONLY, sizeof( cl_float4 ) * g->sizex * g->sizey
    );
    env.add_buffer( "g", CL_MEM_READ_WRITE, sizeof( global_buffers ) );

    events[ 0 ] = env.write_buffer( "txRand_f", g->g_txRand_f );
    events[ 1 ] = env.write_buffer( "txRand_f4", g->g_txRand_f4 );
    events[ 2 ] = env.write_buffer( "g", g );

    status = clSetKernelArg(
        env.kern[ "set_buffers" ], 0, sizeof( cl_mem ), env.buf( "txRand_f" )
    );
    clSuccess( status, "setting set_buffers task arg 0" );

    status = clSetKernelArg(
        env.kern[ "set_buffers" ], 1, sizeof( cl_mem ), env.buf( "txRand_f4" )
    );
    clSuccess( status, "setting set_buffers task arg 1" );

    status = clSetKernelArg(
        env.kern[ "set_buffers" ], 2, sizeof( cl_mem ), env.buf( "g" )
    );
    clSuccess( status, "setting set_buffers task arg 2" );

    status = clEnqueueTask(
        env.queue, env.kern[ "set_buffers" ], 3, events, events + 3
    );
    clSuccess( status, "enqueing set_buffers task" );

    clSuccess( clFlush( env.queue ), "flushing queue" );
    clSuccess( clWaitForEvents( 4, events ), "waiting for events" );
    stop( "buffers" );

    for ( int i = 4; i > 0; --i )
        clSuccess( clReleaseEvent( events[ i - 1 ] ), "releasing event" );

    delete g;
}

const char *
plot_cl(
    cl_env & env,
    unsigned int & raster_w,
    unsigned int & raster_h,
    unsigned int subpixels,
    msaa_t msaa_alg,
    unsigned int msaa,
    float scale_factor,
    bool expand,
    const char * coeff,
    float & plot_time
) {
    float step = 1.0f;
    if ( expand ) {
        raster_w *= subpixels;
        raster_h *= subpixels;
        step /= subpixels;
        subpixels = 1;
    }

    env.free_buffer( "result" );
    env.add_buffer(
        "result", CL_MEM_READ_WRITE,
        sizeof( cl_float4 ) * raster_w * raster_h * subpixels * subpixels
    );
    if ( subpixels > 1 ) {
        env.free_buffer( "reduced" );
        env.add_buffer(
            "reduced", CL_MEM_WRITE_ONLY,
            sizeof( cl_float4 ) * raster_w * raster_h
        );
    }
    if ( msaa > 1 ) {
        size_t extent = raster_w * raster_h * subpixels * subpixels;
        env.free_buffer( "state" );
        env.add_buffer(
            "state", CL_MEM_READ_WRITE, sizeof( cl_uint2 ) * extent
        );
        cl_uint2 * state = new cl_uint2[ extent ];
        switch ( msaa_alg ) {
            case MSAA_GAUSSIAN:
                for ( size_t i = 0; i < extent; ++i ) {
                    state[ i ].s[0] = rand();
                    state[ i ].s[1] = 0;
                }
                break;
            case MSAA_HAMMERSLEY:
                for ( size_t i = 0; i < extent; ++i ) {
                    state[ i ].s[0] = msaa;
                    state[ i ].s[1] = 0;
                }
                break;
        }
        cl_event evt = env.write_buffer( "state", state );
        clSuccess( clFlush( env.queue ), "flushing queue" );
        clSuccess( clWaitForEvents( 1, &evt ), "waiting for write" );
        clSuccess( clReleaseEvent( evt ), "releasing event" );
        delete[] state;
    }

    cl_int status;

    status = clSetKernelArg(
        env.kern[ "plot" ], 0, sizeof( cl_mem ), env.buf( "g" )
    );
    clSuccess( status, "setting plot kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "plot" ], 1,
        sizeof( cl_mem ), msaa > 1 ? env.buf( "state" ) : NULL
    );
    clSuccess( status, "setting plot kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "plot" ], 2, sizeof( cl_float ), &scale_factor
    );
    clSuccess( status, "setting plot kernel arg 2" );

    status = clSetKernelArg( env.kern[ "plot" ], 3, sizeof( cl_float ), &step );
    clSuccess( status, "setting plot kernel arg 3" );

    status = clSetKernelArg(
        env.kern[ "plot" ], 4, sizeof( cl_mem ), env.buf( "result" )
    );
    clSuccess( status, "setting plot kernel arg 4" );

    status = clSetKernelArg(
        env.kern[ "plot" ], 5, sizeof( cl_mem ),
        coeff == NULL ? NULL : env.buf( coeff )
    );
    clSuccess( status, "setting plot kernel arg 5" );

    cl_event events[ 4 ];

    size_t global_size[] = { raster_w * subpixels, raster_h * subpixels };
    if ( msaa > 1 ) {
        cl_event * iters = new cl_event[ msaa ];
        for ( unsigned i = 0; i < msaa; ++i ) {
            status = clEnqueueNDRangeKernel(
                env.queue, env.kern[ "plot" ],
                2, NULL, global_size, NULL,
                i == 0 ? 0 : 1, i == 0 ? NULL : iters + (i - 1), iters + i
            );
            clSuccess( status, "launching kernel" );
        }
        events[ 0 ] = iters[ 0 ];
        events[ 1 ] = iters[ msaa - 1 ];
        for ( unsigned i = msaa - 1; i > 1; --i )
            clSuccess( clReleaseEvent( iters[ i - 1 ] ), "releasing event" );
        delete[] iters;
    } else {
        status = clEnqueueNDRangeKernel(
            env.queue, env.kern[ "plot" ],
            2, NULL, global_size, NULL,
            0, NULL, events + 0
        );
        clSuccess( status, "launching kernel" );
        events[ 1 ] = events[ 0 ];
    }

    const char * result;
    if ( subpixels > 1 ) {
        status = clSetKernelArg(
            env.kern[ "supersample" ], 0,
            sizeof( cl_mem ), env.buf( "result" )
        );
        clSuccess( status, "setting supersample kernel arg 0" );

        status = clSetKernelArg(
            env.kern[ "supersample" ], 1,
            sizeof( cl_mem ), env.buf( "reduced" )
        );
        clSuccess( status, "setting supersample kernel arg 1" );

        status = clSetKernelArg(
            env.kern[ "supersample" ], 2, sizeof( cl_uint ), &subpixels
        );
        clSuccess( status, "setting supersample kernel arg 2" );

        global_size[ 0 ] = raster_w;
        global_size[ 1 ] = raster_h;
        status = clEnqueueNDRangeKernel(
            env.queue, env.kern[ "supersample" ],
            2, NULL, global_size, NULL,
            1, events + 1, events + 2
        );
        clSuccess( status, "enqueing supersample kernel" );

        result = "reduced";
    } else {
        result = "result";
    }

    status = clSetKernelArg(
        env.kern[ "flatten_image" ], 0, sizeof( cl_mem ), env.buf( result )
    );
    clSuccess( status, "setting flatten kernel arg 0" );

    // global_size is now set correctly on both branches
    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "flatten_image" ],
        2, NULL, global_size, NULL,
        1, events + ( subpixels > 1 ? 2 : 1 ), events + 3
    );
    clSuccess( status, "enqueing flattening kernel" );

    clSuccess( clFlush( env.queue ), "flushing queue" );
    status = clWaitForEvents( 1, events + 3 );
    clSuccess( status, "waiting for queue to finish" );

    clGetEventInfo(
        events[ 3 ], CL_EVENT_COMMAND_EXECUTION_STATUS,
        sizeof( status ), &status, NULL
    );
    if ( status != CL_COMPLETE )
        clSuccess( status, "running kernel" );

    cl_ulong t1, t2;
    status = clGetEventProfilingInfo(
        events[ 0 ], CL_PROFILING_COMMAND_START, sizeof( t1 ), &t1, 0
    );
    clSuccess( status, "getting plot start time" );
    status = clGetEventProfilingInfo(
        events[ 1 ], CL_PROFILING_COMMAND_END, sizeof( t2 ), &t2, 0
    );
    clSuccess( status, "getting plot stop time" );

    plot_time = 1e-9f * (float)( t2 - t1 );

    clSuccess( clReleaseEvent( events[ 3 ] ), "releasing event 3" );
    if ( subpixels > 1 )
        clSuccess( clReleaseEvent( events[ 2 ] ), "releasing event 2" );
    if ( events[ 0 ] != events[ 1 ] )
        clSuccess( clReleaseEvent( events[ 1 ] ), "releasing event 1" );
    clSuccess( clReleaseEvent( events[ 0 ] ), "releasing event 0" );

    return result;
}

const char *
load_golden( cl_env & env, const char * fname, unsigned int w, unsigned int h )
{
    unsigned int w2, h2;
    start( "read" );
    float4 * golden = readPPM( fname, w2, h2 );
    stop( "read" );
    if ( w2 != w || h2 != h ) {
        std::cerr << "golden image is wrong size: "
                  << w2 << 'x' << h2
                  << " (should be " << w << 'x' << h << ")\n";
        exit( 3 );
    }

    start( "copy" );
    cl_event event;
    env.free_buffer( "golden" );
    env.add_buffer( "golden", CL_MEM_READ_ONLY, sizeof( cl_float4 ) * w * h );
    event = env.write_buffer( "golden", golden );
    clSuccess( clWaitForEvents( 1, &event ), "writing golden image" );
    clSuccess( clReleaseEvent( event ), "releasing event" );
    stop( "copy" );

    delete[] golden;

    return "golden";
}

cl_event
get_heatmap(
    cl_env & env,
    const char * delta,
    const char * img,
    size_t w,
    size_t h,
    cl_event parent
) {
    cl_int status;

    status = clSetKernelArg(
        env.kern[ "heatmap" ], 0, sizeof( cl_mem ), env.buf( delta )
    );
    clSuccess( status, "setting heatmap kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "heatmap" ], 1, sizeof( cl_mem ), env.buf( img )
    );
    clSuccess( status, "setting heatmap kernel arg 1" );

    size_t global_size[] = { w, h };
    cl_event event;
    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "heatmap" ],
        2, NULL, global_size, NULL,
        parent == NULL ? 0 : 1, &parent, &event
    );
    clSuccess( status, "enquing heatmap kernel" );

    return event;
}

void
upsample(
    cl_env & env,
    const char * img,
    unsigned & w,
    unsigned & h,
    unsigned factor
) {
    cl_int status;

    env.free_buffer( "upsample" );
    env.add_buffer(
        "upsample", CL_MEM_READ_WRITE,
        sizeof( cl_float4 ) * w * h * factor * factor
    );

    status = clSetKernelArg(
        env.kern[ "upsample_img" ], 0, sizeof( cl_mem ), env.buf( img )
    );
    clSuccess( status, "setting upsample_img kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "upsample_img" ], 1, sizeof( cl_mem ), env.buf( "upsample" )
    );
    clSuccess( status, "setting upsample_img kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "upsample_img" ], 2, sizeof( cl_uint ), &factor
    );
    clSuccess( status, "setting upsample_img kernel arg 2" );

    size_t global_size[] = { w, h };
    cl_event event;

    start( "upsampling" );
    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "upsample_img" ],
        2, NULL, global_size, NULL,
        0, NULL, &event
    );
    clSuccess( status, "enqueuing upsample_img kernel" );

    clSuccess( clWaitForEvents( 1, &event ), "running upsample_img kernel" );
    stop( "upsampling" );

    clSuccess( clReleaseEvent( event ), "releasing event" );

    env.rename_buffer( "upsample", img );
    w *= factor;
    h *= factor;
}

float
reduce(
    cl_env & env,
    const char * delta,
    size_t w,
    size_t h,
    cl_event parent
) {
    cl_int status;

    start( "average" );
    cl_event event;
    float * data = env.read_buffer< float >( delta, &parent, &event );
    clSuccess( clWaitForEvents( 1, &event ), "waiting for data" );
    clSuccess( clReleaseEvent( event ), "releasing event" );

    float accum = data[ 0 ];
    for ( size_t i = 1; i < w * h; ++i )
        accum += ( data[ i ] - accum ) / i;
    stop( "average" );

    return accum;
}

const char *
l2_diff_cl(
    cl_env & env,
    const char * img1,
    const char * img2,
    unsigned int raster_w,
    unsigned int raster_h,
    cl_event & event
) {
    cl_int status;

    env.free_buffer( "difference" );
    env.add_buffer(
        "difference", CL_MEM_READ_WRITE,
        sizeof( cl_float ) * raster_w * raster_h
    );

    status = clSetKernelArg(
        env.kern[ "l2_diff" ], 0, sizeof( cl_mem ), env.buf( img1 )
    );
    clSuccess( status, "setting l2_diff kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "l2_diff" ], 1, sizeof( cl_mem ), env.buf( img2 )
    );
    clSuccess( status, "setting l2_diff kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "l2_diff" ], 2, sizeof( cl_mem ), env.buf( "difference" )
    );
    clSuccess( status, "setting l2_diff kernel arg 2" );

    size_t global_size[] = { raster_w, raster_h };
    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "l2_diff" ],
        2, NULL, global_size, NULL,
        0, NULL, &event
    );
    clSuccess( status, "enqueing l2_diff kernel" );

    return "difference";
}

const char *
ssim_diff_cl(
    cl_env & env,
    const char * img1,
    const char * img2,
    unsigned int raster_w,
    unsigned int raster_h,
    cl_event & event
) {
    cl_int status;

    const cl_uint winsize = 8;

    env.free_buffer( "difference" );
    env.add_buffer(
        "difference", CL_MEM_READ_WRITE,
        sizeof( cl_float4 ) * raster_w * raster_h
    );

    status = clSetKernelArg(
        env.kern[ "ssim_diff" ], 0, sizeof( cl_mem ), env.buf( img1 )
    );
    clSuccess( status, "setting ssim_diff kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "ssim_diff" ], 1, sizeof( cl_mem ), env.buf( img2 )
    );
    clSuccess( status, "setting ssim_diff kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "ssim_diff" ], 2, sizeof( cl_mem ), env.buf( "difference" )
    );
    clSuccess( status, "setting ssim_diff kernel arg 2" );

    status = clSetKernelArg(
        env.kern[ "ssim_diff" ], 3, sizeof( cl_float ) * winsize * winsize, 0
    );
    clSuccess( status, "setting ssim_diff kernel arg 3" );

    status = clSetKernelArg(
        env.kern[ "ssim_diff" ], 4, sizeof( cl_uint ), &winsize
    );
    clSuccess( status, "setting ssim_diff kernel arg 4" );

    size_t global_size[] = { raster_w, raster_h };
    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "ssim_diff" ],
        2, NULL, global_size, NULL,
        0, NULL, &event
    );
    clSuccess( status, "launching ssim_diff kernel" );

    return "difference";
}

float
compute_difference(
    cl_env & env,
    diff_t alg,
    const char * golden,
    const char * imgname,
    unsigned int w,
    unsigned int h
) {
    cl_event events[ 2 ];
    start( "diff" );
    const char * bufname;
    switch ( alg ) {
        case DIFF_L2:
            bufname = l2_diff_cl( env, golden, imgname, w, h, events[ 0 ] );
            break;
        case DIFF_SSIM:
            bufname = ssim_diff_cl( env, golden, imgname, w, h, events[ 0 ] );
            break;
        default:
            std::cerr << "unsupported diff algorithm (" << alg << ")\n";
            exit( 3 );
    }

    events[ 1 ] = get_heatmap( env, bufname, imgname, w, h, events[ 0 ] );
    float avg = reduce( env, bufname, w, h, events[ 0 ] );

    cl_int status = clWaitForEvents( 1, events + 1 );
    clSuccess( status, "waiting for queue to finish" );
    stop( "diff" );

    for ( int i = 0; i < 2; ++i )
        clSuccess( clReleaseEvent( events[ i ] ), "releasing event" );

    return avg;
}

fftwf_complex *
get_dft_fftw(
    cl_env & env,
    const char * image,
    unsigned int w,
    unsigned int h,
    cl_event parent
) {
    fftwf_complex * buffer;
    fftwf_plan p;
    cl_int status;
    cl_event event;

    buffer = (fftwf_complex*) fftwf_malloc( sizeof( fftw_complex ) * w * h );
    status = clEnqueueReadBuffer(
        env.queue, *env.buf( image ), CL_FALSE,
        0, sizeof( cl_float2 ) * w * h,
        buffer,
        1, &parent, &event
    );
    clSuccess( status, "enqueueing read of dft" );

    clSuccess( clFlush( env.queue ), "flushing queue" );
    clSuccess( clWaitForEvents( 1, &event ), "waiting for events" );

    p = fftwf_plan_dft_2d( h, w, buffer, buffer, FFTW_FORWARD, FFTW_ESTIMATE );
    fftwf_execute( p );
    fftwf_destroy_plan( p );

    clSuccess( clReleaseEvent( event ), "releasing event" );

    return buffer;
}

#if USE_CLFFT
float2 *
get_dft_cl(
    cl_env & env,
    const char * image,
    unsigned int w,
    unsigned int h,
    cl_event parent
) {
    cl_int status;
    clfftPlanHandle plan;
    size_t length[] = { w, h };
    cl_event events[ 2 ];

    env.init_clFFT();

    status = clfftCreateDefaultPlan( &plan, env.context, CLFFT_2D, length );
    clSuccess( status, "creating clFFT plan" );

    status = clfftSetPlanPrecision( plan, CLFFT_SINGLE );
    clSuccess( status, "setting clFFT precision" );
    status = clfftSetLayout(
        plan, CLFFT_COMPLEX_INTERLEAVED, CLFFT_COMPLEX_INTERLEAVED
    );
    clSuccess( status, "setting clFFT layout" );
    status = clfftSetResultLocation( plan, CLFFT_INPLACE );
    clSuccess( status, "setting clFFT result location" );

    status = clfftEnqueueTransform(
        plan, CLFFT_FORWARD,
        1, &env.queue,
        1, &parent, events + 0,
        env.buf( image ), NULL, NULL
    );
    clSuccess( status, "enqueing transform" );

    float2 * dft = env.read_buffer< float2 >( image, events + 0, events + 1 );

    clSuccess( clFlush( env.queue ), "flushing queue" );
    clSuccess( clWaitForEvents( 1, events + 1 ), "waiting for events" );

    for ( size_t i = 0; i < 2; ++i )
        clSuccess( clReleaseEvent( events[ i ] ), "releasing event" );

    status = clfftDestroyPlan( &plan );
    clSuccess( status, "cleaning up plan" );

    return dft;
}
#endif

float
get_nyquist_energy(
    cl_env & env,
    const char * image,
    unsigned int & w,
    unsigned int & h,
    float rate,
    bool restore_image
) {
    cl_event events[ 3 ];
    int next_event;
    cl_int status;

    size_t global_size[] = { w, h };

    env.add_buffer( "dft", CL_MEM_READ_WRITE, sizeof( fftwf_complex ) * w * h );

    status = clSetKernelArg(
        env.kern[ "make_complex" ], 0, sizeof( cl_mem ), env.buf( image )
    );
    clSuccess( status, "setting make_complex kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "make_complex" ], 1, sizeof( cl_mem ), env.buf( "dft" )
    );
    clSuccess( status, "setting make_complex kernel arg 1" );

    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "make_complex" ],
        2, NULL, global_size, NULL,
        0, NULL, events
    );
    clSuccess( status, "launching make_complex kernel" );
    next_event = 1;

#if USE_CLFFT
    float2 * dft = (float2 *) get_dft_cl( env, "dft", w, h, events[ 0 ] );
#else
    float2 * dft = (float2 *) get_dft_fftw( env, "dft", w, h, events[ 0 ] );
#endif

    float total   = 0.0f;
    float outside = 0.0f;
    float peak    = 0.0f;
    unsigned int rx = w / rate * 0.5;
    unsigned int ry = h / rate * 0.5;
    for ( size_t y = 0, i = 0; y < h; ++y ) {
        for ( size_t x = 0; x < w; ++x, ++i ) {
            float amp =
                sqrt( dft[ i ].x * dft[ i ].x + dft[ i ].y * dft[ i ].y );
            total += amp;
            if ( ( ry <= y && y < h - ry ) || ( rx <= x && x < w - rx ) )
                outside += amp;
            peak = max( peak, log( sqrt( amp ) + 1 ) );
        }
    }

    if ( restore_image ) {
        w /= rate;
        h /= rate;

        env.free_buffer( image );
        env.add_buffer( image, CL_MEM_READ_WRITE, sizeof( float4 ) * w * h );

        events[ 1 ] = env.write_buffer( "dft", dft );

        status = clSetKernelArg(
            env.kern[ "dft_image" ], 0, sizeof( cl_mem ), env.buf( "dft" )
        );
        clSuccess( status, "setting dft_image kernel arg 0" );

        status = clSetKernelArg(
            env.kern[ "dft_image" ], 1, sizeof( cl_mem ), env.buf( image )
        );
        clSuccess( status, "setting dft_image kernel arg 1" );

        status = clSetKernelArg(
            env.kern[ "dft_image" ], 2, sizeof( cl_float ), &peak
        );
        clSuccess( status, "setting dft_image kernel arg 2" );

        status = clSetKernelArg(
            env.kern[ "dft_image" ], 3, sizeof( cl_float ), &w
        );
        clSuccess( status, "setting dft_image kernel arg 3" );

        status = clSetKernelArg(
            env.kern[ "dft_image" ], 4, sizeof( cl_float ), &h
        );
        clSuccess( status, "setting dft_image kernel arg 4" );

        status = clEnqueueNDRangeKernel(
            env.queue, env.kern[ "dft_image" ],
            2, NULL, global_size, NULL,
            1, events + 1, events + 2
        );
        clSuccess( status, "enqueing dft_image kernel" );
        next_event = 3;

        clSuccess( clFlush( env.queue ), "flushing queue" );
        clSuccess( clWaitForEvents( 1, events + 2 ), "waiting for events" );
    }

    for ( size_t i = 0; i < next_event; ++i )
        clSuccess( clReleaseEvent( events[ i ] ), "freeing event" );
    env.free_buffer( "dft" );

    return outside / total;
}

float
compute_similarity(
    cl_env & env,
    const char * golden,
    const char * imgname,
    unsigned int w,
    unsigned int h
) {
    size_t global_size[] = { w, h };
    cl_uint window = 7;
    unsigned int evt = 0;
    cl_int status;

    cl_event * events = new cl_event[ ( w - window ) * ( h - window ) + 3 ];

    start( "bidirsim" );
    env.add_buffer( "complete", CL_MEM_READ_WRITE, sizeof( cl_float ) * w * h );
    env.add_buffer( "cohere", CL_MEM_READ_WRITE, sizeof( cl_float ) * w * h );

    status = clSetKernelArg(
        env.kern[ "init_bidir_sim" ], 0, sizeof( cl_mem ), env.buf( "complete" )
    );
    clSuccess( status, "setting init_bidir_sim kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "init_bidir_sim" ], 1, sizeof( cl_mem ), env.buf( "cohere" )
    );
    clSuccess( status, "setting init_bidir_sim kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "init_bidir_sim" ], 2, sizeof( cl_uint ), &window
    );
    clSuccess( status, "setting init_bidir_sim kernel arg 2" );

    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "init_bidir_sim" ],
        2, NULL, global_size, NULL,
        0, NULL, events + evt
    );
    clSuccess( status, "enqueing init_bidir_sim kernel" );

    status = clSetKernelArg(
        env.kern[ "bidirectional_similarity" ], 0,
        sizeof( cl_mem ), env.buf( golden )
    );
    clSuccess( status, "setting bidirectional_similarity kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "bidirectional_similarity" ], 1,
        sizeof( cl_mem ), env.buf( imgname )
    );
    clSuccess( status, "setting bidirectional_similarity kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "bidirectional_similarity" ], 2,
        sizeof( cl_mem ), env.buf( "complete" )
    );
    clSuccess( status, "setting bidirectional_similarity kernel arg 2" );

    status = clSetKernelArg(
        env.kern[ "bidirectional_similarity" ], 3,
        sizeof( cl_mem ), env.buf( "cohere" )
    );
    clSuccess( status, "setting bidirectional_similarity kernel arg 3" );

    status = clSetKernelArg(
        env.kern[ "bidirectional_similarity" ], 6,
        sizeof( cl_uint ), &window
    );
    clSuccess( status, "setting bidirectional_similarity kernel arg 6" );

    for ( cl_uint i = 0; i < h - window; ++i ) {
        for ( cl_uint j = 0; j < w - window; ++j, ++evt ) {
            status = clSetKernelArg(
                env.kern[ "bidirectional_similarity" ], 4,
                sizeof( cl_uint ), &i
            );
            clSuccess( status, "setting bidirectional_similarity kernel arg 4" );

            status = clSetKernelArg(
                env.kern[ "bidirectional_similarity" ], 5,
                sizeof( cl_uint ), &j
            );
            clSuccess( status, "setting bidirectional_similarity kernel arg 5" );

            status = clEnqueueNDRangeKernel(
                env.queue, env.kern[ "bidirectional_similarity" ],
                2, NULL, global_size, NULL,
                1, events + evt, events + evt + 1
            );
            clSuccess( status, "enquing bidirectional_similarity kernel" );
        }
    }

    status = clSetKernelArg(
        env.kern[ "merge_bidir_sim" ], 0,
        sizeof( cl_mem ), env.buf( "complete" )
    );
    clSuccess( status, "setting merge_bidir_sim kernel arg 0" );

    status = clSetKernelArg(
        env.kern[ "merge_bidir_sim" ], 1,
        sizeof( cl_mem ), env.buf( "cohere" )
    );
    clSuccess( status, "setting merge_bidir_sim kernel arg 1" );

    status = clSetKernelArg(
        env.kern[ "merge_bidir_sim" ], 2,
        sizeof( cl_uint ), &window
    );
    clSuccess( status, "setting merge_bidir_sim kernel arg 2" );

    status = clEnqueueNDRangeKernel(
        env.queue, env.kern[ "merge_bidir_sim" ],
        2, NULL, global_size, NULL,
        1, events + evt, events + evt + 1
    );
    clSuccess( status, "enqueing merge_bidir_sim kernel" );
    evt += 1;

    events[ evt + 1 ] =
        get_heatmap( env, "complete", imgname, w, h, events[ evt ] );
    float avg = reduce( env, "complete", w, h, events[ evt ] );

    status = clWaitForEvents( 1, events + evt + 1 );
    clSuccess( status, "waiting for queue to finish" );

    stop( "bidirsim" );

    env.free_buffer( "complete" );
    env.free_buffer( "cohere" );

    for ( evt += 2; evt > 0; --evt )
        clSuccess( clReleaseEvent( events[ evt - 1 ] ), "releasing event" );
    delete[] events;


    return avg;
}

float dispatch( float * c, void * plotter );

class SimplexPlotter {
public:
    SimplexPlotter(
        cl_env & e,
        const std::vector< image_options > & img_opts,
        diff_t d,
        msaa_t a,
        unsigned nterms,
        float s
    )
        : env( e ),
          images( img_opts ),
          init( nterms, 0.0f ),
          iter( 0 ),
          terms( nterms ),
          seed( s ),
          alg( a ),
          diff( d )
    {
        if ( diff == NO_DIFF )
            diff = DIFF_L2;

        std::set< std::string > names;
        for ( size_t i = 0; i < images.size(); ++i ) {
            const char * fname = images[ i ].golden.c_str();
            if ( names.count( fname ) > 0 )
                continue;
            const char * key = load_golden(
                env,
                images[ i ].golden.c_str(),
                images[ i ].width,
                images[ i ].height
            );
            env.rename_buffer( key, fname );
            names.insert( fname );
        }
    }

    ~SimplexPlotter() {
        for ( size_t i = 0; i < images.size(); ++i ) {
            const char * fname = images[ i ].golden.c_str();
            env.free_buffer( fname );
        }
    }

    void setCoefficients( const std::vector< float > & cs ) {
        init = cs;
        init.resize( terms, 0.0f );
    }

    const char * solve() {
        env.free_buffer( cname );
        env.add_buffer( cname, CL_MEM_READ_ONLY, sizeof( cl_float ) * terms );

        float * c = new float[ terms ];
        std::copy( init.begin(), init.end(), c );
        if ( seed == 1.0f ) {
            for ( unsigned j = 0; j < terms; ++j )
                c[ j ] += 1.0f;
        } else if ( seed > 1.0f ) {
            std::default_random_engine gen( seed );
            std::normal_distribution< float > rand( 0.0f, 1.0f );
            for ( unsigned j = 0; j < terms; ++j )
                c[ j ] += rand( gen ) * c[ j ];
        }

        iter = 0;
        OptimizerSimplex * opt =
            new OptimizerSimplex( dispatch, this, c, terms );

        unsigned numIter = opt->solve();
        float * best = opt->pos();

        std::cout << opt->bestError() << ',' << numIter;
        for ( unsigned i = 0; i < terms; ++i )
            std::cout << ',' << best[ i ];
        std::cout << '\n';

        delete opt;
        delete[] c;

        return cname;
    }

    float getFitness( float * coefficients ) {
        update_coefficients( coefficients );

        float avg = 0.0f;
        unsigned npix = 0;
        for ( size_t i = 0; i < images.size(); ++i ) {
            const char * bufname;
            unsigned w    = images[ i ].width;
            unsigned h    = images[ i ].height;
            unsigned msaa = images[ i ].msaa;
            float zoom    = images[ i ].zoom;
            float plot_time;

            start( "plot" );
            bufname = plot_cl(
                env, w, h, 1, alg, msaa, zoom, false, cname, plot_time
            );
            stop( "plot" );

            avg += compute_difference(
                env, diff, images[ i ].golden.c_str(), bufname, w, h
            ) * w * h;
            npix += w * h;
        }

        iter += 1;
        if ( iter % 100 == 0 )
            std::cerr << "INFO: " << iter << " fitness evaluations (" << images.size() << " images)\n";
        return avg / npix;
    }

private:
    void update_coefficients( float * c ) {
        cl_event event = env.write_buffer( cname, c );
        cl_int status = clWaitForEvents( 1, &event );
        clSuccess( status, "waiting to initialize coefficient array" );
    }

    constexpr static const char * cname = "coeff";

    cl_env & env;
    std::vector< image_options > images;
    std::vector< float > init;
    unsigned iter, terms;
    float seed;
    msaa_t alg;
    diff_t diff;
};

float
dispatch( float * c, void * plotter )
{
    SimplexPlotter * simplex = ( SimplexPlotter * )( plotter );
    return simplex->getFitness( c );
}

int
main( int argc, char ** argv )
{
    options opts;

    start( "parse" );
    int i = parse_args( argc, argv, opts );
    if ( i != 0 )
        return i;
    stop( "parse" );

    cl_env env( opts.gpu );
    init_opencl( env, opts );

    const char * coeff  = NULL;
    if ( opts.simplex.terms > 0 ) {
        SimplexPlotter simplex(
            env, opts.images,
            opts.diff_alg, opts.msaa_alg,
            opts.simplex.terms, opts.seed
        );
        if ( opts.coefficients.size() > 0 )
            simplex.setCoefficients( opts.coefficients );
        coeff = simplex.solve();
    } else if ( opts.coefficients.size() > 0 ) {
        coeff = "coeff";
        env.free_buffer( coeff );
        env.add_buffer(
            coeff, CL_MEM_READ_ONLY,
            sizeof( cl_float ) * opts.coefficients.size()
        );
        cl_event event;
        event = env.write_buffer( coeff, &( opts.coefficients[ 0 ] ) );
        cl_int status = clWaitForEvents( 1, &event );
        clSuccess( status, "waiting to initialize coefficient array" );
    }

    for ( i = 0; i < opts.images.size(); ++i ) {
        const image_options & img = opts.images[ i ];
        unsigned w, h;
        const char * bufname = NULL;
        const char * golden = NULL;
        float plot_time;

        if ( opts.diff_alg != NO_DIFF ) {
            w = img.width;
            h = img.height;
            start( "plot" );
            bufname = plot_cl(
                env, w, h, img.subsamples,
                opts.msaa_alg, img.msaa, img.zoom,
                false, coeff, plot_time
            );
            stop( "plot" );
            if ( golden == NULL )
                golden = load_golden( env, img.golden.c_str(), w, h );
            float avg = compute_difference(
                env, opts.diff_alg, golden, bufname, w, h
            );
            printf( "%g ", avg );
        }

        if ( opts.similarity ) {
            w = img.width;
            h = img.height;
            start( "plot" );
            bufname = plot_cl(
                env, w, h, img.subsamples,
                opts.msaa_alg, img.msaa, img.zoom,
                false, coeff, plot_time
            );
            stop( "plot" );

            if ( golden == NULL )
                golden = load_golden( env, img.golden.c_str(), w, h );
            float avg = compute_similarity( env, golden, bufname, w, h );
            printf( "%g ", avg );
        }

        if ( opts.aliasing ) {
            w = img.width;
            h = img.height;
            start( "plot" );
            bufname = plot_cl(
                env, w, h, img.zoom,
                opts.msaa_alg, img.msaa, img.zoom,
                true, coeff, plot_time
            );
            stop( "plot" );

            start( "dft" );
            float energy = get_nyquist_energy(
                env, bufname, w, h, img.zoom,
                !opts.similarity && img.outfile != ""
            );
            printf( "%g ", energy );
            stop( "dft" );
        }

        if ( img.outfile != "" ) {
            if ( bufname == NULL ) {
                w = img.width;
                h = img.height;
                start( "plot" );
                bufname = plot_cl(
                    env, w, h, img.subsamples,
                    opts.msaa_alg, img.msaa, img.zoom,
                    false, coeff, plot_time
                );
                stop( "plot" );
            }
            if ( img.upsample > 1 )
                upsample( env, bufname, w, h, img.upsample );

            start( "copy" );
            float4 * raster = env.read_buffer< float4 >( bufname, NULL, NULL );
            clSuccess( clFinish( env.queue ), "finishing queue" );
            stop( "copy" );

            start( "write" );
            writePPM( img.outfile.c_str(), raster, w, h );
            stop( "write" );

            delete[] raster;
        }

        if ( opts.fps )
            printf( "%g ", 1.0f / plot_time );

        printf( "\n" );
    }

#if DO_TIMING
    std::map< std::string, unsigned long >::iterator it;
    float total = 0.0f;
    for ( it = timings.begin(); it != timings.end(); ++it )
        total += (float)it->second * 1e-9;
    for ( it = timings.begin(); it != timings.end(); ++it )
        printf(
            "%s time\t= %1.4f\t(%5.2f%%)\n",
            it->first.c_str(), it->second * 1e-9f, 
            it->second * 1e-7f / total
        );
#endif

    return 0;
}

