#!/usr/bin/python

from collections import OrderedDict
from contextlib import contextmanager
from cStringIO import StringIO
import hashlib
from optparse import OptionGroup, OptionParser, OptionValueError, Values
import os
import random
import shelve
import shlex
import shutil
from subprocess import check_call, Popen, PIPE, CalledProcessError
import sys

from utils import *

def lex_args( stream, fname ):
    lexer = shlex.shlex( stream, fname, True )
    lexer.whitespace_split = True
    return list( lexer )

parser = OptionParser( usage = "%prog [options] [config-file...]" )
parser.add_option(
    "--prog", metavar = "source.c", help = "program source file to process"
)
parser.add_option(
    "--initial", metavar = "genomes",
    help = "comma-separated genomes to start from"
)
parser.add_option(
    "--keep-opencl", action = "store_true",
    help = "save every variant OpenCL file as XXXXXX.cl"
)
parser.add_option(
    "--keep-sources", action = "store_true",
    help = "save every variant source file as XXXXXX.c"
)
parser.add_option(
    "--search", metavar = "strategy", choices = ( "brute", "ga", "hill" ),
    help = "the search algorithm to use (brute, ga, or hill)"
)
parser.add_option(
    "--test-command", help = "shell command to evaluate variant fitness"
)
parser.add_option(
    "--save", metavar = "dirname",
    help = "write the members of the final pareto frontier to the named directory"
)
parser.add_option(
    "--seed", metavar = "N", type = int, default = 0,
    help = "initialize the random seed for the search"
)
parser.add_option(
    "--stats", action = "store_true",
    help = "show stats on where time is being spent"
)
parser.add_option(
    "--cache-file", metavar = "file",
    help = "cache variant fitness evaluations to this file"
)

group = OptionGroup( parser, "Genetic Search" )
group.add_option(
    "--crossp", metavar = "prob", type = float, default = 0.5,
    help = "probability between 0.0 and 1.0 of applying crossover"
)
group.add_option(
    "--generations", metavar = "N", type = int, default = 10,
    help = "number of generations to run"
)
group.add_option(
    "--popsize", metavar = "N", type = int, default = 40,
    help = "size of the population"
)
group.add_option(
    "--tourney-size", metavar = "N", type = int, default = 2,
    help = "number of participants in each tournament"
)
parser.add_option_group( group )

group = OptionGroup( parser, "Hill-Climbing Search" )
group.add_option(
    "--restarts", metavar = "N", type = int, default = 10,
    help = "number of random restarts to use"
)
parser.add_option_group( group )

options, args = parser.parse_args()
config_opts = Values()
while len( args ) > 0:
    arg = args.pop( 0 )
    with open( arg ) as fh:
        config_args = lex_args( fh, arg )
    args.extend( parser.parse_args( config_args, config_opts )[ 1 ] )
for key in config_opts.__dict__:
    if not key in options.__dict__ or options.__dict__[ key ] is None:
        options.__dict__[ key ] = config_opts.__dict__[ key ]

if not hasattr( options, "search" ):
    options.search = "brute"

for flag in [ "prog", "test_command" ]:
    if not hasattr( options, flag ) or getattr( options, flag ) is None:
        flag = flag.replace( "_", "-" )
        print >>sys.stderr, "ERROR: --%s is a required flag\n" % flag
        parser.print_help()
        exit( 1 )

random.seed( options.seed )

shader = os.path.splitext( os.path.basename( options.prog ) )[ 0 ]

if options.cache_file is None:
    options.cache_file = shader + ".cache"
    
root = os.path.abspath( os.path.dirname( sys.argv[ 0 ] ) )
compilesh = os.path.join( root, "compile.sh" )
testpy    = os.path.join( root, "test.py" )
transform = os.path.join( root, "transform" )

check_executable( [ compilesh ] )
check_executable( [ testpy, "--help" ] )
check_executable( [ transform, "--help" ] )

def dcmp( a, b ):
    return a < b

def dmax( iterable, key = lambda x: x ):
    iterable = map( lambda x: ( x, key( x ) ), iterable )
    best, _ = reduce(
        lambda (x1, k1), (x2, k2): (x1, k1) if dcmp( k1, k2 ) else (x2, k2),
        iterable
    )
    return best

def log( *args ):
    msg = " ".join( map( str, args ) )
    print >>logfile, msg
    logfile.flush()
    try:
        os.fsync( logfile.fileno() )
    except: pass

    print msg
    sys.stdout.flush()
    try:
        os.fsync( sys.stdout.fileno() )
    except: pass

def keeper( keep, suffix ):
    if keep:
        @contextmanager
        def get_fname( counter = [ 0 ] ):
            name = "%06d%s" % ( counter[ 0 ], suffix )
            counter[ 0 ] += 1
            yield name
    else:
        @contextmanager
        def get_fname():
            with mktemp( suffix ) as name:
                yield name
    return get_fname

source_name = keeper( options.keep_sources, ".c" )
opencl_name = keeper( options.keep_sources, ".cl" )

class Genome:
    def __init__( self, stmts, initial = False ):
        self.genes = OrderedDict()
        for stmt in stmts:
            self.genes[ stmt ] = initial

    def __hash__( self ):
        return reduce(
            lambda h, (s, b): ( h * 31 + hash( b ) ) & 0xffffffff,
            self.genes.iteritems(),
            0
        )

    def __eq__( self, other ):
        return self.genes == other.genes

    def __str__( self ):
        return "".join( map( lambda b: "1" if b else "0", self.genes.values() ) )

    def __copy( self ):
        copy = Genome( list() )
        copy.genes = self.genes.copy()
        return copy

    def onePointCrossover( self, other ):
        c1 = self.__copy()
        c2 = other.__copy()
        pivot = random.randrange( len( self.genes ) )
        for i, stmt in zip( range( pivot ), self.genes ):
            c1.genes[ stmt ], c2.genes[ stmt ] = c2.genes[ stmt ], c1.genes[ stmt ]
        return c1, c2

    def flip( self, stmt ):
        copy = self.__copy()
        copy.genes[ stmt ] = not copy.genes[ stmt ]
        return copy

    def getTransform( self ):
        transform = list()
        for s, b in self.genes.iteritems():
            if b:
                transform.append( s )
        return ",".join( transform )

    def mutate( self ):
        stmt = random.choice( self.genes.keys() )
        return self.flip( stmt )

    def randomize( self ):
        copy = self.__copy()
        for stmt in copy.genes:
            copy.genes[ stmt ] = random.choice( [ True, False ] )
        return copy

class Variant:
    cache = shelve.open( options.cache_file, protocol = 2 )

    @staticmethod
    def getKey( genome ):
        digest = hashlib.md5( options.test_command ).hexdigest()
        return digest + "," + str( genome )

    @classmethod
    def getVariant( cls, genome ):
        key = cls.getKey( genome )
        if not key in cls.cache:
            cls.cache[ key ] = cls( genome )
        return cls.cache[ key ]

    def __init__( self, genome ):
        self.fitness = None
        self.stmts   = None
        self.genome  = genome

    def __eq__( self, other ):
        return self.genome == other.genome

    def __str__( self ):
        return str( self.genome )

    def __get_test_command( self, **kwargs ):
        cmd = lex_args( StringIO( options.test_command ), "test-command" )
        cmd = map( lambda arg: kwargs.get( arg, arg ), cmd )
        return cmd

    def getGenome( self ):
        return self.genome

    def build( self, fname ):
        with source_name() as source:
            with open( "/dev/null", 'w' ) as fh:
                with timer( "compile" ):
                    self.save( source )
                    check_call( [ compilesh, shader, fname, source ] )

    def save( self, fname ):
        targets = self.genome.getTransform()
        cmd = [
            transform, ppfile, shader,
                "--transform", targets,
                "--change-indicator", "transformed",
                "-o", fname
        ]
        with open( "/dev/null", 'w' ) as fh:
            check_call( cmd, stderr = fh )

    def getstatements( self ):
        if self.stmts is None:
            cmd = [ transform, ppfile, shader, "--get-statements" ]
            p = Popen( cmd, stdout = PIPE, stderr = PIPE )
            lines, errs = p.communicate()[ :2 ]
            if p.returncode != 0:
                log( errs )
                exit( p.returncode )
            self.stmts = list()
            for line in lines.splitlines():
                self.stmts.append( line )
            self.__class__.cache[ self.getKey( self.genome ) ] = self
        return self.stmts

    def __test( self ):
        if self.fitness is None:
            with opencl_name() as clfile:
                self.build( clfile )

                with mktemp( "" ) as fitnessfile:
                    with timer( "test" ):
                        cmd = self.__get_test_command(
                            __SHADER__     = shader,
                            __BINARY_FILE__  = clfile,
                            __FITNESS_FILE__ = fitnessfile
                        )
                        try:
                            check_call( cmd )
                        except CalledProcessError as e:
                            shutil.copyfile( clfile, "bad.cl" )
                            log( "failed program is at bad.cl" )
                            raise e

                    self.fitness = list()
                    with open( fitnessfile ) as fh:
                        for line in fh:
                            for i, value in enumerate( line.split() ):
                                while len( self.fitness ) <= i:
                                    self.fitness.append( list() )
                                self.fitness[ i ].append( float( value ) )
            self.fitness = map( lambda c: sum( c ) / len( c ), self.fitness )
            self.__class__.cache[ self.getKey( self.genome ) ] = self

        return self.fitness

    def test( self ):
        result = self.__test()
        log( "   ", self, ":", " ".join( map( str, result ) ) )
        return result

    def dominates( self, other ):
        x = self.__test()
        y = other.__test()

        better = False
        for a, b in zip( x, y ):
            if dcmp( a, b ):
                better = True
            elif dcmp( b, a ):
                return False
        return better

class Frontier:
    def __init__( self ):
        self.members = list()

    def __iter__( self ):
        return iter( map(
            lambda (x,_): x,
            sorted( self.members, key = lambda (_,y): y )
        ) )

    def add( self, v, key ):
        new_members = list()
        ejected = list()
        for vold, keyold in self.members:
            if vold.dominates( v ) or vold == v:
                return [ v ]
            elif v.dominates( vold ):
                ejected.append( vold )
            else:
                new_members.append( ( vold, keyold ) )
        else:
            new_members.append( ( v, key ) )
            self.members = new_members

        return ejected

    def extend( self, other ):
        ejected = list()
        for elem, key in other.members:
            ejected.extend( self.add( elem, key ) )
        return ejected

    def save( self, dirname ):
        if not os.path.exists( dirname ):
            os.makedirs( dirname )
        elements = sorted( self.members, key = lambda (_,y): y )
        for i, (elem, key) in enumerate( elements ):
            fname = os.path.join( dirname, "%06d.c" % i )
            elem.save( fname )
            with open( os.path.join( dirname, "%06d.fitness" % i ), 'w' ) as fh:
                try:
                    print >>fh, ",".join( map( str, key ) )
                except Exception as e:
                    print >>fh, key
            with open( os.path.join( dirname, "%06d.genome" % i ), 'w' ) as fh:
                print >>fh, elem

@contextmanager
def preprocess():
    with mktemp( ".c" ) as ppfile:
        check_call( [
            "gcc", "-E", options.prog, "-DFILTER_MODE_TENT", "-o", ppfile
        ] )
        yield ppfile

def build_initial_pop( stmts ):
    if options.initial is None:
        return list()

    pop = list()
    for rep in options.initial.split( "," ):
        genome = Genome( stmts, False )
        for bit, stmt in zip( rep, stmts ):
            if bit == "1":
                genome = genome.flip( stmt )
        pop.append( genome )
    return pop

def brute_force_gen( stmts, index ):
    if index < 0:
        yield Genome( stmts )
    else:
        for genome in brute_force_gen( stmts, index - 1 ):
            yield genome
            yield genome.flip( stmts[ index ] )

def brute_force( stmts ):
    log( "brute force search begins" )
    best = Frontier()
    for genome in brute_force_gen( stmts, len( stmts ) - 1 ):
        v = Variant.getVariant( genome )
        best.add( v, v.test() )
    return best

def tournament( pop, key ):
    tourney = list()
    while len( tourney ) < options.tourney_size:
        tourney.append( random.choice( pop ) )
    return dmax( tourney, key = key )

def add_to_pop( pop, g ):
    key = Variant.getVariant( g ).test()
    pop.append( ( g, key ) )

def offspring( pop, key ):
    p1 = tournament( pop, key )[ 0 ]
    p2 = tournament( pop, key )[ 0 ]
    if random.random() < options.crossp:
        c1, c2 = p1.onePointCrossover( p2 )
    else:
        c1, c2 = p1, p2
    return c1.mutate(), c2.mutate()

def classic_ga( pop ):
    key = lambda (_,k): k

    best = dmax( pop, key = key )

    for gen in range( options.generations ):
        log( "generation %d starts" % gen )
        next_pop = list()
        while len( next_pop ) < options.popsize:
            c1, c2 = offspring( pop, key )
            add_to_pop( next_pop, c1 )
            add_to_pop( next_pop, c2 )
        if dcmp( key( best ), key( dmax( next_pop, key = key ) ) ):
            next_pop.append( best )
        pop = next_pop

    return pop

def nsga_ii( pop ):
    class NSGAitem:
        def __init__( self, genome, fitness ):
            self.genome  = genome
            self.fitness = fitness
            self.S       = list()
            self.n       = 0
            self.dist    = 0
    key = lambda (_,k): k

    for gen in range( options.generations ):
        log( "NSGA generation", gen, "begins" )
        new_pop = list()
        while len( new_pop ) < options.popsize:
            c1, c2 = offspring( pop, key )
            add_to_pop( new_pop, c1 )
            add_to_pop( new_pop, c2 )
        pop.extend( new_pop )

        if gen == options.generations - 1:
            break

        # 3.2 Non-dominated sort

        log( "NSGA: computing Pareto frontiers" )
        pop = map( lambda (g,f): NSGAitem( g, f ), pop )

        front = list()
        for p in pop:
            vp = Variant.getVariant( p.genome )
            for q in pop:
                vq = Variant.getVariant( q.genome )
                if vp.dominates( vq ):
                    p.S.append( q )
                elif vq.dominates( vp ):
                    p.n += 1
            if p.n == 0:
                front.append( p )

        fronts = list()
        while len( front ) > 0:
            fronts.append( front )
            new_front = list()
            for p in front:
                for q in p.S:
                    q.n -= 1
                    if q.n == 0:
                        new_front.append( q )
                p.S = list()
            front = new_front
        log( "NSGA: %d fronts found" % len( fronts ) )

        # 3.3 Crowding distance

        remaining = options.popsize
        new_pop = list()
        while len( fronts[ 0 ] ) <= remaining:
            remaining -= len( fronts[ 0 ] )
            for p in fronts.pop( 0 ):
                new_pop.append( ( p.genome, p.fitness ) )
        log( "NSGA: computing crowding for %d of %d elements"
             % ( remaining, len( fronts[ 0 ] ) ) )

        front = fronts[ 0 ]
        inf = float( "inf" )
        nObjectives = len( front[ 0 ].fitness )
        for i in range( nObjectives ):
            front.sort( key = lambda p: p.fitness[ i ] )
            lo = front[  0 ].fitness[ i ]
            hi = front[ -1 ].fitness[ i ]
            front[  0 ].dist = inf
            front[ -1 ].dist = inf
            for j in range( 1, len( front ) - 1 ):
                l = front[ j - 1 ].fitness[ i ]
                r = front[ j + 1 ].fitness[ i ]
                front[ j ].dist += ( r - l ) / ( hi - lo )
        if dcmp( 0, 1 ):
            front.sort( key = lambda p: p.dist )
        else:
            front.sort( key = lambda p: - p.dist )
        while len( new_pop ) < options.popsize:
            p = front.pop( -1 )
            new_pop.append( ( p.genome, p.fitness ) )

        pop = new_pop

    return pop

def genetic_algorithm( stmts ):
    log( "generating initial population" )
    pop = list()
    for genome in build_initial_pop( stmts ):
        add_to_pop( pop, genome )
    if len( pop ) == 0:
        add_to_pop( pop, Genome( stmts, False ) )
        add_to_pop( pop, Genome( stmts, True ) )
    while len( pop ) < options.popsize:
        add_to_pop( pop, pop[ 0 ][ 0 ].randomize() )

    if len( pop[ 0 ][ 1 ] ) == 1:
        pop = classic_ga( pop )
    else:
        pop = nsga_ii( pop )

    frontier = Frontier()
    for genome, key in pop:
        v = Variant.getVariant( genome )
        frontier.add( v, key )
    return frontier

def hill_climb( stmts ):
    log( "hill climbing search begins" )

    best = Frontier()
    for i in range( options.restarts ):
        log( "hill climbing restart: %d" % i )
        fitness = dict()
        def get_fitness( genome, fitness = fitness ):
            if genome not in fitness:
                fitness[ genome ] = Variant.getVariant( genome ).test()
            return fitness[ genome ]

        frontier = Frontier()
        if options.initial is not None:
            pop = build_initial_pop( stmts )
            options.initial = None
            for genome in pop:
                frontier.add(
                    Variant.getVariant( genome ),
                    get_fitness( genome )
                )
        else:
            genome = Genome( stmts, False ).randomize()
            frontier.add( Variant.getVariant( genome ), get_fitness( genome ) )

        do_work = True
        j = 0
        while do_work:
            size = len( list( frontier ) )
            log( "hill climbing step: %d (%d in front)" % ( j, size ) )
            do_work = False
            for variant in frontier:
                genome = variant.getGenome()
                for stmt in stmts:
                    candidate = genome.flip( stmt )
                    variant = Variant.getVariant( candidate )
                    result = frontier.add( variant, get_fitness( candidate ) )
                    if not variant in result:
                        do_work = True

            j += 1
        best.extend( frontier )
    return best

def main():
    global ppfile
    with preprocess() as ppfile:
        tmp = Variant( dict() )
        stmts = tmp.getstatements()
        del tmp

        best = {
            "brute": brute_force,
            "ga":    genetic_algorithm,
            "hill":  hill_climb,
        }[ options.search ]( stmts )

        if options.save is not None:
            best.save( options.save )

        log( "Final frontier:" )
        for v in best:
            v.test()

with open( "search.%s.%d" % ( shader, options.seed ), 'w' ) as logfile:
    try:
        main()
    finally:
        if options.stats:
            show_stats( logfile )
            show_stats( sys.stdout )

