#!/usr/bin/python2

from collections import defaultdict
from contextlib import contextmanager
from math import sqrt
from optparse import OptionParser
import os
from pipes import quote
from subprocess import call, check_call, Popen, PIPE, STDOUT
import sys
import tempfile
import time

from utils import *

parser = OptionParser(
    usage = "%prog [options] shader shader-file fitness-file size"
)
parser.add_option(
    "--imagew", metavar = "weight", type = float, default = 0.0,
    help = "weight to apply to the difference between rendered images."
)
parser.add_option(
    "--aliasw", metavar = "weight", type = float, default = 0.0,
    help = "weight to apply to energy contributing to aliasing"
)
parser.add_option(
    "--similw", metavar = "weight", type = float, default = 0.0,
    help = "weight to apply to similarity metric"
)
parser.add_option(
    "--timew", metavar = "weight", type = float, default = 0.0,
    help = "weight to apply to rendering time"
)
parser.add_option(
    "--coeff", metavar = "c1,c2", help = "comma-separated list of coefficients"
)
parser.add_option(
    "--diff", metavar = "func", default = "l2",
    help = "difference algorithm to apply to images"
)
parser.add_option(
    "--gpu", metavar = "N", type = int,
    help = "GPU device number to use for OpenCL computations"
)
parser.add_option(
    "--gray", action = "store_true",
    help = "evaluate grayscale images instead of RGB images"
)
parser.add_option(
    "--low-var", metavar = "p", type = float,
    help = "repeat timing measurements to get variance under p times the mean"
)
parser.add_option(
    "--multifit", action = "store_true",
    help = "report multi-dimensional fitness instead of a single fitness value"
)
parser.add_option(
    "--msaa", metavar = "N", type = int,
    help = "evaluate the performance of taking multiple samples with the shader"
)
parser.add_option(
    "--perspective", action = "store_true",
    help = "render perspective instead of orthographic"
)
parser.add_option(
    "--timeout", metavar = "secs", type = int, default = 10,
    help = "max number of seconds to allow tested application to run"
)
parser.add_option(
    "--stats", action = "store_true",
    help = "show stats on where time is spent during testing"
)
parser.add_option(
    "--zoom", metavar = "x", default = "1",
    help = "comma-separated list of zoom values to render"
)
options, args = parser.parse_args()

if len( args ) < 4:
    parser.print_help()
    exit()
options.zoom = map( int, options.zoom.split( "," ) )

root = os.path.abspath( os.path.dirname( sys.argv[ 0 ] ) )
if "LOG_SCRIPT_ARGS" in os.environ:
    with open( os.path.join( root, "..", scriptlog ), 'a' ) as fh:
        print >>fh, " ".join( map( quote, sys.argv ) )

shader, clfile, fname, size = args[ :4 ]
clfile = os.path.abspath( clfile )
compilesh = os.path.join( root, "compile.sh" )
renderer  = os.path.join( root, "renderer" )

# checking for executables can add significantly to run time. Don't do it.

#check_executable( [ "diff-img", "-h" ] )
#check_executable( [ "convert", "-h" ] )
#check_executable( [ compilesh ] )
#check_executable( [ renderer, "-h" ] )

no_orig = True
def make_orig():
    global no_orig

    orig = os.path.join( root, "..", shader + ".cl" )
    if no_orig:
        with mktemp( ".c" ) as preprocessed:
            check_call( [
               "gcc", "-E", os.path.join( root, "..", "src", shader + ".c" ),
                    "-DFILTER_MODE_NONE",
                    "-o", preprocessed
            ] )
            with timer( "compile" ):
                check_call( [ compilesh, shader, orig, preprocessed ] )
        no_orig = False
    return orig

def run_exe( global_flags, img_flags, ss ):
    cmd = [ renderer, clfile ] + global_flags
    if options.gpu is not None:
        cmd += [ "--gpu", str( options.gpu ) ]
    if options.gray:
        cmd += [ "--gray" ]
    if options.perspective:
        cmd += [ "--perspective" ]
    if options.msaa is not None and options.msaa <= 16:
        cmd += [ "--sampling", "gaussian" ]
    if options.coeff is not None:
        cmd += [ "--coeff", options.coeff ]
    for scale in ss:
        cmd += [ "-i",
            "--subsamples", "1",
            "--zoom", str( scale ),
            "--size", size,
        ]
        if options.msaa is not None:
            cmd += [ "--msaa", str( options.msaa ) ]
        cmd += img_flags( scale )

    with timer( "render image" ):
        p = Popen(
            [ "timeout", str( options.timeout ) ] + cmd,
            stdout = PIPE
        )
        lines = p.communicate()[ 0 ].splitlines()

    if len( lines ) != len( ss ):
        with open( fname, 'w' ) as fh:
            print >>fh, 0
            exit( 2 )

    vals = list()
    for line in lines:
        for i, val in enumerate( map( float, line.split() ) ):
            while len( vals ) <= i:
                vals.append( list() )
            vals[ i ].append( val )

    return vals

def get_fitness( ss ):
    global_flags = list()
    if options.imagew > 0:
        global_flags += [ "--diff", options.diff.strip( '-' ).lower() ]
    if options.aliasw > 0:
        global_flags += [ "--aliasing" ]
    if options.similw > 0:
        global_flags += [ "--similarity" ]
    if options.timew > 0:
        global_flags += [ "--fps" ]

    if options.imagew > 0 or options.similw > 0:
        def get_reference( scale ):
            return [
                "-g", get_golden( shader, size, scale, options.gray, options.perspective )
            ]
    else:
        get_reference = lambda scale: list()
    
    vals = run_exe( global_flags, get_reference, ss )

    if options.timew > 0 and options.low_var is not None:
        global_flags = [ "--fps" ]
        write_null = lambda scale: [ "-o", "/dev/null" ]

        remaining = 10
        n  = 0
        m = [ 0 ] * len( ss )
        while remaining > 0:
            times = run_exe( global_flags, write_null, ss * remaining )
            for i in range( 0, len( times[ -1 ] ), len( ss ) ):
                n = n + 1
                for j in range( i, i + len( ss ) ):
                    delta = times[ -1 ][ j ] - vals[ -1 ][ j - i ]
                    vals[ -1 ][ j - i ] += delta / n
                    m[ j - i ] += delta * ( times[ -1 ][ j ] - vals[ -1 ][ j - i ] )
            for i, (mean, var) in enumerate( zip( vals[ -1 ], m ) ):
                ratio = sqrt( var / ( n - 1 ) ) / mean
                if ratio > options.low_var:
                    break
            else:
                remaining = 0

    if options.timew > 0:
        vals[ -1 ] = map( lambda v: 1/v, vals[ -1 ] )

    return map( lambda vs: sum( vs ), vals )

def invert( val, zero ):
    #return zero if val == 0.0 else 1.0 / val
    return val

# Pre-generate the golden images for image-based fitness values

ss = list()
cmd = list()
for scale in options.zoom:
    golden = get_golden( shader, size, scale, options.gray, options.perspective )
    if not os.path.exists( os.path.dirname( golden ) ):
        os.makedirs( os.path.dirname( golden ) )
    if not os.path.exists( golden ):
        cmd += [
            "-i", "-o", golden,
            "--msaa", "2048",
            "--zoom", str( scale ),
            "--size", size,
        ]

    ss.append( scale )
if len( cmd ) > 0:
    orig = make_orig()
    cmd = [ renderer, orig ] + cmd
    if options.gpu is not None:
        cmd.extend( [ "--gpu", str( options.gpu ) ] )
    if options.gray:
        cmd.append( "--gray" )
    if options.perspective:
        cmd.append( "--perspective" )
    with timer( "golden" ):
        check_call( cmd )
if not no_orig:
    os.remove( orig )

# Measure the fitness

try:
    fitness = list()
    vals = get_fitness( ss )
    if options.imagew > 0:
        fitness.append( vals.pop( 0 ) * options.imagew )
    if options.similw > 0:
        fitness.append( vals.pop( 0 ) * options.similw )
    if options.aliasw > 0:
        fitness.append( vals.pop( 0 ) * options.aliasw )
    if options.timew > 0:
        fitness.append( vals.pop( 0 ) * options.timew )
    fitness += vals

    # Fix the fitness so that larger values are preferred

    if options.multifit:
        afit  = 0 if options.aliasw > 0 else -3
        afit += 1 if options.similw > 0 else 0
        afit += 1 if options.imagew > 0 else 0
        for i in range( len( fitness ) ):
            if i == afit:
                fitness[ i ] = invert( fitness[ i ], 0 )
            else:
                fitness[ i ] = invert( fitness[ i ], sys.float_info.max )
        with open( fname, 'w' ) as fh:
            print >>fh, " ".join( map( str, fitness ) )
    else:
        fitness = invert( sum( fitness ), sys.float_info.max )
        with open( fname, 'w' ) as fh:
            print >>fh, fitness

    # exit with failure so that the search doesn't immediately end

    exit( 0 )
finally:
    if options.stats:
        show_stats()

