open Cil
open Utils

let inlineonly = ref false
let genome     = ref None
let get_stmts  = ref false
let changevar  = ref ""
let outfile    = ref ""
let skipfiles  = ref []
let tstmts     = ref None

let append v x = v := x :: !v
let opt v x = v := Some(x)

let options = [
  ("--debug", Arg.Set(debug), " show debugging information");
  ("--change-indicator", Arg.Set_string(changevar),
   "var variable to set to true if the source was changed");
  ("--do-not-print", Arg.String(append skipfiles),
   "header Do not print definitions from the named header file");
  ("--genome", Arg.String(opt genome),
   "genome Bit string indicating which statements to transform");
  ("--inline-only", Arg.Set(inlineonly),
   " only perform inlining, do not band-limit");
  ("--get-statements", Arg.Set(get_stmts),
   " print the list of statements that may be transformed");
  ("--transform", Arg.String(opt tstmts),
   "n1,n2 comma-separated list of statement numbers to transform");
  ("-o", Arg.Set_string(outfile), "file Write transformed file to a file");
]

let func_defs = Hashtbl.create 257

let flip f a b = f b a
let is_none = function None -> true | _ -> false

module Transform = Transformer.Make(Widthexp)
(*
module UsedVarTransfer =
struct
  let name = "Used Variables"

  let debug = debug

  type t = IntSet.t

  let pretty () t =
    seq (text " ") num (IntSet.elements t)

  let copy t = t

  let stmtStartData = Inthash.create 257

  class getUseVisitor uses =
    object
      inherit nopCilVisitor

      method vvrbl vi =
        uses := IntSet.add vi.vid !uses;
        DoChildren
    end

  let initStartData stmts formals =
    Inthash.clear stmtStartData;
    List.iter (fun stmt ->
      let init = ref IntSet.empty in

      Inthash.add stmtStartData stmt.sid !init
    ) stmts

  let funcExitData = 

  let combineStmtStartData s ~old t =

  let combineSuccessors lhs rhs =

  let doInstr i data =
    match i with
    | Set(lv, e, _) ->
      let uses = 
  Dataflow.Default

  let doStmt _ _ = Dataflow.Default

  let filterStmt _ = true
end

let UsedVarAnalysis = Dataflow.BackwardsDataFlow(UsedVarTransfer)
*)

class flattenVisitor =
object (self)
  inherit nopCilVisitor

  val mutable current_fd = None

  method private make_assignment e =
    match current_fd, e with
    | None, _ -> e
    | _, Const(_) -> e
    | _, Lval(_)  -> e
    | Some(fd), _ ->
      let lv = var (makeTempVar fd (typeOf e)) in
      self#queueInstr [Set(lv, e, locUnknown)];
      Lval(lv)

  method vfunc fd =
    let old_fd = current_fd in
    current_fd <- Some(fd);
    ChangeDoChildrenPost(fd, fun fd -> current_fd <- old_fd; fd)

  method vblock b =
    ChangeDoChildrenPost(b, fun b ->
      let stmts =
        List.map (fun stmt ->
          match stmt.skind with
          | Instr([]) -> [stmt]
          | Instr(first::rest) ->
            stmt.skind <- Instr([first]);
            stmt :: List.map (fun i -> mkStmtOneInstr i) rest
          | Block(b') when (List.length b'.battrs) = 0 ->
            b'.bstmts
          | _ -> [stmt]
        ) b.bstmts
      in
      let stmts = List.flatten stmts in
        { b with bstmts = stmts }
    )

  method vstmt s =
    ChangeDoChildrenPost(s, fun s ->
      let skind =
        match s.skind with
        | Return(Some(e), loc)  -> Return(Some(self#make_assignment e), loc)
        | If(e, th, el, loc)    -> If(self#make_assignment e, th, el, loc)
        | Switch(e, b, ss, loc) -> Switch(self#make_assignment e, b, ss, loc)
        | _ -> s.skind
      in
      s.skind <- skind;
      s
    )

  method vinst i =
    ChangeDoChildrenPost([i], function
      | [Call(rv, f, args, loc)] ->
        let args = List.map self#make_assignment args in
        [Call(rv, f, args, loc)]
      | is -> is
    )

  method vexpr e =
    ChangeDoChildrenPost(e, function
      | SizeOfE(e')          -> SizeOfE(self#make_assignment e')
      | AlignOfE(e')         -> AlignOfE(self#make_assignment e')
      | CastE(t, e')         -> CastE(t, self#make_assignment e')
      | UnOp(op, e', t)      -> UnOp(op, self#make_assignment e', t)
      | BinOp(op, e1, e2, t) ->
        let e1 = self#make_assignment e1 in
        let e2 = self#make_assignment e2 in
        BinOp(op, e1, e2, t)
      | e -> e
    )

  method vlval lv =
    ChangeDoChildrenPost(lv, function
      | Mem(e), off -> Mem(self#make_assignment e), off
      | lv -> lv
    )
end

class numVisitor =
object
  inherit nopCilVisitor

  val mutable next = 1

  method vstmt s =
    s.sid <- next;
    next <- next + 1;
    DoChildren
end

class removeGotoNextLine =
object
  inherit nopCilVisitor

  method vblock b =
    let stmts, pending =
      List.fold_left (fun (stmts,pending)  s ->
        let stmts =
          match pending with
          | Some(_, sid) when sid = s.sid -> stmts
          | Some(s', _) -> s'::stmts
          | _ -> stmts
        in
          match s.skind with
          | Goto(s', _) -> stmts, Some(s, !s'.sid)
          | _ -> s::stmts, None
      ) ([],None) b.bstmts
    in
    let stmts =
      match pending with
      | Some(s', _ ) -> s'::stmts
      | _ -> stmts
    in
      b.bstmts <- List.rev stmts;
      DoChildren
end

class collectGotoTargets targets =
object
  inherit nopCilVisitor

  method vstmt s =
    begin match s.skind with
    | Goto(s', _) -> targets := ISet.add !s'.sid !targets
    | _ -> ()
    end;
    DoChildren
end

class removeUnusedLabels targets =
object
  inherit nopCilVisitor

  method vstmt s =
    if not (ISet.mem s.sid targets) then
      s.labels <- [];
    DoChildren
end

let parse fname =
  Errormsg.hadErrors := false;
  let cilfile = Frontc.parse fname () in
  if !Errormsg.hadErrors then
    raise Parsing.Parse_error;
  cilfile

let build_namespaces cilfile =
  let noinline (Attr(n,_)) = n = "noinline" in
  iterGlobals cilfile (fun g ->
    match g with
    | GFun(fd,_) ->
      if (List.length (extract_function_attributes noinline fd)) = 0 then
        Hashtbl.add func_defs fd.svar.vname fd
    | _ -> ()
  )

let is_interesting_function g =
  let check_function name loc =
    if not (Hashtbl.mem builtinFunctions name) then
      let base = Filename.basename loc.file in
      List.for_all (fun fname -> base <> (Filename.basename fname)) !skipfiles
    else
      false
  in
  match g with
  | GVarDecl(vi,_) when isFunctionType vi.vtype ->
    check_function vi.vname (get_globalLoc g)
  | GFun(fd,_) ->
    check_function fd.svar.vname (get_globalLoc g)
  | _ -> true

(*********
 *
 *********)

let main () = begin
  let usage   = "Usage: " ^ Sys.argv.(0) ^ " [options] filename fut" in
  let options = Arg.align options in
  let args = ref [] in
  Arg.parse options (fun x -> args := x :: !args) usage;

  let genome, tstmts =
    match !genome, !tstmts with
    | Some(_), Some(_) ->
      failwith "cannot specify both --genome and --transform"
    | Some(bits), None ->
      let genome = ref [] in
      String.iter (fun c ->
        if c = '0' then
          genome := false :: !genome
        else
          genome := true :: !genome
      ) bits;
      List.rev !genome, ""
    | None, Some(stmts) -> [], stmts
    | _ -> [], ""
  in

  let fname, fut =
    match List.rev !args with
    | [fname; fut] -> fname, fut
    | _ ->
      prerr_string "ERROR: you must provide both file and function names\n\n";
      Arg.usage options usage;
      exit 1
  in

  Cil.initCIL ();
  let cilfile = parse fname in
  build_namespaces cilfile;
  let futfd =
    foldGlobals cilfile (fun fd g ->
      match g with
      | GFun(fd', _) when fd'.svar.vname = fut -> fd'
      | _ -> fd
    ) dummyFunDec
  in

  ignore (visitCilFunction (new flattenVisitor) futfd);
  ignore (Inline.inline futfd func_defs);
  ignore (visitCilFunction (new flattenVisitor) futfd);
  ignore (visitCilFunction (new numVisitor) futfd);
  ignore (visitCilFunction (new removeGotoNextLine) futfd);

  let gotoTargets = ref ISet.empty in
  ignore (visitCilFunction (new collectGotoTargets gotoTargets) futfd);
  ignore (visitCilFunction (new removeUnusedLabels !gotoTargets) futfd);

  if !debug then begin
    let chan = if !outfile = "" then stdout else open_out !outfile in
    dumpFile defaultCilPrinter chan fname cilfile;
    if !outfile <> "" then
      close_out chan
  end;

  let available = 
    if !inlineonly then
      ISet.empty
    else
      let _ = Transform.init !debug cilfile in
      Transform.apply futfd ISet.empty
  in
  if not !inlineonly then begin
    let tstmts =
      if tstmts = "all" then
        available
      else if (List.length genome) > 0 then
        List.fold_left2 (fun tstmts b stmt ->
          if b then ISet.add stmt tstmts else tstmts
        ) ISet.empty genome (ISet.elements available)
      else
        List.fold_left (fun set s ->
          ISet.add (int_of_string s) set
        ) ISet.empty (Str.split (Str.regexp ",") tstmts)
    in
    let _ = Transform.apply futfd tstmts in
    if !changevar <> "" then
      iterGlobals cilfile (function
        | GVar(vi, ini, loc) when vi.vname = !changevar ->
          ini.init <- Some(SingleInit(one))
        | _ -> ()
      )
  end;

  if !get_stmts then
    ISet.iter (fun i -> Printf.printf "%d\n" i) available;

  if not !get_stmts || !outfile <> "" then begin
    let globals = List.filter is_interesting_function cilfile.globals in
    let cilfile = {cilfile with globals = globals} in

    let chan = if !outfile = "" then stdout else open_out !outfile in
    dumpFile defaultCilPrinter chan fname cilfile;
    if !outfile <> "" then
      close_out chan;
  end

end ;;

main () ;;

