from collections import defaultdict
from contextlib import contextmanager
from subprocess import check_call, Popen, PIPE, STDOUT
import os
import sys
import tempfile
import time

stats = defaultdict( lambda: 0.0 )

class ContextList:
    def __init__( self ):
        self.managers = list()

    def __enter__( self ):
        return self

    def __exit__( self, typ, exc, trace ):
        suppress = False
        for manager, val in reversed( self.managers ):
            try:
                suppress = manager.__exit__( typ, exc, trace )
                if suppress:
                    typ, exc, trace = None, None, None
            except Exception as e:
                typ, exc, trace = sys.exc_info()
                suppress = False
        return suppress or typ is None

    def append( self, manager ):
        val = manager.__enter__()
        self.managers.append( ( manager, val ) )
        return val

def check_executable( cmd, handler = lambda: exit( 2 ) ):
    try:
        p = Popen( cmd, stdout = PIPE, stderr = STDOUT )
        if len( p.communicate()[ 0 ].strip() ) == 0:
            raise Exception
    except:
        print >>sys.stderr, "command %s not found" % ( cmd[ 0 ] )
        handler()

def get_golden( shader, size, scale, grayscale = False, perspective = False ):
    root = os.path.abspath( os.path.dirname( sys.argv[ 0 ] ) )
    ppmfile = "%s-%s-%d" % ( shader, size, scale )
    if grayscale:
        ppmfile += "-gray"
    if perspective:
        ppmfile += "-persp"
    ppmfile += ".ppm"
    return os.path.join( root, "..", "golden", ppmfile )

@contextmanager
def mktemp( suffix = ".ppm" ):
    tmp = tempfile.NamedTemporaryFile( suffix = suffix )
    tmp.close()
    try:
        yield tmp.name
    finally:
        if os.path.exists( tmp.name ):
            os.remove( tmp.name )

def show_stats( fh = sys.stderr ):
    if len( stats ) > 0:
        total = sum( stats.values() )
        width = max( map( len, stats.keys() ) ) + 1
        fmt = "%%-%ds %%6.3f (%%5.2f)%%%%" % width
        for stat in sorted( stats ):
            pct = 100 * stats[ stat ] / total
            print >>fh, fmt % ( stat + ":", stats[ stat ], pct )

@contextmanager
def timer( key ):
    start = time.time()
    try:
        yield
    finally:
        stop = time.time()
        stats[ key ] += stop - start

