open Cil
open Utils

module E = Errormsg
;;
E.logChannel := stderr
;;

let opt v s = v := Some(s)

let outfile = ref None

let options = [
  "-o", Arg.String(opt outfile), "file output file to generate"
]

let get_pure_functions cilfile =
  let is_api3d = filter_api3d () in
  let pure = Hashtbl.create 257 in
  iterGlobals cilfile (fun g ->
    match g with
    | GVarDecl(vi, _) when is_api3d g -> Hashtbl.add pure vi ()
    | GFun(fd, _) when is_api3d g -> Hashtbl.add pure fd.svar ()
    | _ -> ()
  ) ;
  fun instr ->
    match instr with
    | Call(_, Lval(Var(vi), NoOffset), _, _) -> Hashtbl.mem pure vi
    | _ -> false

class zero_init =
  let rec initialize accum lv =
    let assign e =
      Set(lv, e, locUnknown) :: accum
    in
    let t = unrollType (typeOfLval lv) in
    match t with
    | TVoid(_)              -> accum
    | TInt(ik, _)           -> assign (kinteger ik 0)
    | TFloat(fk, _)         -> assign (Const(CReal(0.0, fk, None)))
    | TPtr(_)               -> assign (mkCast zero t)
    | TEnum(ei, _)          -> assign (mkCast zero t)
    | TComp(ci, _)          ->
      List.fold_left (fun accum fi ->
        initialize accum (addOffsetLval (Field(fi, NoOffset)) lv)
      ) accum (List.rev ci.cfields)
    | TNamed(ti, _)         ->
      E.s (E.error "unrollType did not unroll type %a" d_type t)
    | TFun(rt, args, va, _) ->
      E.s (E.error "should not have lvals of function type: %a" dn_lval lv)
    | TBuiltin_va_list(_)   ->
      E.s (E.error "va_list not supported")
    | TArray(t', len, _) ->
      let rec init_array accum i =
        let accum =
          initialize accum (addOffsetLval (Index(integer (i-1), NoOffset)) lv)
        in
        if i > 1 then init_array accum (i-1) else accum
      in
      try
        init_array accum (lenOfArray len)
      with LenOfArray -> accum
  in
object
  inherit nopCilVisitor

  method vfunc fd =
    let inits =
      List.fold_left (fun inits vi ->
        initialize inits (var vi)
      ) [] (List.rev fd.slocals)
    in
    fd.sbody.bstmts <- mkStmt (Instr(inits)) :: fd.sbody.bstmts;
    SkipChildren
end

let main () = begin
  let usage   = "Usage: " ^ Sys.argv.(0) ^ " [options] filename" in
  let options = Arg.align options in
  let args = ref [] in
  Arg.parse options (fun x -> args := x :: !args) usage;

  let fname =
    match List.rev !args with
    | [fname] -> fname
    | _ ->
      prerr_string "ERROR: you must provide a filename\n\n";
      Arg.usage options usage;
      exit 1
  in

  let cilfile = Frontc.parse fname () in
  if !Errormsg.hadErrors then
    exit 2;
  cilfile.globals <-
    List.filter (fun g -> not (filter_builtin g)) cilfile.globals;
  visitCilFileSameGlobals (new zero_init) cilfile;
  Deadcodeelim.callHasNoSideEffects := get_pure_functions cilfile;
  Cfg.computeFileCFG cilfile;
  Deadcodeelim.dce cilfile;
  Rmtmps.removeUnusedTemps cilfile;

  let chan, close =
    match !outfile with
    | Some(outfile) ->
      let chan = open_out outfile in
      chan, fun () -> close_out chan
    | None -> stdout, fun () -> ()
  in
  lineDirectiveStyle := Some(LinePreprocessorOutput);
  dumpFile defaultCilPrinter chan fname cilfile;
  close ()
end ;;

main () ;;

